program CDSize;
{ A demonstration of module EDirTree, XtdFPC package
                             (C) by Marco van de Voort 1998

Of all demos, this is the one I use most.

The program scans directories/files passed as 1st parameter.
The first parameter has two forms:

1 - textfile with directory and file names
            (if first character of parameter 1 is '@')
2 - one directory (or filename)

The second, optional parameter is a value for the clustersize, default is none.

The output is the size in bytes, the number of directories and files scanned,
per directory scanned (if multiple directories in the textfile, in that case
overall stats are also printed)

Used to estimate how big directories on a 16k clustersize disk will be on a 2k
clustersize CDROM, in order to get the cdrom as full as possible.

(I also use this program to convince neighbours, friends and familymembers to
1 - partition their harddisk
2 - Use FAT32 (If they don't use kinky applications)
iow anything but 32k clusters. My favorite partition scheme is 1 Gb FAT16 for
 Win95. (speed reasons), and the rest FAT32, unless it is a really huge HD.)

Assume you want to make a CD with directories c:\games,c:\apps and C:\modem.
An index is made with Indexer.mod, and is called c:\index.txt.

You can include a root directory in this file, but you can also include individual and
directories names, like c:\autorun.inf, c:\windows etc. Wildcards are supported.

The textfile would simply be :

c:\games
c:\apps
C:\modem
c:\index.txt
c:\autorun.*

and the commandline would be CdSize @textfile.txt 2048

You can also specify the clustersize in a second parameter in the textfile:

d:\
d:\ 512
d:\ 1024
d:\ 2048
d:\ 4096
d:\ 8192
d:\ 16384
d:\ 32768
d:\ 65536

The totals will be nonsense, but you can see the effect of a clustersize on
your directory tree.

Lines without clustersize specification take the clustersize specified on
the commandline if existant, else 0 (don't round to clustersize, just add)

When burning a CD, the returned size in bytes which should be smaller
than 681 million. Also subtract about 1/2 a MB for every 1000 files above 1000.
}

Uses Dos,Strings,       EPasStr,EDirTree,EFIO;
{$Mode FPC}

CONST CR=CHR(13)+CHR(10);

      Usage=CR+CR+
 'Computes the total size of directories+files, stored in INFILE (ASCII).'+CR+
 'The second optional parameter is the clustersize in bytes, default 2048.'+CR+
 'Usage : CDSIZE <INFILE> [clustersize]'+ CR +CR+
 #9'E.g.'#9'cdsize @c:\newcd.txt 4096'#9'OR'+CR+
 #9#9'cdsize c:\windows 4096';

PROCEDURE ShowRecord(Dir:PathStr;Fil:SearchRec);
{Do nothing, we are only interested in the cumulatives, EDirTree keeps those
    for us. }

BEGIN
END;

VAR Param1,Param2,                      {Strings}
    Par22         : PathStr;

    Bytes         : CARDINAL;           {Totals}
    Files,Dirs    : CARDINAL;
    F             : Text;               {Textfile where paths are read from}
    D             : SearchRec;          {For wildcard expansion}
    DefClus,                            {default clustersize(commandline or 0) }
    Code          : LONGINT;            {Variable used for multiple purposes}
    P1,P2         : PChar;              {Used to interface with EDirTree}

PROCEDURE DoOne(LocalStat:BOOLEAN);

BEGIN
   Param1:=FExpand(Param1);
   FindFirst(Param1,AnyFile-VolumeId,D);   {Wildcard handling}
   IF DosError=0 THEN
    BEGIN
     REPEAT
      Par22:=Param1;                 {Create full path from Param1 and D.Name}
      {$IFDEF Linux}
       Code:=RCharPos(Par22,'/');
      {$ELSE}
       Code:=RCharPos(Par22,'\');
      {$ENDIF}
      IF Code<>-1 THEN              {Remove last backslash and everything behind it}
       Par22[0]:=CHR(Code-1);
      Par22:=Par22+'\'+D.Name;
      Writeln(Par22);               {Output newly created path}
      IF (D.attr and directory)>0 THEN       {Directory? Then recursive search with EDirTree}
       BEGIN
        StrPCopy(P1,Par22); StrCopy(P2,'*.*');
        EDirTree.FileScan(P1,P2,@ShowRecord);
        INC(Files,EDirTree.FoundCount);      { Update summuries}
        INC(Bytes,EDirTree.TotalBytes);
        INC(Dirs, EDirTree.FoundDirs+1);     { +1=current dir}
        IF LocalStat THEN
         BEGIN
          Write(' C. siz:',EdirTree.ClusterSize:5,' ',EDirTree.TotalBytes:15,' bytes in ');
          WriteLn(EDirTree.FoundCount:5,' files and ',EDirTree.FoundDirs:5,' directories.');
         END;
       END
     ELSE
      BEGIN
       INC (Files);
        IF EdirTree.ClusterSize=0 THEN
        Code:=D.Size
       ELSE
        Code:=(1+(D.Size DIV EDirTree.ClusterSize))*EdirTree.ClusterSize;
       INC (Bytes,Code);
       IF LocalStat THEN
        BEGIN
         Write(' C. siz:',EdirTree.ClusterSize:5,' ',Code:15,' bytes in ');
         WriteLn(1:5,' files');
        END;
     END;
    IF LocalStat THEN
     Writeln;
    FindNext(D);
    UNTIL DosError<>0;
   END;
END;

PROCEDURE DoFile;               {Read textfile with paths}
BEGIN
 IF NOT FileExists(Param1) THEN
  BEGIN
   Writeln('Error : File doesn'#39't exist.');
   HALT
  END;
 Assign(F,Param1);
 Reset(F);
 REPEAT
  Readln(F,Param2);                                 {Read a line}
  ItemS(Param1,Param2,' ',0);
  IF (Param1[1]<>';') AND (Param1[0]<>#0) THEN      {Comment of empty line}
   BEGIN
    ItemS(Par22,Param2,' ',1);                      {split into two parts}
    IF Length(Par22)=0 THEN                         {Handle clustersize if exists}
     EDirTree.ClusterSize:=DefClus
    ELSE
     Val(Par22,EDirTree.ClusterSize,Code);
    DoOne(TRUE);                                    {Do size-calculation including wildcard expansion}
   END;
  UNTIL EOF(F);
 Close(F);
END;

BEGIN
 DefClus:=0;                        { Default Clustersize=0, no rounding to clustersize}

 GetMem(P1,260);  GetMem(P2,260);       {For interfacing with EDirTree}

 Write ('CDSIZE 1.0, EXE is PD, source is (modified) GPL by Marco van de Voort,'+cr+' see xtdfpc/docs/copying.txt.'+Cr);

 IF ParamCount <1 THEN                  {Parameters?}
  BEGIN
   WriteLn(Usage);
   HALT;
  END;
 Param1:=ParamStr(1);                  {Param 1}
 IF ParamCount=2 THEN
  BEGIN
   Param2:=ParamStr(2);                {Param 2 if existant}
   Val(Param2,DefClus,Code);           {load param 2 to default clustersize}
   EDirTree.ClusterSize:=DefClus;
   IF Code<>0 THEN
    BEGIN
     Writeln('Invalid clustersize ',Param2);
     HALT;
    END;
  END;

 Files:=0; Bytes:=0; Dirs:=0;       {Init global stats}
 EDirTree.DirsToo:=FALSE;           {Don't report directories back}
 Writeln;

 IF Param1[1]='@' THEN              {Reading one directory, or bunch of dirs from file?}
  BEGIN
   Delete(Param1,1,1);              {File, delete @}
   DoFile;                          {Read file}
  END
 ELSE
   DoOne(FALSE);                    {One dir/file only}

 {Write global stats}

 Writeln;
 Write(Bytes,' bytes in ',Files,' files and ');
 WriteLn(Dirs,' directories.'); Writeln;

 {Clean-up}
 Freemem(P1,260);  Freemem(P2,260);
END.
