PROGRAM Dictionary;
{$M 65000 0 650000} { Some recursive procedures}

{
Compile with -Ch10000000 at least. Very memory hungry!

Dictionary, part of the XtdLib/FPC package.
   (C) Marco van de Voort 1997

  All rights of Borland version donated to the SWAG organisation.

  Original (SWAG author, UNKNOWN) comments:
  simple dictionary using a btree.
  The program reads in an ASCII file with one word per line and stores the
  words in an btree.
  A btree is something like binary tree but every node
  can have more than two descent nodes. This is done by a linked list.

  This method has two advantages:
    * when a word is wrong you can easily give some proposes how the word
      is written correctly (just change the path in the tree a little)
    * bigger dict. may save space. E.g "base, basicly, basement" etc.
      share the same path on the first three niveaus.

  ATTENTION! I don't free any mem I've allocated. This is done by the
  heap manager (i.e. he allocates large blockes and releases them ) when
  the program ends. But this can be added easily.

  Also, there is no PROCEDURE included that deletes words (I don't need it in
  my project). I suggest it is NOT that easy TO add such a PROCEDURE but
  have a try ;-))

  ---------------------------------------------------------------------
MvdV comments:
   -  It is an ordinary binary list isn't it? Ok, normally using the nodes right
       and left are considered to be the same level as the node itself. But that's
       only a point of view?
   -  Space save argument is nonsens. Two pointers per char!(8 bytes!)
      Every character should be used 9 times to get even.
   -  Instead of a nicely textfile, with input on each line, a standard
      textfile is used. I use XTDLIB.DOC as a test. Hmm, maybe I should use
      4dos.doc next :-)
      I tried 00global.txt (800k, index simtel CD set), but it went out of
      memory. (both TopSpeed version and Borland, works with the FPK version)
      (That's TopSpeed. I fed the FPC version over 16 MB, I used modified
      FindFile to find all textfiles on my harddisk)
   -  This program has no output! So I added an output procedure.
      The dictionary is spooled to a textfile, with a word per line.
      It's sorted because the list is kept sorted.
   -  The original didn't implement building a sorted list, allthough it
      was only 4 extra statements or so. I did :-)
   -  Added global deallocate to free memory.
   -  DeleteWord was more difficult than I thought, but not as much as
       the original author claims. I had a more simple approach, but the
       current approach doesn't use temporary variable arrays.
       (original idea was scanning for the word, and saving a pointer for
        each character. The size of the array would however limit the size
        of the word, so I chose this algoritm. Limited by
        string size only.
 (BP/TS:) Rough estimation of max words : 5000-10000 depending on "double" words.
        It's a demo, not an application, requires an extender to work
         properly. Memory usage is too heavy.
 (FPK:) 100.000 entries is about 5-6 MB, so if you keep in mind that the DPMI
         version of FPK can allocate about 64MB, you could reach up to 750000-
         1000000 words, less or more depending on how many words are shared.
         (sister and sisters is only one character extra, 9 bytes. The more
          overlapping like this, the more words you get in a MB. The more words
          you have in memory, the higher is the change that these kind of
          overlaps happen.
   -  Memory allocation seems to work, afterwards all bytes free again.

How to use it? Make a list of all your textfiles per language,
 path inclusive. Run Dictionary with one of the files as a parameter.
 The system loads your first file into memory, and then you type
 templ <filename>

with <filename>=the textfile you created.

Now type "end", and your rough dictionary is in newdict.DOC

The reason I was interested in this, was because I wanted to build a packer for
english (and maybe a separate one for other languages) with a fixed tables.

Somebody told me that an average language teacher which studied that language
on university level, maybe even got promoted, knows about 30000 words in that
language. Two bytes can hold a value 65535.
The idea is to replace known words by a two byte code, and to leave unknown
words untouched, but prefixed with some escape code. A different escape code
could be used for RLE. And after that LZW compression overall. I curious if
that'll work!

You could put one huge dictionary on disk with some indexes, and compress all
textfiles with it.
}

Uses EPasStr,EFIO,ELib;

CONST
      CrLf         = CHR(13)+CHR(10); { Ease helptext creation }
      { Displayed when HELP is typed in interrogation mode }
      { Indentation mangled to avoid wrapping }
HelpMenu='END               : end interrogation mode.'+CrLf+
         'DEL <word>        : delete word <word>'+CrLf+
         'HELP              : this menu'+CrLf+
  'LOAD <File>       : merges file <filename> with current dictionary'
         +CrLf;
  Help2=
   'TEMPL <File>      : merges data of (several) files in textfile <File>'
         +CrLf+
         'OUTPUTFILE <File> : Set name of output file to <File>'+CrLf+
         'SAVE <File>       : Save all data to the file'+CrLf+
         '<word>            : Does <word> exist?'+CrLf+ CrLf;

     { The main type. 10 bytes, but 16 are allocated under TopSpeed,
        don't know about Borland. }
TYPE  PNode     = ^TNode;
      TNode   = RECORD
               Character : CHAR;    { the current Character }
               WordEnd   : BOOLEAN; { is this char. the last of one word?}
               right,down: PNode;   { right: points to next char on the
                                              same niveau
                                       down : points to the next char in
                                              word }
                  END;


VAR BTree: PNode;                       { our tree }

PROCEDURE Memo;
{ Memory available. Specially meant to see if another "LOAD" is usefull.
 Remember when you crash (memory) everything will be gone!}
BEGIN
 WriteLn('Memory available :',MemAvail);
END;

PROCEDURE CreateBTree;
{ just initalizes the tree w/ a dummy element }
BEGIN
  BTree:=NIL;
  NEW(BTree);
  BTree^.Character:=CHR($1A); { #$1A is END-OF-FILE. shouldn't be used in any word }
  BTree^.right:=NIL;
  BTree^.down:=NIL;
  BTree^.WordEnd:=TRUE;
  { The dummy element can be deleted once words have been added. It isn't
  done yet, not worth the trouble. }
END;

FUNCTION GetNode(Character: CHAR; LevelPtr: PNode): PNode;
{ returns the node in Level "LevelPtr" that contains "Character".
  if there is no node, it is created, in an alphabetical way.
  This is the most import procedure. It builds the BTree!}
VAR P,p2: PNode;

BEGIN { All allocation is done in this procedure, so one check is enough}
  IF MaxAvail<SIZEOF(TNode) THEN { Safety measure. Don't delete, the
                                         system will lock up very tight, and
                                         even crash Windows 95!}
    BEGIN
    WRITELN('Out of Memory, trying to spool memory would crash system. So bye bye!');
    HALT;
    END;
  IF LevelPtr=NIL THEN { New level}
   BEGIN
    NEW(P);
    P^.right:=NIL;
    P^.down:=NIL;
    P^.Character:=Character;
    P^.WordEnd:=FALSE;
    GetNode:=P;
   END
   ELSE    { Search for a pointer in this level, if not found, create
              a new one.}
   BEGIN
    P:=LevelPtr;
    WHILE (P^.right<>NIL) AND (P^.Character<>Character) DO
     P:=P^.right;
    IF P^.Character=Character THEN { Found! }
     GetNode:=P
      ELSE
       BEGIN
       p2:=LevelPtr;    { Search level again}
       P:=p2;
       WHILE (p2<>NIL) AND (p2^.Character<Character) DO
        { Search for a place to insert}
        BEGIN
        P:=p2;
        p2:=p2^.right
        END;
       IF p2=P THEN             { Maybe insert before first pointer,
                                   or only one pointer}
        BEGIN
         NEW(p2);               { Create a new pointer}
         p2^.right:=P^.right;   { link it in }
         P^.right:=p2;
         IF P^.Character<Character THEN { Insert after 1st pointer}
          P:=p2
         ELSE
          BEGIN
          p2^.Character:=P^.Character; { Insert before 1st pointer}
          p2^.WordEnd:=P^.WordEnd;     { move record}
          p2^.down:=P^.down;
          END;
         P^.Character:=Character;       { add new data to "new" record}
         P^.down:=NIL;
         P^.WordEnd:=FALSE;
         END
        ELSE
         BEGIN
         p2:=P^.right; { Save link }
         NEW(P^.right);   { Create new link }
         P:=P^.right;     { and follow it }
         P^.right:=p2;    { Make link to next record }
         P^.Character:=Character; { rest of record}
         P^.down:=NIL;
         P^.WordEnd:=FALSE; { Is set by calling procedure if necessary.}
         END;

        { p^.right is NIL! }

       GetNode:=P; { Return the (newly created or found) pointer }
      END; {if}
  END; { if }
END;

PROCEDURE InsertWord(WOORD: STRING);
{ inserts the word "WOORD" into btree }
VAR P1,p2,p3: PNode;
    i: WORD;
BEGIN
  IF  Length(WOORD)<2 THEN EXIT;
  p2:=BTree;
  FOR i:=1 TO Length(WOORD) DO
   BEGIN
    P1:=GetNode(WOORD[i],p2);        { Search for pointer on this level}
    IF p2=NIL THEN                  { Make link of past level to newly
                                       created record }
     p3^.down:=P1;
    p3:=P1;
    p2:=P1^.down;                   { Search next level}
    END;
  P1^.WordEnd:=TRUE;
END;

FUNCTION ProofWord(WOORD: STRING): BOOLEAN;
{ returns true if "WOORD" is in our dictionary. }
VAR P1,p2: PNode;
    I: WORD;
BEGIN
  IF WOORD[0]=CHR(0) THEN BEGIN ProofWord:=FALSE; EXIT; END;
  P1:=BTree;
  I:=1;
  WHILE (P1<>NIL) AND (Length(WOORD)>=I) DO
   BEGIN
    WHILE (P1^.right<>NIL) AND (P1^.Character<>WOORD[I]) DO
      P1:=P1^.right; { Search current level for character}
    IF P1^.Character=WOORD[I] THEN
     BEGIN
      INC(I); { If found, then move a level down}
      p2:=P1;
      P1:=P1^.down;
      END
    ELSE P1:=NIL;       { else bye bye! }
   END;
  IF ((I-1)=Length(WOORD)) AND (p2^.WordEnd) THEN
   ProofWord:=TRUE
  ELSE
   ProofWord:=FALSE;
END;

PROCEDURE KillWord(Woord : String);
{ Based on Proofword. Contains a modified version of proofword which
   finds the last character. If last character has no links, then
   the record is deleted. The last character is also deleted from the
   string, and the search for the (new) last character is repeated.
   Advantage of this method is the possible unlimited size of the string.
   If Woord was a PChar, the length of a word could be 64K, without
   problems. Disadvantage is that the tree is parsed maximal
   Length(Woord) times. With an array of pointers the parsing could
   be limited to 1 }

VAR Len : WORD;
    P1,P2,P3: PNode;
    I    : WORD;

BEGIN
 UpperCase(Woord);
 Len:=Length(Woord);    { Save original length word }
 P2:=NIL;
 REPEAT
  P1:=BTree;               { Start at beginning of tree }
  I:=1;
  WHILE (P1<>NIL) AND (Length(Woord)>=I) DO { Repeat until end of word }
   BEGIN
    WHILE (P1^.right<>NIL) AND (P1^.Character<>Woord[I]) DO
    { search level for matching character}
     P1:=P1^.right;
    IF P1^.Character=Woord[I] THEN { Found on this level? Next level}
    BEGIN
      INC(I);
      P3:=P2;
      P2:=P1;
      P1:=P1^.down;
      END
     ELSE      P1:=NIL { Not found of end of word}
   END;
  IF (I-1)=Length(Woord) THEN { Entire word found?}
   BEGIN
   IF (I-1)=Len THEN

     P2^.WordEnd:=FALSE; { If last character we're deleting? Then WordEnd
                              becomes false since the word no longer exists}
    IF (P2^.right<>NIL) OR (P2^.down<>NIL) THEN { Links to other (parts) of
                                                 words?}
     EXIT
     ; { Yes, stop deleting }
    IF P2^.WordEnd = TRUE THEN       { another word stops here? }
     EXIT; { Yes, have to keep this record. Stop deleting}
    IF P3<>NIL THEN
    P3^.down:=NIL;                   { Delete reference to current level
                                         (in P2) }
    DISPOSE(P2);                     { Kill current level}
    DEC(Woord[0]);              { Remove last character of word}
    END
   ELSE EXIT;
 UNTIL Length(Woord)=0;
END;

PROCEDURE KillTree(R:PNode);
{ Recursive, delete a binary tree. All nodes except Root-node are deleted.
   Root-node can't be killed because of recursion.
   Requires extra procedure calling this one and an extra dispose. Not worth
   the trouble. Inserted the dispose into the main program }

BEGIN
 REPEAT
  IF R^.right <> NIL THEN
   IF (R^.right^.right = NIL) AND (R^.right^.down=NIL) THEN
    BEGIN { Record one "right" lower empty?}
    DISPOSE(R^.right); { Yes then kill it }
    R^.right:=NIL
    END
   ELSE
    KillTree(R^.right); { No, do recursive kill }
  IF R^.down <> NIL THEN { Same for "down"}
   IF (R^.down^.right = NIL) AND (R^.down^.down=NIL) THEN
    BEGIN { Record one more "down" empty?}
     DISPOSE(R^.down);    { Yes, kill it }
     R^.down:=NIL;
    END
   ELSE
    KillTree(R^.down);    { No, recursive kill }
  UNTIL ((R^.right=NIL) AND (R^.down=NIL))
  { repeat until this record is clean}
END;

VAR Deep   : WORD;              { Recursionlevel, and index in OutStr}
    OutStr : String;            { Words are "saved" in here }
    DictFile: Text;             { our ascii dictionary }

PROCEDURE Output(R:PNode);
{Recursive output of words to file. }
BEGIN
 IF Deep>255 THEN               { Safety, don't write behind string }
  Deep:=2;
 OutStr[Deep]:=R^.Character; { Save character of current node }
 IF R^.WordEnd THEN          { End of word? Then output! }
  BEGIN
  OutStr[0]:=CHR(Deep);  { Fix string length}
  WriteLn(DictFile,OutStr); { out with the stuff }
  END;
 IF R^.down = NIL THEN      { No "longer" words, or more characters in }
 BEGIN                      {                             this word?}
  IF R^.right <> NIL THEN    { No words starting to differ on this point?}
   Output(R^.right)
   END
 ELSE
  BEGIN
  INC(Deep);                { Deeper words}
  Output(R^.down);
  DEC(Deep);
  IF R^.right <> NIL THEN    { Longer words}
   Output(R^.right);
  END;
END;

{ The following three sets are used to split a line into words.
   The combination is quite powerfull. I tried some variants,
   and even managed to get the words out of a pascal source, so all
   keywords and words in a comment.

Moving characters from the other sets to dontscan increase quality (more
correct words, less random letters, but some words might not be found.
   }

       { Characters trimmed from both sides of the words }
CONST KillSet:CHARSET=['#',CHR(39),'?','0'..'9','/','-',':','\','*','"',
               '$','&','+','<','>','=',';','@','^','|','~',']',
               '[','_',#123,#125,'!',CHR(9),CHR(135)];
{ Don't scan words containing the following characters (after trimming)}
      DontScan:CHARSET=[CHR(2),'\','%','/',CHR(39),'0'..'9','$','&','"'];
      { Separators used }
      ItemString: CHARSET=['_','+' , '.' , ',' , ' ' , '(' , ')' , '-' , ':' , '=',
      '>', '<', '`', ']','[',';','@','*','!'];
      TotKillSet:CHARSET=[#0..#31,#155..#255];

      { Possible commands in interrogation mode}
      CommandStr=   'END@LOAD@DEL@HELP@OUTPUTFILE@TEMPL@ADD@SAVE';

      InitTxt=
' Dictionary v0.02, by Marco van de Voort (but based on a SWAG source)';
      Usage=
' A textfile as input is required, but is only read, not written'+CrLf;


PROCEDURE LoadOrMergeFile(FileName:STRING);
{ Load a textfile, split lines to words, and add words to the tree}

VAR A : WORD;
    S,S2 : String;

BEGIN
  Assign(DictFile,FileName);
  Reset(DictFile);
  WriteLn('Loading/Merging file ',FileName,' dictionary...');
  A:=0;
  WHILE NOT EOF(DictFile) DO
  BEGIN
   READLN(DictFile,S);       { read a line }
   { Change all tabs to chr(9)}
   ReplaceChar(S,CHR(9),' ');
   KillChrTot(S,TotKillSet); {delete some high and low ascii from the string}
   UpperCase(S);
   A:=0;
   REPEAT
    Item(S2,S,ItemString,A); { Split into rough words}
    IF Length(S2)>0 THEN
     BEGIN
     KillChar(S2,KillSet);      { Trim characters from back and front}
     KillBChar(S2,KillSet);     { of string}

     IF CharPosSet(S2,DontScan)=0 THEN
      BEGIN
       { CharPosSet prevents words like "I'll" to be added to tree,
          and other words with garbadge in it. }
      InsertWord(S2); { Insert word to the tree. One letter words are not
                          inserted, but ignored by InsertWord}
      END;
     INC(A);
     END;
    UNTIL Length(S2)=0; { Untill no more words from this read line}
  END {WHILE};
  Close(DictFile);
  Memo;
END;

PROCEDURE DOTempl(S:String);

VAR F :Text;

BEGIN
IF FileExists(s) THEN         {TEMPL}
             BEGIN
             Assign(F,S);
             Reset(F);
             WHILE NOT EOF(F) DO
              BEGIN
              READLN(F,S);
              IF FileExists(s) THEN
               LoadOrMergeFile(s);
              END;
              Close(F);
             END;
END;
VAR OutName : String;

PROCEDURE OutWithIt;
{ Output file tree to textfile, the real work is done by output.}

BEGIN
   Assign(DictFile,OutName);
   ReWrite(DictFile);
   Deep:=1; { Very important! Can't be done in
               Output because of recursion!}
   Output(BTree^.right); { Don't use the dummy record}
   Close(DictFile);
END;

{ Main code}
VAR s,S2,s3 : String;
    A       : WORD;
    Fl           : BOOLEAN;

BEGIN
  WriteLn(InitTxt);
  Memo;
  OutName:='newdict.doc'; { Default output name }
  IF ParamCount<1 THEN
   BEGIN
    WriteLn(Usage);      { No parameters}
    HALT;
   END;
  s:=ParamStr(1);
  CreateBTree;           { Set up tree}
  IF S[1]='@' THEN
   BEGIN
    Delete(S,1,1);
    DoTempl(S);
   END
  ELSE
   BEGIN
    IF FileExists(s) THEN
     LoadOrMergeFile(s)  { Load file }
    ELSE
     BEGIN
      WriteLn(Usage);
      HALT;
     END;
    END;
  WriteLn
 ('Dictionary maintenance/interrogation mode. Type Help for command index!');
  s:='';
  REPEAT
    Write('OK>');   { Interrogation mode }
    ReadLn(s);      { Read string from keyboard}
    UpperCase(s);        { Everything uppercase, the TREE is uppercase}
    Item(S2,s,[' '],0); { Get 1st substring separated by spaces}
    A:=0; Fl:=FALSE;
    REPEAT          { Compare the substring with the commands}
     Item(s3,CommandStr,['@'],A);
     IF Compare(S2,S3) THEN
      Fl:=TRUE             { Found!}
     ELSE
      INC(A);
    UNTIL Fl OR (Length(s3)=0);
    IF Fl THEN
     BEGIN { A is index in commandline }
     Delete(s,1,Length(S2)+1); { Determine argument,
                                Item(s2,s,[' '],1); is also possible }
     CASE A OF
       0 : Break;                         { END}
       1 : IF FileExists(s) THEN          { LOAD}
            LoadOrMergeFile(s)
           ELSE
            WriteLn('File not found');
       2 :  KillWord(s);                  { DEL}
       3 :  BEGIN
             WriteLn(HelpMenu);
             WriteLn(Help2);
            END;
       4 :  OutName:=s;                   {OutputName}
       5 : DoTempl(S);
        6: BEGIN
                            { ADD }
            upperCase(s);
            InsertWord(S);
            END;
        7 : BEGIN                          {SAVE}
            WriteLn('Saving file ',S);
            OutName:=S;
            OutWithit;
            END;
        END;
       END
    ELSE          { No keyword? Test of word exists in Tree}
     BEGIN
     Write('WORD ',s,' ');
     IF NOT ProofWord(s) THEN
      Write('NOT ');
     WriteLn('FOUND!');
     END;
    UNTIL 3=4;   { Go on forever, the exit wil do, saves a boolean:-) }
   OutWithIt;                   { Output tree to textfile}
   Writeln('Deleting tree from memory');
   { Deleting the tree from memory, nicely and cleanly, works, but is
   slow (2.5 MB full heap)
   KillTree(BTree);             { Remove tree from memory, slow }
   DISPOSE(BTree);              { Remove root-node}
   }
   Memo;
END. { Aargh, 644 lines already?}
