Unit ECompr;

INTERFACE

(* LZSS compression/decompression, core routines.
 TopSpeed 3.1 translation, part of XTDLIB/FPK package (C) 1998

Based on:
LZHUF.C (c)1989 by Haruyasu Yoshizaki, Haruhiko Okumura, and Kenji Rikitake.
All rights reserved. Permission granted for non-commercial use.
 * LZHUF.C English version 1.0
 * Based on Japanese version 29-NOV-1988
 * LZSS coded by Haruhiko OKUMURA
 * Adaptive Huffman Coding coded by Haruyasu YOSHIZAKI
 * Edited and translated to English by Kenji RIKITAKE
 * Translated to Modula-2 by Don Milne, June 1989.
 * TopSpeed M2 3.1 conversion by Marco van de Voort, june 1997
    Several datastructures dynamically allocated instead of static.
    Stripped down to bare compression/decompression routines only for
    generic use: DJARC.DEF DJARC.MOD, integral translation
    moved to DJARC2.MOD. Some files however were missing to compile
    DJARC2. (ArgHandler and FileNames)
 * FPK Pascal port by Marco van de Voort, August 1998
*)


(* Watch out! n in DJArc.Unpacker is size of UNCOMPRESSED data, while
   Pack.Unpacker is is the compressed size(as returned by Packer *)

PROCEDURE Packer(n: CARDINAL; in: ADDRESS; out: ADDRESS): CARDINAL;
PROCEDURE Unpacker(n: CARDINAL; in: ADDRESS; out: ADDRESS);

(* CoreArc.A imports data below, it's not interesting for the user,
    please beware: the constants are also fixed in CoreArc.A! Some
     others (in DJARC.MOD) are also duplicated in CoreArc.A!*)


CONST  N	    = 4096;    (* Size of string buffer *)
       F	    = 60;      (* Size of look-ahead buffer *)
       BigSize      = N+F+2*(N+1)+2*(N+256)+100;

TYPE texttype=ARRAY[0..N+F-1] OF CHAR;
     TPtr    =POINTER TO texttype;
     dadtype =ARRAY[0..N] OF INTEGER;
     dadptr  =POINTER TO dadtype;
     rtype   =ARRAY[0..N+256] OF INTEGER;
     rptr    =POINTER TO rtype;
     BigType =ARRAY[0..BigSize] OF BYTE;
     BigPtr  =POINTER TO BigType;

VAR textbuf     :    TPtr;                   (* 4156 *)
    lson,
    dad         :    dadptr;            (* 8194 each *)
    rson        :    rptr;                   (* 8706 *)
    mpos,mlen:INTEGER;

IMPLEMENTATION;

CONST BUFFSIZE   = 10240;
      FIONULLFILE=MAX(CARDINAL);

TYPE
     BufType = ARRAY[0..BUFFSIZE-1] OF CHAR;
     BufPtr  = POINTER TO BufType;

VAR
   inp,inbytes,outp:CARDINAL;
    codesize,textsize:LONGCARD;
    input,output:BufPtr;

(* LZSS Parameters *)

CONST
       THRESHOLD    = 2;
       NILPTR       = N;       (* End of tree's node  *)


(* Huffman coding parameters *)

CONST NCHAR    = (256 - THRESHOLD + F); (* character code (= 0.NCHAR-1) *)
      T	= (NCHAR * 2 - 1);       (* Size of table *)
      R	= (T - 1);	       (* root position *)
      MAXFREQ  = 8000H;		 (* update when cumulative frequency *)
					(* reaches this value *)

(* Tables FOR encoding/decoding upper 6 bits of
 * sliding dictionary pointer
 *)
(* encoder table *)
TYPE str64 = ARRAY[0..63] OF CHAR;
     str256 = ARRAY[0..255] OF CHAR;

CONST plen = str64(
	CHR(03H),CHR(04H),CHR(04H),CHR(04H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),
	CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H));

CONST pcode = str64(
	CHR(00H),CHR(020H),CHR(030H),CHR(040H),CHR(050H),CHR(058H),CHR(060H),CHR(068H),
	CHR(070H),CHR(078H),CHR(080H),CHR(088H),CHR(090H),CHR(094H),CHR(098H),CHR(09CH),
	CHR(0A0H),CHR(0A4H),CHR(0A8H),CHR(0ACH),CHR(0B0H),CHR(0B4H),CHR(0B8H),CHR(0BCH),
	CHR(0C0H),CHR(0C2H),CHR(0C4H),CHR(0C6H),CHR(0C8H),CHR(0CAH),CHR(0CCH),CHR(0CEH),
	CHR(0D0H),CHR(0D2H),CHR(0D4H),CHR(0D6H),CHR(0D8H),CHR(0DAH),CHR(0DCH),CHR(0DEH),
	CHR(0E0H),CHR(0E2H),CHR(0E4H),CHR(0E6H),CHR(0E8H),CHR(0EAH),CHR(0ECH),CHR(0EEH),
	CHR(0F0H),CHR(0F1H),CHR(0F2H),CHR(0F3H),CHR(0F4H),CHR(0F5H),CHR(0F6H),CHR(0F7H),
	CHR(0F8H),CHR(0F9H),CHR(0FAH),CHR(0FBH),CHR(0FCH),CHR(0FDH),CHR(0FEH),CHR(0FFH));

(* decoder table *)
CONST dcode = str256(
	CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),
	CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),
	CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),
	CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),CHR(00H),
	CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),
	CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),CHR(01H),
	CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),
	CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),CHR(02H),
	CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),
	CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),
	CHR(09H),CHR(09H),CHR(09H),CHR(09H),CHR(09H),CHR(09H),CHR(09H),CHR(09H),
	CHR(0AH),CHR(0AH),CHR(0AH),CHR(0AH),CHR(0AH),CHR(0AH),CHR(0AH),CHR(0AH),
	CHR(0BH),CHR(0BH),CHR(0BH),CHR(0BH),CHR(0BH),CHR(0BH),CHR(0BH),CHR(0BH),
	CHR(0CH),CHR(0CH),CHR(0CH),CHR(0CH),CHR(0DH),CHR(0DH),CHR(0DH),CHR(0DH),
	CHR(0EH),CHR(0EH),CHR(0EH),CHR(0EH),CHR(0FH),CHR(0FH),CHR(0FH),CHR(0FH),
	CHR(010H),CHR(010H),CHR(010H),CHR(010H),CHR(011H),CHR(011H),CHR(011H),CHR(011H),
	CHR(012H),CHR(012H),CHR(012H),CHR(012H),CHR(013H),CHR(013H),CHR(013H),CHR(013H),
	CHR(014H),CHR(014H),CHR(014H),CHR(014H),CHR(015H),CHR(015H),CHR(015H),CHR(015H),
	CHR(016H),CHR(016H),CHR(016H),CHR(016H),CHR(017H),CHR(017H),CHR(017H),CHR(017H),
	CHR(018H),CHR(018H),CHR(019H),CHR(019H),CHR(01AH),CHR(01AH),CHR(01BH),CHR(01BH),
	CHR(01CH),CHR(01CH),CHR(01DH),CHR(01DH),CHR(01EH),CHR(01EH),CHR(01FH),CHR(01FH),
	CHR(020H),CHR(020H),CHR(021H),CHR(021H),CHR(022H),CHR(022H),CHR(023H),CHR(023H),
	CHR(024H),CHR(024H),CHR(025H),CHR(025H),CHR(026H),CHR(026H),CHR(027H),CHR(027H),
	CHR(028H),CHR(028H),CHR(029H),CHR(029H),CHR(02AH),CHR(02AH),CHR(02BH),CHR(02BH),
	CHR(02CH),CHR(02CH),CHR(02DH),CHR(02DH),CHR(02EH),CHR(02EH),CHR(02FH),CHR(02FH),
	CHR(030H),CHR(031H),CHR(032H),CHR(033H),CHR(034H),CHR(035H),CHR(036H),CHR(037H),
	CHR(038H),CHR(039H),CHR(03AH),CHR(03BH),CHR(03CH),CHR(03DH),CHR(03EH),CHR(03FH));

CONST dlen = str256(
	CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),
	CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),
	CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),
	CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),CHR(03H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),CHR(04H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),CHR(05H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),CHR(06H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),CHR(07H),
	CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),
	CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H),CHR(08H));

TYPE FreqTable=ARRAY[0..T] OF CARDINAL;
     PrntTable=ARRAY[0..T+NCHAR] OF CARDINAL;
     FreqPtr  = POINTER TO FreqTable;
     PrntPtr  = POINTER TO PrntTable;

VAR freq:FreqPtr   (* cumulative freq table *);

(*
 * pointing parent nodes.
 * area [T.(T + NCHAR - 1)] are pointers for leaves
 *)

VAR prnt:PrntPtr;

(* pointing children nodes (son[], son[] + 1) *)
VAR son:FreqPtr;
    code,len:CARDINAL;
    getbuf,putbuf:CARDINAL;
    getlen,putlen:SHORTCARD;

(**************************** UTILITY ROUTINES *****************************)



PROCEDURE getc(VAR c:INTEGER):INTEGER;

VAR SizeToRead:CARDINAL;

BEGIN
     IF inp=inbytes THEN
      RETURN -1;
     END;
     c := INTEGER(ORD(input^[inp]));
     INC(inp);
     RETURN c;
END getc;

PROCEDURE putc(c:INTEGER);
BEGIN
     output^[outp] := CHR(CARDINAL(c)); INC(outp);
END putc;

(* initialize freq tree *)

PROCEDURE StartHuff();

VAR i,j:INTEGER;

BEGIN
     i:=0;
     WHILE i < NCHAR DO
	 freq^[i] := 1;
	 son^[i] := i + T;
	 prnt^[i + T] := i;
	 INC(i);
     END;
     i := 0; j := NCHAR;
     WHILE (j <= R) DO
	 freq^[j] := freq^[i] + freq^[i + 1];
	 son^[j] := i;
	 prnt^[i] := j;
	 prnt^[i + 1] := j;
	 INC(i,2); INC(j);
     END;
     freq^[T] := 0FFFFH;
     prnt^[R] := 0;
END StartHuff;

(************************ COMPRESSION ROUTINES ***************************)

PROCEDURE InitTree();  (* Initializing tree *)

VAR i:CARDINAL;

BEGIN
     i:=N+1;
     WHILE i<=N+256 DO rson^[i]:=NILPTR; INC(i) END; (* root *)
     i:=0;
     WHILE (i<N) DO dad^[i]:=NILPTR; INC(i) END;     (* node *)
END InitTree;

PROCEDURE Putcode(l:INTEGER; c:CARDINAL);	 (* output c bits *)
BEGIN
     putbuf := CARDINAL(BITSET(putbuf) + BITSET(c >> CARDINAL(putlen)));
     INC(putlen,SHORTCARD(l));
     IF putlen >= 8 THEN
	 putc(putbuf >> 8);
	 DEC(putlen,8);
	 IF putlen >= 8 THEN
	     putc(putbuf);
	     INC(codesize,2);
	     DEC(putlen,8);
	     putbuf := c << (CARDINAL(l) - CARDINAL(putlen));
	   ELSE
	     putbuf := putbuf << 8;
	     INC(codesize);
	 END;
     END;
END Putcode;

(* reconstruct freq tree *)

PROCEDURE reconst();

VAR i,j,k:INTEGER;
    f,l:CARDINAL;

BEGIN
     (* halven cumulative freq FOR leaf nodes *)
     i:=0; j:=0;
     WHILE i<T DO
	 IF son^[i] >= T THEN
	     freq^[j] := (freq^[i] + 1) >> 1;
	     son^[j] := son^[i];
	     INC(j);
	 END;
	 INC(i);
     END;

     (* make a tree : first, connect children nodes *)
     i:=0; j:=NCHAR;
     WHILE j<T DO
	 k := i + 1;
	 f := freq^[i] + freq^[k];
	 freq^[j] := f;
	 k:=j-1;
	 WHILE f < freq^[k] DO DEC(k) END;
	 INC(k);
	 l := (j - k) * 2;
	 Lib.Move(ADR(freq^[k]), ADR(freq^[k+1]), l);
	 freq^[k] := f;
	 Lib.Move(ADR(son^[k]), ADR(son^[k+1]), l);
	 son^[k] := i;
	 INC(i,2); INC(j);
     END;

     (* connect parent nodes *)
     i:=0;
     WHILE i<T DO
	 k := son^[i];
	 IF k >= T THEN
	     prnt^[k] := i;
	   ELSE
	     prnt^[k] := i;
	     prnt^[k + 1] := i;
	 END;
	 INC(i);
     END;
END reconst;

(* update freq tree *)

PROCEDURE update(c:INTEGER);

VAR i,j,k,l:INTEGER;

BEGIN
     IF freq^[R] = MAXFREQ THEN reconst() END;

     c := prnt^[c + T];
     REPEAT
	   (* keep the outer loop together so stupid compilers
	    * can optimize.
	    *)
	   LOOP
	       INC(freq^[c]); k:=freq^[c];
	       (* swap nodes to keep the tree freq-ordered *)
	       IF CARDINAL(k) > freq^[c + 1] THEN EXIT END;
	       c := prnt^[c];
	       IF c=0 THEN RETURN END;
	   END;

	   l := c + 1;
	   (* this is the inner loop -- unroll it a few times *)
	   LOOP
	       INC(l);
	       IF CARDINAL(k)<=freq^[l] THEN EXIT END;
	   END;
	   DEC(l);
	   freq^[c] := freq^[l];
	   freq^[l] := k;

	   i := son^[c];
	   prnt^[i] := l;
	   IF i < T THEN prnt^[i + 1] := l END;

	   j := son^[l];
	   son^[l] := i;

	   prnt^[j] := c;
	   IF j < T THEN prnt^[j + 1] := c END;
	   son^[c] := j;

	   c := l;
	   c := prnt^[c];
     UNTIL c=0;	     (* do it until reaching the root *)
END update;

PROCEDURE EncodeChar(c:CARDINAL);

VAR i:CARDINAL;
    j,k:INTEGER;

BEGIN
     i:=0; j:=0; k:=prnt^[c + T];

     (* search connections from leaf node to the root *)
     REPEAT
	   i := i>>1;
	   (*
	   IF node's address is odd, output 1
	   ELSE output 0
	   *)
	   IF ODD(k) THEN INC(i,8000H) END;
	   INC(j);
	   k := prnt^[k];
     UNTIL k = R;
     Putcode(j, i);
     code := i;
     len := j;
     update(c);
END EncodeChar;

PROCEDURE EncodePosition(c:CARDINAL);

VAR i:CARDINAL;

BEGIN
     (* output upper 6 bits with encoding *)
     i := c >> 6;
     Putcode(INTEGER(ORD(plen[i])), ORD(pcode[i]) << 8);

     (* output lower 6 bits directly *)
     Putcode(6, CARDINAL(BITSET(c)*BITSET(3FH)) << 10);
END EncodePosition;

PROCEDURE EncodeEnd();
BEGIN
     IF putlen<>0 THEN
	 putc(putbuf >> 8);
	 INC(codesize);
     END;
END EncodeEnd;

PROCEDURE DeleteNode(p:INTEGER);  (* Deleting node from the tree *)

VAR q:INTEGER;

BEGIN
     IF dad^[p] = NILPTR THEN RETURN END;  (* unregistered *)
     IF rson^[p] = NILPTR THEN
	 q := lson^[p];
       ELSIF lson^[p] = NILPTR THEN
	 q := rson^[p];
       ELSE
	 q := lson^[p];
	 IF rson^[q] <> NILPTR THEN
	     REPEAT
		   q := rson^[q];
	     UNTIL rson^[q] = NILPTR;
	     rson^[dad^[q]] := lson^[q];
	     dad^[lson^[q]] := dad^[q];
	     lson^[q] := lson^[p];
	     dad^[lson^[p]] := q;
	 END;
	 rson^[q] := rson^[p];
	 dad^[rson^[p]] := q;
     END;
     dad^[q] := dad^[p];
     IF rson^[dad^[p]] = p THEN
	 rson^[dad^[p]] := q;
       ELSE
	 lson^[dad^[p]] := q;
     END;
     dad^[p] := NILPTR;
END DeleteNode;

PROCEDURE InsertNode(r:INTEGER);  (* Inserting node to the tree *)

VAR i,p:INTEGER;
    c:CARDINAL;
    cmp:INTEGER;

BEGIN
     cmp := 1;
     p := N + 1 + INTEGER(textbuf^[r]);
     rson^[r] := NILPTR;
     lson^[r] := NILPTR;
     mlen := 0;

     LOOP
         IF (cmp >= 0) THEN
             IF rson^[p] <> NILPTR THEN
                 p := rson^[p];
               ELSE
                 rson^[p] := r;
                 dad^[r] := p;
                 RETURN;
             END
           ELSIF (lson^[p] <> NILPTR) THEN
             p := lson^[p];
           ELSE
             lson^[p] := r;
             dad^[r] := p;
             RETURN;
         END;
         i:=1;
         LOOP
             IF i=F THEN EXIT END;
             cmp := INTEGER(textbuf^[i+r])-INTEGER(textbuf^[i+p]);
             IF cmp<>0 THEN EXIT END;
             INC(i);
         END;
         IF (i > THRESHOLD) THEN
             IF (i > mlen) THEN
                 mpos := INTEGER(BITSET(r-p) * BITSET(N-1)) - 1;
                 mlen := i;
                 IF mlen>=F THEN EXIT END;
               ELSIF i = mlen THEN
                 c := INTEGER(BITSET(r-p) * BITSET(N-1)) - 1;
                 IF c < CARDINAL(mpos) THEN mpos:=c END;
             END;
         END;
     END;

     dad^[r] := dad^[p];
     lson^[r] := lson^[p];
     rson^[r] := rson^[p];
     dad^[lson^[p]] := r;
     dad^[rson^[p]] := r;
     IF rson^[dad^[p]] = p THEN
         rson^[dad^[p]] := r;
       ELSE
         lson^[dad^[p]] := r;
     END;
     dad^[p] := NILPTR;  (* remove p *)
END InsertNode;

PROCEDURE Encode();  (* Encoding/Compressing *)

VAR i,c,len,r,s,lastmatchlength:INTEGER;
    printcount:LONGCARD;

BEGIN
     textsize := 0;		   (* rewind and rescan *)
     printcount := 256;

     StartHuff();
     InitTree();
     s := 0;
     r := N - F;
     i:=s;
     WHILE i<r DO textbuf^[i] := ' '; INC(i) END;
     len:=0;
     LOOP
	 IF len>=F THEN EXIT END;
	 IF getc(c)<0 THEN EXIT END;
	 textbuf^[r + len] := CHR(c);
	 INC(len);
     END;
     i:=1; textsize:=LONGCARD(len);
     WHILE i<=F DO InsertNode(r - i); INC(i) END;
     InsertNode(r);
     REPEAT
	   IF mlen > len THEN mlen := len END;
	   IF mlen <= THRESHOLD THEN
	       mlen := 1;
	       EncodeChar(ORD(textbuf^[r]));
	     ELSE
	       EncodeChar(255 - THRESHOLD + mlen);
	       EncodePosition(mpos);
	   END;
	   i:=0; lastmatchlength:=mlen;
	   LOOP
	       IF i=lastmatchlength THEN EXIT END;
	       IF getc(c)<0 THEN EXIT END;
	       DeleteNode(s);
	       textbuf^[s] := CHR(c);
	       IF (s < F - 1) THEN textbuf^[s + N] := CHR(c) END;
	       s := INTEGER(BITSET(s+1) * BITSET(N-1));
	       r := INTEGER(BITSET(r+1) * BITSET(N-1));
	       InsertNode(r);
	       INC(i);
	   END;
	   INC(textsize,LONGCARD(i));

	   IF (textsize > printcount) OR (c<0) THEN
	       INC(printcount,256);
	   END;

	   LOOP
	       IF i>=lastmatchlength THEN EXIT END;
	       INC(i);
	       DeleteNode(s);
	       s := INTEGER(BITSET(s+1) * BITSET(N-1));
	       r := INTEGER(BITSET(r+1) * BITSET(N-1));
	       DEC(len);
	       IF len<>0 THEN InsertNode(r) END;
	   END;
     UNTIL len=0;
     EncodeEnd();
END Encode;

(************************ DECOMPRESSION ROUTINES **************************)

PROCEDURE GetBit():INTEGER;	(* get one bit *)

VAR i:INTEGER;

BEGIN
     WHILE getlen <= 8 DO
	 IF getc(i)<0 THEN i:=0 END;
	 getbuf := CARDINAL(BITSET(getbuf) + BITSET(i << INTEGER(8 - getlen)));
	 INC(getlen,8);
     END;
     i := getbuf;
     getbuf := getbuf << 1;
     DEC(getlen);
     RETURN ORD(INTEGER(i)<0);
END GetBit;



PROCEDURE GetByte():INTEGER;      (* get a byte *)

VAR i:CARDINAL;
    temp:INTEGER;

BEGIN
     WHILE getlen <= 8 DO
	 IF getc(temp)<0 THEN i:=0 ELSE i:=temp END;
	 getbuf := CARDINAL(BITSET(getbuf) + BITSET(i << (8 - CARDINAL(getlen))));
	 INC(getlen,8);
     END;
     i := getbuf;
     getbuf := getbuf<<8;
     DEC(getlen,8);
     RETURN (i >> 8);
END GetByte;

PROCEDURE DecodeChar():INTEGER;

VAR c:CARDINAL;

BEGIN
     c := son^[R];

     (*
      * start searching tree from the root to leaves.
      * choose node #(son[]) IF input bit = 0
      * ELSE choose #(son[]+1) (input bit = 1)
      *)
     WHILE c < T DO
	 IF getlen<>0 THEN
	     DEC(getlen);
	     IF getbuf>=8000H THEN INC(c) END;
	     getbuf := getbuf<<1;
	   ELSE
	     INC(c,GetBit());
	 END;
	 c := son^[c];
     END;
     DEC(c,T);
     update(c);
     RETURN INTEGER(c);
END DecodeChar;



PROCEDURE DecodePosition():INTEGER;

VAR i,j,c:CARDINAL;

BEGIN
     (* decode upper 6 bits from given table *)
     i := GetByte();
     c := ORD(dcode[i]) << 6;
     j := ORD(dlen[i]);

     (* input lower 6 bits directly *)
     DEC(j,2);
     WHILE j<>0 DO
	 DEC(j);
	 i := i<<1;
	 IF getlen<>0 THEN
	     DEC(getlen);
	     IF getbuf>=8000H THEN INC(i) END;
	     getbuf := getbuf<<1;
	   ELSE
	     INC(i,GetBit());
	 END;
     END;
     RETURN INTEGER(BITSET(c)+(BITSET(i)*BITSET(3FH)));
END DecodePosition;

PROCEDURE Decode(textsize:LONGCARD);  (* Decoding/Uncompressing *)

VAR i,j,k,r,c:INTEGER;
    count:LONGCARD;

BEGIN
     StartHuff();
     i:=0;
     WHILE i<(N-F) DO textbuf^[i] := ' '; INC(i) END;
     count:=0; r:=N-F;
     WHILE count < textsize DO
	 c := DecodeChar();
	 IF c < 256 THEN
	     putc(c);
	     textbuf^[r]:=CHR(c);
	     r := INTEGER(BITSET(r+1) * BITSET(N-1));
	     INC(count);
	   ELSE
	     i := INTEGER(BITSET(r-DecodePosition()-1) * BITSET(N-1));
	     j := c - 255 + THRESHOLD;
	     IF (r+j<N) AND (i+j<N) AND ((i+j<=r) OR (i>=r)) THEN
		 Lib.Move(ADR(textbuf^[i]),ADR(textbuf^[r]),j);
		 i:=r;
		 WHILE i<r+j DO
		     putc(INTEGER(ORD(textbuf^[i]))); INC(i);
		 END;
		 INC(r,j);
		 INC(count,LONGCARD(j));
	       ELSE
		 k:=i; INC(j,i);
		 WHILE k<j DO
		     c := ORD(textbuf^[CARDINAL(BITSET(k)*BITSET(N-1))]);
		     putc(c);
		     textbuf^[r] := CHR(c);
		     r := INTEGER(BITSET(r+1) * BITSET(N-1));
		     INC(count);
		     INC(k);
		 END;
	     END;
	 END;
     END;
END Decode;

PROCEDURE DecodeRaw(textsize:LONGCARD);

VAR c:INTEGER;
    count:LONGCARD;

BEGIN
     count := 0;
     WHILE count<textsize DO
	 IF getc(c)<0 THEN RETURN END;
	 putc(c);
	 INC(count);
     END;
END DecodeRaw;



PROCEDURE AllocateStruct;

BEGIN
 NEW(prnt); NEW(son); NEW(freq);
 NEW(textbuf); NEW(dad); NEW(rson); NEW(lson);
 inp:=0; outp:=0;
 code:=0;len:=0;
 getbuf:=0;putbuf:=0;
 getlen:=0;putlen:=0;
END AllocateStruct;

PROCEDURE DeAllocateStruct;

BEGIN
 DISPOSE(prnt); DISPOSE(son); DISPOSE(freq);
 DISPOSE(textbuf); DISPOSE(dad); DISPOSE(rson); DISPOSE(lson);
END DeAllocateStruct;


PROCEDURE Packer(n: CARDINAL; in: ADDRESS; out: ADDRESS):CARDINAL;

BEGIN
 AllocateStruct;
 input:=in; output:=out;
 inbytes:=n;
 Encode;
 DeAllocateStruct;
RETURN outp;
END Packer;

PROCEDURE Unpacker(n: CARDINAL; in: ADDRESS; out: ADDRESS);

BEGIN
 AllocateStruct;
 input:=in; output:=out;
 inbytes:=n;
 Decode(LONGCARD(n));
  DeAllocateStruct;
END Unpacker;

END DJArc.
