{****************************************************************************
*                                                                           *
*                                    MMX                                    *
*                                                                           *
*                                                                           *
* Language:             FPC Pascal v0.99.11+                                *
* Environment:          MS-DOS                                              *
* Supported extenders:  GO32V2                                              *
*                                                                           *
* Required switches:                                                        *
*                                                                           *
* Author:               Thomas Schatzl                                      *
* License:              LGPL, see the file COPYING.TXT, included in this    *
*                       distribution for details about the copyright.       *
* Date:                 12.05.1999                                          *
* Version:              Rel1                                                *
*                                                                           *
*      Send bug reports and feedback to tom_at_work@geocities.com           *
*                                                                           *
* Description:  Compatibility unit to the RTL. You only need this unit if   *
*               you want to compile old code using the RTL MMX unit. The    *
*               CPU unit is usually sufficient for any detection purposes.  *
*                                                                           *
*               Original code from the FreePascal RTL library, done by the  *
*               Free Pascal Development team.                               *
*																			*
*               Full compatibility to the RTL units (CPU and MMX)           *
*                                                                           *
*****************************************************************************}

unit mmx;

interface

type
	tmmxshortint = array[0..7] of shortint;
	tmmxbyte = array[0..7] of byte;
	tmmxword = array[0..3] of word;
	tmmxinteger = array[0..3] of integer;
	tmmxfixed = array[0..3] of fixed16;
	tmmxlongint = array[0..1] of longint;
	tmmxcardinal = array[0..1] of cardinal;
	{ for the AMD 3D }
	tmmxsingle = array[0..1] of single;

	pmmxshortint = ^tmmxshortint;
	pmmxbyte = ^tmmxbyte;
	pmmxword = ^tmmxword;
	pmmxinteger = ^tmmxinteger;
	pmmxfixed = ^tmmxfixed;
	pmmxlongint = ^tmmxlongint;
	pmmxcardinal = ^tmmxcardinal;
	{ for the AMD 3D }
	pmmxsingle = ^tmmxsingle;

const
	is_mmx_cpu : boolean = false;
	is_amd_3d_cpu : boolean = false;

{ sets all floating point registers to empty
(use this after mmx usage) }

procedure emms;

implementation

uses
	cpu;

{$ASMMODE ATT}

{ returns true, if the processor supports the mmx instructions }
function mmx_support : boolean;
begin
	mmx_support := cpu_supported(cHAVE_MMX);
end;

function amd_3d_support : boolean;
begin
	amd_3d_support := cpu_supported(cHAVE_3DNOW);
end;

procedure emms;assembler;
asm
	emms
end;

var
	oldexitproc : pointer;

procedure mmxexitproc;
begin
	exitproc:=oldexitproc;
	emms;
end;

begin
	if (cpu_supported(cHAVE_MMX)) then begin
		is_mmx_cpu := true;
        { the exit code sets the fpu stack to empty }
        oldexitproc:=exitproc;
        exitproc:=@mmxexitproc;
		is_amd_3d_cpu := cpu_supported(cHAVE_3DNOW);
	end;
end.
