{****************************************************************************
*                                                                           *
*                                  DPMI 0.9                                 *
*                                                                           *
*                                                                           *
* Language:             FPC Pascal 0.99.11                                  *
* Environment:          IBM PC (MS-DOS)                                     *
* Supported extenders:  GO32V2                                              *
*                                                                           *
* Required switches:                                                        *
*                                                                           *
* Author:               Thomas Schatzl                                      *
* License:              LGPL, see the file COPYING.TXT, included in this    *
*                       distribution for details about the copyright.       *
* Date:                 09.05.1999                                          *
* Version:              rel1rev7                                            *
*                                                                           *
*      Send bug reports and feedback to tom_at_work geocities.com           *
*                                                                           *
* Description:  Complete interface to all DPMI 0.9 calls. Some additional   *
*               (DOS specific) routines were added to provide some enhanced *
*               TP/BP compatibility.                                        *
*                                                                           *
*               Another feature of this unit is the possibility to enable   *
*               "near pointers" on your machine.                            *
*                                                                           *
*                                                                           *
* See the history at the end too                                            *
* See the docs (DPMI.HTM) too                                               *
*                                                                           *
*****************************************************************************}
unit DPMI;

interface

{$MODE OBJFPC}

type
     Descriptor = array[0..7] of Byte;

     Registers = record
               case integer of
                    0 : (edi, esi, ebp, res1, ebx, edx, ecx, eax : DWord;
                         Flags, es, ds, fs, gs, ip, cs, sp, ss : Word);
                    1 : (di, di2, si, si2, bp, bp2, res2, res3,
                         bx, bx2, dx, dx2, cx, cx2, ax, ax2 : Word);
                    2 : (res4 : array[1..4] of DWord;
                         bl, bh, bl2, bh2, dl, dh, dl2, dh2,
                         cl, ch, cl2, ch2, al, ah, al2, ah2 : Byte);
     end;

const
  { Flags constants }
      fCarry     = $0001;    { carry flag }
      fParity    = $0004;    { parity flag }
      fAuxiliary = $0010;    { auxiliary flag }
      fZero      = $0040;    { zero flag }
      fSign      = $0080;    { sign flag }
      fTrap      = $0100;    { trap flag }
      fInterrupt = $0200;    { interrupt flag }
      fDirection = $0400;    { direction flag }
      fOverflow  = $0800;    { overflow flag }

type
     pm_Addr = record
             offset : DWord;
             selector : Word;
     end;

     rm_Addr = record
             offset : Word;
             segment : Word;
     end;

     MemInfoBuf = record
                largest_available_free_block : Longint;
                max_unlocked_pages : Longint;
                max_lockable_pages : Longint;
                linear_address_space_size : Longint;
                number_of_unlocked_pages : Longint;
                number_of_free_pages : Longint;
                number_of_physical_pages : Longint;
                free_linear_address_space : Longint;
                size_of_paging_file : Longint;
                reserved : array[0..2] of Longint;
     end;

var
	dpmi_error : DWord;
	dpmi_halt_on_error : Boolean;

{ Initialization services }

function dpmi_get_cpu_mode(var in_pm : Boolean) : Boolean;

{ LDT management services }

function dpmi_allocate_ldt_descriptors( number : Word; var basesel : Word) : Boolean;
function dpmi_free_ldt_descriptor( sel : Word) : Boolean;
function dpmi_segment_to_descriptor( seg : Word; var sel : Word) : Boolean;
function dpmi_get_next_selector_increment_value(var incval : Word) : Boolean;

function dpmi_get_segment_base_address( sel : Word; var baseaddr : DWord) : Boolean;
function dpmi_set_segment_base_address( sel : Word;  baseaddr : DWord) : Boolean;
function dpmi_get_segment_limit( sel : Word; var limit : DWord) : Boolean;
function dpmi_set_segment_limit( sel : Word;  limit : DWord) : Boolean;

function dpmi_get_descriptor_access_rights( sel : Word; var rights : Word) : Boolean; { lar }
function dpmi_set_descriptor_access_rights( sel : Word;  rights : Word) : Boolean;
function dpmi_create_code_segment_alias_descriptor( codesel : Word; var sel : Word) : Boolean;

function dpmi_get_descriptor( sel : Word; var descr : Descriptor) : Boolean;
function dpmi_set_descriptor( sel : Word; descr : Descriptor) : Boolean;

function dpmi_allocate_specific_descriptor( sel : Word) : Boolean;

{ Memory management services }

function dpmi_get_free_memory_information(var mem : meminfobuf) : Boolean;

function dpmi_allocate_memory_block( size : DWord; var linearaddr : DWord; var blockhandle : DWord) : Boolean;
function dpmi_free_memory_block( blockhandle : DWord) : Boolean;
function dpmi_resize_memory_block( newsize : DWord; var blockhandle : DWord; var linearaddr : DWord) : Boolean;

{ Physical address mapping }

function dpmi_physical_address_mapping( physaddr : DWord;  size : DWord; var linearaddr : DWord) : Boolean;

{ DOS memory management }

function dpmi_allocate_DOS_memory_block( size : DWord; var realaddr : rm_Addr; var sel : Word) : Boolean;
function dpmi_free_DOS_memory_block( sel : Word) : Boolean;
function dpmi_resize_DOS_memory_block( newsize : DWord;  sel : Word) : Boolean;

function dpmi_allocate_DOS_memory_block( size : DWord; var realseg : Word; var sel : Word) : Boolean;

{ Interrupt services }

function dpmi_get_rm_interrupt( number : Byte; var addr : rm_Addr) : Boolean;
function dpmi_set_rm_interrupt( number : Byte;  addr : rm_Addr) : Boolean;
function dpmi_get_exception_handler( number : Byte; var addr : pm_Addr) : Boolean;
function dpmi_set_exception_handler( number : Byte;  addr : pm_Addr) : Boolean;
function dpmi_get_pm_interrupt( number : Byte; var addr : pm_Addr) : Boolean;
function dpmi_set_pm_interrupt( number : Byte;  addr : pm_Addr) : Boolean;

{ Virtual interrupt state functions }

function dpmi_get_and_disable_virtual_interrupts(var prevstate : Boolean) : Boolean;
function dpmi_get_and_enable_virtual_interrupts(var prevstate : Boolean) : Boolean;
function dpmi_get_virtual_interrupt_state(var state : Boolean) : Boolean;

{ Translation services }

function dpmi_simulate_rm_interrupt( number : Byte; var regs : Registers) : Boolean;
function dpmi_call_rm_procedure_with_retf_frame(var regs : Registers) : Boolean;
function dpmi_call_rm_procedure_with_iret_frame(var regs : Registers) : Boolean;

function dpmi_allocate_rm_callback( procaddr : pm_Addr;  regs : pm_Addr; var realaddr : rm_Addr) : Boolean;
function dpmi_free_rm_callback( procaddr : rm_Addr) : Boolean;

{ Page locking services }

function dpmi_lock_linear_region( linaddr : DWord;  size : DWord) : Boolean;
function dpmi_unlock_linear_region( linaddr : DWord;  size : DWord) : Boolean;

function dpmi_mark_rm_region_as_pageable( startaddr : DWord;  size : DWord) : Boolean;
function dpmi_relock_rm_region( startaddr : DWord;  size : DWord) : Boolean;

function dpmi_get_page_size(var pagesize : DWord) : Boolean;

function dpmi_mark_page_as_demand_paging_candidate( startaddr : DWord;  size : DWord) : Boolean;
function dpmi_discard_page_contents( startaddr : DWord;  size : DWord) : Boolean;

{ Miscellaneous services }

function dpmi_get_version(var ver : Word; var flags : Word; var processor : Byte; var picbase : Word) : Boolean;

{ Commonly used combinations of the above }

function create_selector(var sel : Word;  baseaddr, size : DWord) : Boolean;
function change_selector( sel : Word;  new_baseaddr, new_size : DWord) : Boolean;
function free_selector( sel : Word) : Boolean;

function get_linear_address( sel : Word;  offset : DWord) : DWord;
function get_linear_address( addr : rm_Addr) : DWord;

function map_physical_memory( physaddr : DWord;  size : DWord; var adr : pm_Addr) : Boolean;
function map_physical_memory( physaddr : DWord;  size : DWord; var sel : Word) : Boolean;

function intr( num : Byte; var r : registers) : Boolean;
function realintr( num : Byte; var r : registers) : Boolean;

function lock_data(var data;  size : DWord) : Boolean;
function lock_code( faddr : pointer;  size : DWord) : Boolean;

function unlock_data(var data;  size : DWord) : Boolean;
function unlock_code( faddr : Pointer;  size : DWord) : Boolean;

{ Return the current contents of the segment registers }

function CSeg : Word;   { returns %cs segment }
function DSeg : Word;   { returns %ds segment }
function DSegAlias : Word; { returns ___djgpp_ds_alias }
function ESeg : Word;   { returns %es segment }
function FSeg : Word;   { returns %fs segment }
function GSeg : Word;   { returns %gs segment }
function SSeg : Word;   { returns %ss segment }

{ Port access }

procedure outportb( port : Word;  data : Byte);
procedure outportw( port : Word;  data : Word);
procedure outportl( port : Word;  data : Longint);

function inportb( port : Word) : Byte;
function inportw( port : Word) : Word;
function inportl( port : Word) : Longint;

{ Enable / disable hardware interrupts }

procedure Enable;
procedure Disable;

{ Transfer buffer access }

function tb_size : DWord;
function tb_address : DWord;

{ "near pointer" handling }
{ ! use carefully ! this may destroy everything on your computer ;) }

function dpmi_enable_nearptr : Boolean;
function dpmi_disable_nearptr : Boolean;

function dpmi_nearptr_enabled(var enabled : Boolean) : Boolean;

function dpmi_nearptr_address_mapping( physaddr, size : DWord; var nearptr : Pointer) : Boolean;

implementation

{$ASMMODE ATT}

const
      err_allocate_ldt_descriptors     = $00010000;
      err_free_ldt_descriptor          = $00010001;
      err_segment_to_descriptor        = $00010002;
      err_get_next_selector_increment  = $00010003;
      err_get_segment_base_address     = $00010004;
      err_set_segment_base_address     = $00010005;
      err_get_segment_limit            = $00010006;
      err_set_segment_limit            = $00010007;
      err_get_descriptor_access_rights = $00010008;
      err_set_descriptor_access_rights = $00010009;
      err_create_code_segment_alias    = $0001000A;
      err_get_descriptor               = $0001000B;
      err_set_descriptor               = $0001000C;
      err_allocate_specific_descriptor = $0001000D;

      err_get_free_memory_information  = $00020000;
      err_allocate_memory_block        = $00020001;
      err_free_memory_block            = $00020002;
      err_resize_memory_block          = $00020003;

      err_physical_address_mapping     = $00030000;

      err_allocate_DOS_memory_block    = $00040000;
      err_free_DOS_memory_block        = $00040001;
      err_resize_DOS_memory_block      = $00040002;

      err_get_rm_interrupt             = $00050000;
      err_set_rm_interrupt             = $00050001;
      err_get_exception_handler        = $00050002;
      err_set_exception_handler        = $00050003;
      err_get_pm_interrupt             = $00050004;
      err_set_pm_interrupt             = $00050005;

      err_get_and_enable_virtual_ir    = $00060000;
      err_get_and_disable_virtual_ir   = $00060001;
      err_get_virtual_ir_state         = $00060002;

      err_simulate_rm_interrupt        = $00070000;
      err_call_rm_proc_with_retf_frame = $00070001;
      err_call_rm_proc_with_iret_frame = $00070002;
      err_allocate_rm_callback         = $00070003;
      err_free_rm_callback             = $00070004;

      err_lock_linear_region           = $00080000;
      err_unlock_linear_region         = $00080001;
      err_mark_rm_region_as_pageable   = $00080002;
      err_relock_rm_region             = $00080003;
      err_get_page_size                = $00080004;
      err_mark_page_as_demand_candidate= $00080005;
      err_discard_page_contents        = $00080006;

      err_get_version                  = $00090000;

      err_enable_nearptr               = $000A0000;
      err_disable_nearptr              = $000A0001;
      err_nearptr_address_mapping      = $000A0002;

type
	error_number = DWord;

	error_msg = record
		number      : error_number; { the error number to be written }
		description : ansistring;   { description of the error occured }
	end;

const
	err_msglist : array[0..46] of error_msg = (
		(number : err_allocate_ldt_descriptors;     description : 'Failed to allocate LDT descriptor'),
		(number : err_free_ldt_descriptor;          description : 'Couldn''t free LDT descriptor'),
		(number : err_segment_to_descriptor;        description : 'Unable to map segment to descriptor'),
		(number : err_get_next_selector_increment;  description : '(This error should not occur)'),
		(number : err_get_segment_base_address;     description : 'Unable to get base address, perhaps invalid selector'),
		(number : err_set_segment_base_address;     description : 'Unable to set base address, maybe invalid selector'),
		(number : err_get_segment_limit;            description : 'Unable to get segment limit, perhaps invalid selector specified'),
		(number : err_set_segment_limit;            description : 'Unable to set segment limit, maybe invalid selector'),
		(number : err_get_descriptor_access_rights; description : 'Unable to get descriptor access rights'),
		(number : err_set_descriptor_access_rights; description : 'Unable to get descriptor access rights'),
		(number : err_create_code_segment_alias;    description : 'Couldn''t create code segment alias'),
		(number : err_get_descriptor;               description : 'Failed to get descriptor'),
		(number : err_set_descriptor;               description : 'Failed to set descriptor'),
		(number : err_allocate_specific_descriptor; description : 'Couldn''t allocate specific descriptor'),
                  
		(number : err_get_free_memory_information;  description : '(This error should not occur)'),
		(number : err_allocate_memory_block;        description : 'Failed to allocate memory block'),
		(number : err_free_memory_block;            description : 'Failed to free memory block'),
		(number : err_resize_memory_block;          description : 'Couldn''t resize memory block'),
                  
		(number : err_physical_address_mapping;     description : 'Unable to map physical address to linear address'),
                  
		(number : err_allocate_DOS_memory_block;    description : 'Couldn''t allocate DOS memory'),
		(number : err_free_DOS_memory_block;        description : 'Couldn''t free DOS memory'),
		(number : err_resize_DOS_memory_block;      description : 'Failed to resize DOS memory block'),
                  
		(number : err_get_rm_interrupt;             description : '(This error should not occur)'),
		(number : err_set_rm_interrupt;             description : '(This error should not occur)'),
		(number : err_get_exception_handler;        description : 'Failed to get exception handler'),
		(number : err_set_exception_handler;        description : 'Failed to set exception handler'),
		(number : err_get_pm_interrupt;             description : '(This error should not occur)'),
		(number : err_set_pm_interrupt;             description : 'Failed to set protected mode interrupt'),

		(number : err_get_and_enable_virtual_ir;    description : '(This error should not occur)'),
		(number : err_get_and_disable_virtual_ir;   description : '(This error should not occur)'),
		(number : err_get_virtual_ir_state;         description : '(This error should not occur)'),

		(number : err_simulate_rm_interrupt;        description : 'Failed to simulate real mode interrupt'),
		(number : err_call_rm_proc_with_retf_frame; description : 'Failed to call real mode procedure with retf frame'),
		(number : err_call_rm_proc_with_iret_frame; description : 'Failed to call real mode procedure with iret frame'),
		(number : err_allocate_rm_callback;         description : 'Failed to allocate real mode callback'),
		(number : err_free_rm_callback;             description : 'Failed to free real mode callback'),

		(number : err_lock_linear_region;           description : 'Couldn''t lock linear region'),
		(number : err_unlock_linear_region;         description : 'Couldn''t unlock linear region'),
		(number : err_mark_rm_region_as_pageable;   description : 'Failed to mark real mode region as pageable'),
		(number : err_relock_rm_region;             description : 'Couldn''t relock real mod region'),
		(number : err_get_page_size;                description : 'Failed to obtain page size'),
		(number : err_mark_page_as_demand_candidate;description : 'Couldn''t mark page as demand paging candidate'),
		(number : err_discard_page_contents;        description : 'Failed at discarding page contents'),
                  
		(number : err_get_version;                  description : '(This error should not occur)'),

		(number : err_enable_nearptr;               description : 'Failed to enable near pointers'),
		(number : err_disable_nearptr;              description : 'Failed to disable near pointers'),
		(number : err_nearptr_address_mapping;      description : 'Failed to obtain pointer to physical memory area'));

procedure dpmi_handler(n : error_number);
var
	i : DWord;
begin
	dpmi_error := n;
	i := 0;
	while ((err_msglist[i].number <> n) and (i <= high(err_msglist))) do i := i + 1;

	if (i = (high(err_msglist)+1)) then begin
		Writeln(stderr, 'Unit DPMI : Unknown error ', hexstr(n, 8));
	end else begin
		Writeln(stderr, 'Unit DPMI : Error ', hexstr(n, 8));
		Writeln(stderr, err_msglist[i].description);
	end;
	if (dpmi_halt_on_error) then
		runerror(216);
end;

var
    save_ds,
    new_ds : Word;
    nearptr_enabled : Boolean;


{ Initialization services }

function dpmi_get_cpu_mode(var in_pm : Boolean) : Boolean; assembler;
asm
    movw $0x1686, %ax
    int $0x2f
    movl in_pm, %edi
    notl %eax
    movb %al, (%edi)
    movl $-1, %eax        // note : This function always succeeds
end;

{ LDT management services }

function dpmi_allocate_ldt_descriptors( number : Word; var basesel : Word) : Boolean; assembler;
asm
   movw $0x0000, %ax
   movw number, %cx
   int $0x31
   jc .LError
   movl basesel, %edi
   movw %ax, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_allocate_ldt_descriptors
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_free_ldt_descriptor( sel : Word) : Boolean; assembler;
asm
   movw $0x0001, %ax
   movw sel, %bx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_free_ldt_descriptor
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_segment_to_descriptor( seg : Word; var sel : Word) : Boolean; assembler;
asm
   movw $0x0002, %ax
   movw seg, %bx
   int $0x31
   jc .LError
   movl sel, %edi
   movw %ax, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_segment_to_descriptor
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_next_selector_increment_value(var incval : Word) : Boolean; assembler;
asm
   movw $0x0003, %ax
   int $0x31
   movl INCVAL, %edi
   movw %ax, (%edi)
   movl $-1, %eax   { this function always succeeds }
end;

function dpmi_get_segment_base_address( sel : Word; var baseaddr : DWord) : Boolean; assembler;
asm
   movw $0x0006, %ax
   movw sel, %bx
   int $0x31
   jc .LError
   shll $16, %ecx
   movw %dx, %cx
   movl baseaddr, %edi
   movl %ecx, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_get_segment_base_address
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_set_segment_base_address( sel : Word;  baseaddr : DWord) : Boolean; assembler;
asm
   movw $0x0007, %ax
   movw sel, %bx
   xorl %edx, %edx
   movl baseaddr, %ecx
   movw %cx, %dx
   shrl $16, %ecx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_set_segment_base_address
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_segment_limit( sel : Word; var limit : DWord) : Boolean; { lsl }assembler;
asm
   movzwl sel, %eax
   lsll %eax, %eax
   jnz .LError
   movl limit, %edi
   movl %eax, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_get_segment_limit
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_set_segment_limit( sel : Word;  limit : DWord) : Boolean; assembler;
asm
   movw $0x0008, %ax
   movw sel, %bx
   movl limit, %ecx
   movw %cx, %dx
   shrl $16, %ecx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_set_segment_limit
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_descriptor_access_rights( sel : Word; var rights : Word) : Boolean; { lar }assembler;
asm
   movzwl sel, %ebx
   larl %ebx, %ebx
   jnz .LError
   movl rights, %edi
   movw %bx, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_get_descriptor_access_rights
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_set_descriptor_access_rights( sel : Word;  rights : Word) : Boolean; assembler;
asm
   movw $0x0009, %ax
   movw sel, %bx
   movw rights, %cx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_set_descriptor_access_rights
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_create_code_segment_alias_descriptor( codesel : Word; var sel : Word) : Boolean; assembler;
asm
   movw $0x000a, %ax
   movw codesel, %bx
   int $0x31
   jc .LError
   movl sel, %edi
   movw %ax, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_create_code_segment_alias
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_descriptor( sel : Word; var descr : descriptor) : Boolean; assembler;
asm
   movw $0x000b, %ax
   movw sel, %bx
   movl descr, %edi { %es = %ds ! }
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_get_descriptor
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_set_descriptor( sel : Word;  descr : Descriptor) : Boolean; assembler;
asm
   movw $0x000c, %ax
   movw sel, %bx
   leal descr, %edi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_set_descriptor
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_allocate_specific_descriptor( sel : Word) : Boolean; assembler;
asm
   movw $0x000d, %ax
   xorl %edi, %edi
   movw sel, %di
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_allocate_specific_descriptor
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_free_memory_information(var mem : meminfobuf) : Boolean; assembler;
asm
    movw $0x500, %ax
    movl mem, %edi
    int $0x31
    movl $-1, %eax      // note : This function always succeeds
end;

function dpmi_allocate_memory_block( size : DWord; var linearaddr : DWord; var blockhandle : DWord) : Boolean; assembler;
asm
   movw $0x501, %ax
   movl size, %ecx
   movw %cx, %bx
   shrl $16, %ebx
   int $0x31
   jc .LError
   movl blockhandle, %edx
   movw %si, 2(%edx)
   movw %di, 0(%edx)
   movl linearaddr, %edx
   movw %bx, 2(%edx)
   movw %cx, 0(%edx)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_allocate_memory_block
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_free_memory_block( blockhandle : DWord) : Boolean; assembler;
asm
   movw $0x502, %ax
   movl blockhandle, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_free_memory_block
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_resize_memory_block( newsize : DWord; var blockhandle : DWord; var linearaddr : DWord) : Boolean; assembler;
asm
   movw $0x503, %ax
   movl newsize, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl blockhandle, %esi
   movl (%esi), %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl linearaddr, %edx
   movw %bx, 2(%edx)
   movw %cx, 0(%edx)
   movl blockhandle, %edx
   movw %si, 2(%edx)
   movw %di, 0(%edx)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_resize_memory_block
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

{ Physical address mapping }

function dpmi_physical_address_mapping( physaddr : DWord;  size : DWord; var linearaddr : DWord) : Boolean; assembler;
asm
   movw $0x800, %ax
   movl physaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx

   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl linearaddr, %edi
   movw %bx, 2(%edi)
   movw %cx, 0(%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_physical_address_mapping
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

{ DOS memory management services }

function dpmi_allocate_DOS_memory_block( size : DWord; var realaddr : rm_Addr; var sel : Word) : Boolean; assembler;
asm
   movw $0x100, %ax
   movl size, %ebx
   addl $15, %ebx
   shrl $4, %ebx
   int $0x31
   jc .LError
   movl realaddr, %edi
   movw %ax, rm_addr.segment(%edi)
   movw $00, rm_addr.offset(%edi)
   movl SEL, %edi
   movw %dx, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_allocate_DOS_memory_block
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_allocate_DOS_memory_block( size : DWord; var realseg : Word; var sel : Word) : Boolean;
var realaddr : rm_Addr;
begin
     result := dpmi_allocate_DOS_memory_block(size, realaddr, sel);
     realseg := realaddr.segment;
end;

function dpmi_free_DOS_memory_block( sel : Word) : Boolean; assembler;
asm
   movw $0x101, %ax
   movw sel, %dx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_free_DOS_memory_block
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_resize_DOS_memory_block( newsize : DWord;  sel : Word) : Boolean; assembler;
asm
   movw $0x102, %ax
   movl newsize, %ebx
   addl $15, %ebx
   shrl $4, %ebx
   movw sel, %dx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_resize_DOS_memory_block
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

{ Interrupt services }

function dpmi_get_rm_interrupt( number : Byte; var addr : rm_Addr) : Boolean; assembler;
asm
   movw $0x200, %ax
   movb number, %bl
   int $0x31
   movl addr, %edi
   movw %cx, rm_addr.segment(%edi)     // cx:dx = segment:offset address
   movw %dx, rm_addr.offset(%edi)
   movl $-1, %eax        // note : This function always succeeds
end;

function dpmi_set_rm_interrupt( number : Byte;  addr : rm_Addr) : Boolean; {assembler;}
begin
asm
   movw $0x201, %ax
   movb number, %bl
   movw addr.segment, %cx
   movw addr.offset, %dx
   int $0x31
   movl $-1, %eax          // note : This function always succeeds
end;
end;

function dpmi_get_exception_handler( number : Byte; var addr : pm_Addr) : Boolean; assembler;
asm
   movw $0x202, %ax
   movb number, %bl
   int $0x31
   jc .LError
   movl addr, %edi
   movw %cx, pm_addr.selector(%edi)
   movl %edx, pm_addr.offset(%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_get_exception_handler
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_set_exception_handler( number : Byte;  addr : pm_Addr) : Boolean; {assembler;}
begin
asm
   movw $0x203, %ax
   movb number, %bl
   movw addr.selector, %cx
   movl addr.offset, %edx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_set_exception_handler
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;
end;

function dpmi_get_pm_interrupt( number : Byte; var addr : pm_Addr) : Boolean; assembler;
asm
   movw $0x204, %ax
   movb number, %bl
   int $0x31
   movl addr, %edi
   movw %cx, pm_addr.selector(%edi)
   movl %edx, pm_addr.offset(%edi)
   movw %cx, 4(%edi)
   movl %edx, 0(%edi)
   movl $-1, %eax       // note : This function always succeeds
end;

function dpmi_set_pm_interrupt( number : Byte;  addr : pm_Addr) : Boolean; {assembler;}
begin
asm
   movw $0x205, %ax
   movb number, %bl
   movw addr.selector, %cx
   movl addr.offset, %edx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_set_pm_interrupt
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;
end;

{ Virtual interrupt state functions }

function dpmi_get_and_disable_virtual_interrupts(var prevstate : Boolean) : Boolean; assembler;
asm
   movw $0x900, %ax
   int $0x31
   movl prevstate, %edi
   movb %al, (%edi)
   movl $-1, %eax      // note : This function always succeeds
end;

function dpmi_get_and_enable_virtual_interrupts(var prevstate : Boolean) : Boolean; assembler;
asm
   movw $0x901, %ax
   int $0x31
   movl prevstate, %edi
   movb %al, (%edi)
   movl $-1, %eax       // note : This function always succeeds
end;

function dpmi_get_virtual_interrupt_state(var state : Boolean) : Boolean; assembler;
asm
   movw $0x902, %ax
   int $0x31
   movl state, %edi
   movb %al, (%edi)
   movl $-1, %eax        // note : This function always succeeds
end;


{ Translation services }

function dpmi_simulate_rm_interrupt( number : Byte; var regs : Registers) : Boolean; assembler;
asm
   movw $0x300, %ax
   movb number, %bl
   xorb %bh, %bh
   xorl %ecx, %ecx
   movl regs, %edi
   movw $0, registers.sp(%edi)    // regs.sp := 0
   movw $0, registers.ss(%edi)    // regs.ss := 0
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_simulate_rm_interrupt
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_call_rm_procedure_with_retf_frame(var regs : Registers) : Boolean; assembler;
asm
   movw $0x301, %ax
   xorl %ebx, %ebx
   xorl %ecx, %ecx
   movl regs, %edi
   movw $0, registers.sp(%edi)
   movw $0, registers.ss(%edi)
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_call_rm_proc_with_retf_frame
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_call_rm_procedure_with_iret_frame(var regs : Registers) : Boolean; assembler;
asm
   movw $0x302, %ax
   xorl %ebx, %ebx
   xorl %ecx, %ecx
   movl regs, %edi
   movw $0, registers.sp(%edi)
   movw $0, registers.ss(%edi)
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_call_rm_proc_with_iret_frame
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_allocate_rm_callback( procaddr : pm_Addr;  regs : pm_Addr; var realaddr : rm_Addr) : Boolean; {assembler;}
begin
asm
   movw $0x303, %ax
   movl procaddr.offset, %esi
   movl regs.offset, %edi
   pushw %ds
   pushw %es

   movw regs.selector, %es
   movw procaddr.selector, %ds
   int $0x31
   
   popw %es
   popw %ds
   jc .LError_0x303
   movl realaddr, %edi
   movw %cx, rm_Addr.segment(%edi)
   movw %dx, rm_Addr.offset(%edi)
   movl $-1, %eax
   jmp .LNoError_0x303
.LError_0x303:
   pushl err_allocate_rm_callback
   call dpmi_handler
.LNoErrorHandler_0x303:
   xorl %eax, %eax
.LNoError_0x303:
end;
end;

function dpmi_free_rm_callback( procaddr : rm_Addr) : Boolean; assembler;
asm
   movw $0x304, %ax
   movl procaddr, %ecx
   movw %cx, %dx
   shrl $16, %ecx
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_free_rm_callback
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_version(var ver : Word; var flags : Word; var processor : Byte; var picbase : Word) : Boolean; assembler;
asm
   movw $0x400, %ax
   int $0x31
   movl ver, %edi
   movw %ax, (%edi)
   movl flags, %edi
   movw %bx, (%edi)
   movl processor, %edi
   movb %cl, (%edi)
   movl picbase, %edi
   movw %dx, (%edi)
   movl $-1, %eax               // note : This function always succeeds
end;

{ Page locking services }

function dpmi_lock_linear_region( linaddr : DWord;  size : DWord) : Boolean; assembler;
asm
   movw $0x600, %ax
   movl linaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_lock_linear_region
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_unlock_linear_region( linaddr : DWord;  size : DWord) : Boolean; assembler;
asm
   movw $0x601, %ax
   movl linaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_unlock_linear_region
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_mark_rm_region_as_pageable( startaddr : DWord;  size : DWord) : Boolean; assembler;
asm
   movw $0x602, %ax
   movl startaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_mark_rm_region_as_pageable
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_relock_rm_region( startaddr : DWord;  size : DWord) : Boolean; assembler;
asm
   movw $0x603, %ax
   movl startaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_relock_rm_region
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_get_page_size(var pagesize : DWord) : Boolean; assembler;
asm
   movw $0x604, %ax
   int $0x31
   jc .LError
   movl pagesize, %edi
   shll $16, %ebx
   movw %cx, %bx
   movl %ebx, (%edi)
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_get_page_size
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_mark_page_as_demand_paging_candidate( startaddr : DWord;  size : DWord) : Boolean; assembler;
asm
   movw $0x702, %ax
   movl startaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_mark_page_as_demand_candidate
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function dpmi_discard_page_contents( startaddr : DWord;  size : DWord) : Boolean; assembler;
asm
   movw $0x703, %ax
   movl startaddr, %ebx
   movw %bx, %cx
   shrl $16, %ebx
   movl size, %esi
   movw %si, %di
   shrl $16, %esi
   int $0x31
   jc .LError
   movl $-1, %eax
   jmp .LNoError
.LError:
   pushl err_discard_page_contents
   call dpmi_handler
.LNoErrorHandler:
   xorl %eax, %eax
.LNoError:
end;

function create_selector(var sel : Word;  baseaddr, size : DWord) : Boolean;
begin
     result := dpmi_allocate_ldt_descriptors(1, sel) and
               dpmi_set_segment_base_address(sel, baseaddr) and
               dpmi_set_segment_limit(sel, size);
end;

function change_selector( sel : Word;  new_baseaddr, new_size : DWord) : Boolean;
begin
     result := dpmi_set_segment_base_address(sel, new_baseaddr) and
               dpmi_set_segment_limit(sel, new_size);
end;

function free_selector( sel : Word) : Boolean;
begin
     result := dpmi_free_ldt_descriptor(sel);
end;

function get_linear_address( sel : Word;  offset : DWord) : DWord;
var addr : DWord;
begin
     if (dpmi_get_segment_base_address(sel, addr)) then
        result := addr + offset
     else
         result := 0;
end;

function get_linear_address( addr : rm_Addr) : DWord;
begin
     result := addr.segment shl 4 + addr.offset;
end;

function map_physical_memory( physaddr : DWord;  size : DWord; var sel : Word) : Boolean;
var linearaddr : DWord;
begin
     result := dpmi_allocate_ldt_descriptors(1, sel) and
               dpmi_physical_address_mapping(physaddr, size, linearaddr) and
               dpmi_set_segment_base_address(sel, linearaddr) and
               dpmi_set_segment_limit(sel, size);
end;

function map_physical_memory( physaddr : DWord;  size : DWord; var adr : pm_Addr) : Boolean;
var linearaddr : DWord;
begin
     adr.offset := 0;
     result := dpmi_allocate_ldt_descriptors(1, adr.selector) and
               dpmi_physical_address_mapping(physaddr, size, linearaddr) and
               dpmi_set_segment_base_address(adr.selector, linearaddr) and
               dpmi_set_segment_limit(adr.selector, size);
end;


function intr( num : Byte; var r : registers) : Boolean;
begin
     result := dpmi_simulate_rm_interrupt(num, r);
end;

function realintr( num : Byte; var r : registers) : Boolean;
begin
     result := dpmi_simulate_rm_interrupt(num, r);
end;

function lock_data(var data;  size : DWord) : Boolean;
begin
     result := dpmi_lock_linear_region(get_linear_address(dseg, DWord(@data)), size);
end;

function lock_code( faddr : pointer;  size : DWord) : Boolean;
begin
     result := dpmi_lock_linear_region(get_linear_address(cseg, DWord(faddr)), size);
end;

function unlock_data(var data;  size : DWord) : Boolean;
begin
     result := dpmi_unlock_linear_region(get_linear_address(dseg, DWord(@data)), size);
end;

function unlock_code( faddr : Pointer;  size : DWord) : Boolean;
begin
     result := dpmi_unlock_linear_region(get_linear_address(cseg, DWord(faddr)), size);
end;

function CSeg : Word; assembler;
asm
   movw %cs, %ax
end;

function DSeg : Word; assembler;
asm
   movw %ds, %ax
end;

function DSegAlias : Word;
var
	ds_alias : word; external name '___v2prt0_ds_alias';
begin
	result := ds_alias;
end;

function ESeg : Word; assembler;
asm
   movw %es, %ax
end;

function FSeg : Word; assembler;
asm
   movw %fs, %ax
end;

function GSeg : Word; assembler;
asm
   movw %gs, %ax
end;

function SSeg : Word; assembler;
asm
   movw %ss, %ax
end;

procedure outportb( port : Word;  data : Byte); assembler;
asm
   movw PORT, %dx
   movb DATA, %al
   outb %al, %dx
end ['EAX', 'EDX'];

procedure outportw( port : Word;  data : Word); assembler;
asm
   movw PORT, %dx
   movw DATA, %ax
   outw %ax, %dx
end ['EAX', 'EDX'];

procedure outportl( port : Word;  data : Longint); assembler;
asm
   movw PORT, %dx
   movl DATA, %eax
   outl %eax, %dx
end ['EAX', 'EDX'];

function inportb( port : Word) : Byte; assembler;
asm
   movw PORT, %dx
   inb %dx, %al
end ['EAX', 'EDX'];

function inportw( port : Word) : Word; assembler;
asm
   movw PORT, %dx
   inw %dx, %ax
end ['EAX', 'EDX'];

function inportl( port : Word) : Longint; assembler;
asm
   movw PORT, %dx
   inl %dx, %eax
end ['EAX', 'EDX'];


procedure Enable; assembler;
asm
   sti
end;

procedure Disable; assembler;
asm
   cli
end;

function tb_size : DWord;
begin
     tb_size := go32_info_block.size_of_transfer_buffer;
end;

function tb_address : DWord;
begin
     tb_address := go32_info_block.linear_address_of_transfer_buffer;
end;

function dpmi_enable_nearptr : Boolean;
var base : DWord;
begin
     if not (nearptr_enabled) then begin
        save_ds := DSeg;
        if (dpmi_allocate_ldt_descriptors(1, new_ds) and
            dpmi_set_segment_limit(new_ds, $FFFFFFFF) and
            dpmi_get_segment_base_address(save_ds, base) and
            dpmi_set_segment_base_address(new_ds, base)) then begin
            asm
               movw NEW_DS, %ax
               movw %ax, %ds
               movw %ax, %es
            end;
            nearptr_enabled := true;
        end else begin
            dpmi_handler(err_enable_nearptr);
        end;
     end;
end;

function dpmi_disable_nearptr : Boolean;
begin
     if (nearptr_enabled) then begin
        asm
           movw SAVE_DS, %ax
           movw %ax, %ds
           movw %ax, %es
        end;
        dpmi_free_ldt_descriptor(new_ds);
        nearptr_enabled := false;
     end;
end;

function dpmi_nearptr_enabled(var enabled : Boolean) : Boolean;
begin
     enabled := nearptr_enabled;
     result := true;            { this function always succeeds }
end;

function dpmi_nearptr_address_mapping( physaddr, size : DWord; var nearptr : Pointer) : Boolean;
var base : DWord;
begin
     if (nearptr_enabled) then begin
        result := dpmi_get_segment_base_address(DSeg, base) and
                  dpmi_physical_address_mapping(physaddr, size, DWord(nearptr));
        dword(nearptr) := dword(nearptr) - base;
     end else begin
	 nearptr := nil;
         dpmi_handler(err_nearptr_address_mapping); { "near pointers" aren't enabled }
         result := false;
     end;
end;

begin
	nearptr_enabled := false;
	dpmi_halt_on_error := true;
end.
{ 31-12-1998 start of history
  - thanks to a deletion of my entire sources, I had to re-do all things :(

  01-01-1999 some "near pointer" bug fixing
  - removed bug in dpmi_nearptr_address_mapping()
  - added "near pointer" example (NEAR.PAS)
  - removed the automatic -1 segment limit adjustements (to keep it uniform in 
  all functions)

  02-01-1999 changed output of error handler
  - now writes errors to stderr
  - includes MEMORY and FARMEM automatically
  
  03-01-1999 some update
  - added change_selector()
  - updated documentation
 
  04-01-1999 changed examples to work correctly again
  - callback.pas : callback handler was buggy
  - ascii.pas, textmess.pas, sel_des.pas, vga_sel.pas, ... : changed them to fit
  the new syntax of MEMORY
  - removed the automatic use of MEMORY and FARMEM again (doesn't give the 
  effect wished)

  06-01-1999 new functions
  - get_linear_address now returns the linear address of a real mode address
  too

  12-01-1999 some bugfixes, new example
  - fixed keyclick.pas : interrupt handler was buggy (segment registers weren't
  saved)
  - added call16.pas : calls a 16 bit real mode procedure from protected mode
  (interrupt call)

  13-01-1999 updated history, new example
  - added dcall16.pas : calls 16 bit real mode code returning with a retf
  instruction
  - updated history, to reflect new year :) (all entries from january were
  dated in 1998.....)
  - moved all examples into \EXAMPLES directory
	
  17-01-1999 updated examples
  - added/changed some comments in the example files
  - start of makefile
  - new dpmi_allocate_DOS_memory_block() which allows a rm_Addr record as
  parameter
  - updated docs to reflect this

  18-01-1999 doc revision
  - fixed some broken links
  - added new syntax of dpmi_allocate_DOS_memory_block()

  03-02-1999 some cleanup
  - textmess.pas : automatically uses entire screen for demo, not only the 
  standard 80x25 text screen (suggestion from AlexS@freepage.de)

  04-02-1999 code cleanup, error handler
  - added LGPL license notice in header
  - error code constants
  - completely new error handling (using ERRHNDLE.PAS)
  - some code cleanup

  06-02-1999 finished error handling
  - ERRHNDLE.PAS now included within this package (still beta)
  - new example: HANDLER.PAS

  13-02-1999 new functions
  - added map_physical_memory() variant with a pm_Addr as parameter

  14-02-1999 bugfixes
  - fixed a bug which I thought was a FPC bug ;) (var parameters in assembler;
  blocks)

  15-02-1999 some fixes
  - the bug from 14-02-1999 is a _real_ bug, not my fault !
  - code cleanup (removed comments, adapted assembler code to 0.99.11)
  - removed obsolete(?) inline; statements (they were a bit annoying)

  21-03-1999 bug fixes
  - now finally works again with the latest snapshots. This was a bug related 
  to the one I thought I had fixed on 14/2/98.....

  30-03-1999 cleanup
  - some code cleanup
  - FPC 0.99.11 is mandatory

  09-05-1999 adaption to new assembler
  - finished adaption to new internal FPC assembler
  - removed errhndle.pas
  - introduced the dpmi_halt_on_error flag - if set any error causes a RTE216
  - error messages are always printed to stderr
}
