UNIT ECrc32;
{ Verified! Both Pascal and Assembler version generate same CRC as ARJ, see
     example in help!

  Compiles (Pascal code) under BP, but is untested. Easiest way to test is to
  calc a crc over some file with the calctest demo, and compare it with ARJ's
  or PKZIP's}

INTERFACE


{$DEFINE USEASM}                {Intel asssembler is probably a lot faster}
{$DEFINE TightLoop}             {When defined code, is smaller. Larger code
                                  copies the internal loop 8 times to allow
                                  better pipelining, tested, but not if it
                                  actually gains speed}

{$I syseq.inc}                  { Some conditionals only, no actual code}

{ Constant definitions - Standard CRC-32 polynominal }
{$IFdef FPC}
{$AsmMode ATT}
{$ENDIF}
{$IFDEF Turbo}
 type CARDINAL=LONGINT;
{$ENDIF}
CONST
   Poly32 : CARDINAL = $0EDB88320;  { Default CRC polynome. Suited for 99.9%
                                       of the CRCs, parameter for
                                       CreateCRC32Table, also used by
                                       all major archivers including GZIP,ZIP,
                                       ARJ,RAR,LZH etc}

{ Usage : This procedure _must_ be called prior to any use of the CRC-
           calculating routines. It will generate an internal 256 elements
           table of 32-bits CRC-codes on the Heap which is needed for fast
           byte/character-based CRC calculations.}

PROCEDURE CreateCRC32Table( Poly32 : CARDINAL);

{ Usage : Calculates the CRC-32 value of any block of date from BufAdr,
           including BufSize number of bytes. This procedure can be used
           for total CRC-32 calculation or optionally for partial CRC
           addition.
           Recommended use:

            crc := CalcCRC( 0FFFFFFFFH, Buffer , SIZEOF( buffer ) );

   Input : crc     = initial crc-value. To be initialized to FFFFFFFFH before
                     the first call of CalcCRC. If CalcCRC is used for
                     partial (step-by-step) CRC-calculation, this parameter
                     should hold the previous result value of CalcCRC.
           Buffer  = The buffer that contains the to be validated
                     data.
           BufSize = Actual size of the actual to be validated data.
   Output: (In returnvalue)
           CRC-32 value over the specified data. Be aware that this value is
           ONLY the calculated CRC value over the data, NOT including any
           trailing zeroes! (a lot of crc-programs add 4 zeros to the end of
           the data) }
FUNCTION CalcCRC (crc: CARDINAL;var Buffer;BufSize:CARDINAL) : CARDINAL;



{ Usage : This procedure used to be called from within CalcCRC itself, for speed
          it's integrated in CalcCRC, but it
          has been included here to provide you with the possibility to add
          any trailing zeroes whenever applicable. (See the notes in
          CRC32.DOC! (RA-SDK, or available on request)

  Input : crc = CRC-32 value over the already processed data.
          ch  = Character value which is to be appended after the already
                processed data for inclusion in the CRC-32 calculation.
  Output: crc = CRC-32 value over the already processed data, including
                the appended character ch!}
PROCEDURE AddCRC( VAR crc : CARDINAL;ch  : BYTE );

{ Usage : When finished calculating a CRC-32 code over a block of data, the
           crc-value will still be inverted, so you need to call this routine
           to convert the CRC into a valid CARDINAL CRC-32 value for represen-
           tation purposes. This is not necessary if you only use the CRC
           internally. Only if you want to display it (and your using the
           standard polynomal), or compare it with the CRC created by external
           programs like ARJ or ZIP.

   Input : crc = CRC-32 value over the already processed data.
   Output: crc = normalized CRC-32 value over the processed data. }
FUNCTION InvertCRC(crc : CARDINAL ):CARDINAL;

IMPLEMENTATION

{Internal CRC table}

TYPE CRCTabType= ARRAY [0..255] OF CARDINAL;
     CRCTabPtr = ^CRCTabType;

VAR CRCTab : CRCTabPtr;


{$IFNDEF USEASM}

PROCEDURE CreateCRC32Table( Poly32:CARDINAL);
VAR
   I,J    : CARDINAL;
   Temp,
   Temp2  : CARDINAL;
BEGIN
   IF CrcTab=NIL THEN
    NEW(CrcTab);
   FOR I := 0 TO 255 DO
    BEGIN
      Temp:=I;
      FOR J := 0 TO 7 DO
       BEGIN
        Temp2:=Temp;
        Temp:=Temp SHR 1;
        IF (Temp2 AND 1)>0 THEN
          Temp:=Temp XOR Poly32;
       END;
      CRCTab^[I]:=Temp;
     END;
END;

PROCEDURE AddCRC( VAR crc : CARDINAL; ch : BYTE );
VAR Temp : CARDINAL;

BEGIN
   Temp:=ch;                            {Byte to cardinal conversion}
   Temp:=Temp xor (crc AND 255);
   crc:=crc shr 8;
   crc:=crc xor CRCTab^[Temp];
END;

FUNCTION CalcCRC (crc: CARDINAL;VAR Buffer;BufSize:CARDINAL) : CARDINAL;

{$IFDEF TURBO}
TYPE ArrType= ARRAY[0..65000] OF BYTE; {64kb array, only used as type}
{$ELSE}
TYPE ArrType= ARRAY[0..2095151] OF BYTE; {2MB array, only used as type}
{$ENDIF}

VAR I,Temp : CARDINAL;
    J : CARDINAL;
    HlpPtr : ^ArrType;

BEGIN
  J:=crc;
  HlpPtr:=@Buffer;
  IF BufSize>0 THEN
   FOR I:=0 TO BufSize-1 DO
    BEGIN
     Temp:=ORD(HlpPtr^[I]);           {Byte to cardinal conversion}
     Temp:=Temp xor (J AND 255);
     J:=J shr 8;
     J:=J xor CRCTab^[Temp];
    END;
  CalcCrc:=J;
END;

FUNCTION InvertCRC(crc : CARDINAL ):CARDINAL;

BEGIN
 InvertCrc:=crc XOR $FFFFFFFF;
END;

{$ELSE}

PROCEDURE CreateCRC32Table( Poly32 : CARDINAL); ASSEMBLER;

ASM
{$IFDEF FrameStack}
                     pushl %ebp
                     mov   %esp,%ebp
{$ENDIF}
                     movl    CRCTAB,%eax
                     testl  %eax,%eax
                     jne    .LTableallreadyallocated
                     pushl   $1024
                     pushl   crctab
                     call    GETMEM
.LTableallreadyallocated:
                     movl    CRCTAB,%edi
                     movl    8(%ebp),%ebx
                     xor     %esi,%esi
.LCreateLoop:        mov     %esi,%eax
                     mov     $8,%ecx
.LCreateInLoop:      mov     %eax,%edx
                     shrl    $1,%eax                // Check if C flag contains
                     test    $1,%edx                // killed bit, remove EDX
                     je      .LCreateNoXor
                     xorl     %ebx,%eax
 .LCreateNoXor:      loop    .LCreateInLoop
                     stosl
                     incl      %esi
                     cmpl      $256,%esi
                     jne .LCreateLoop
{$IFDEF FrameStack}
                     leave
{$ENDIF}
END['EAX','EBX','ECX','EDX','ESI','EDI'];



FUNCTION  CalcCRC( crc     : CARDINAL;var Buffer;BufSize:CARDINAL):CARDINAL;ASSEMBLER;
    ASM
    {$IFDEF FrameStack}
                             pushl %ebp
                             mov   %esp,%ebp
    {$ENDIF}
                             mov   8(%ebp),%ebx    // load previous CRC
                             mov   12(%ebp),%esi   // source address
                             movl    CRCTAB,%edi
                             mov   16(%ebp),%ecx   // Counter.
{$IFNDEF TightLoop}
.LTightTest:                 test $-8,%ecx  // Test if at least 8 bytes left in buffer
                             je   .LCalcEndbigloop

                             lodsb                  // Yes, loop optimalisation
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx

                             // the above base code is copied 7 times. So 7+1=8

                             lodsb                              // 2
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             lodsb                              // 3
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             lodsb                              // 4
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             lodsb                              // 5
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             lodsb                              // 6
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             lodsb                              // 7
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             lodsb                              // 8
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx

                             subl $8,%ecx               // 8 less to go
                             jmp  .LTightTest
.LCalcEndbigloop:            jecxz .LCalcAfterEndLoop
{$ENDIF}
  .LCalcLoop:                lodsb                      // Do (rest) buffer
                             movzbl %bl,%edx
                             shrl  $8,%ebx
                             xorb  %al,%dl
                             xorl  (%edi,%edx,4),%ebx
                             loop  .LCalcLoop

  .LCalcAfterEndLoop:        mov   %ebx,%eax
    {$IFDEF FrameStack}
                             leave
    {$ENDIF}
end['EAX','EBX','ECX','EDX','ESI','EDI'];


PROCEDURE AddCRC( VAR crc : CARDINAL; ch : BYTE );ASSEMBLER;



ASM
{$IFDEF FrameStack}
                         pushl %ebp
                         mov   %esp,%ebp
{$ENDIF}
                         mov   8(%ebp),%ebx
                         mov   (%ebx),%eax
                         movb  12(%ebp),%dl

                         movzbl %al,%ecx
                         shrl  $8,%eax
                         xorb  %dl,%cl
                         movl    CRCTAB,%edx
                         xorl  (%edx,%ecx,4),%eax
                         mov  %eax,(%ebx)
{$IFDEF FrameStack}
                         leave
{$ENDIF}
end['EAX','EBX','ECX','EDX'];


FUNCTION InvertCRC(crc : CARDINAL ):CARDINAL; ASSEMBLER;

ASM
{$IFDEF FrameStack}
                         pushl %ebp
                         mov   %esp,%ebp
{$ENDIF}
                         movl 8(%ebp),%eax
                         xorl $-1,%eax
{$IFDEF FrameStack}
                         leave
{$ENDIF}
end['EAX'];

{$ENDIF USEASM}

BEGIN
 CRCTAB:=NIL;
{CreateCRC32Table(Poly32);}  {Disabled for speed reasons,
                             init in your own program}
End.
