UNIT EDate;
{ Date unit in plain AT&T assembler, cool!, I hope it's fast.

   This is a unit of the XTDLIB/FPK package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)

   Thanks to Andreas Reinicke, who gave me an unit on which most of these
   procedures are based, though the implementations have changed through the
   years, the definitions stayed nearly the same. Also thanks to Holger Jacob,
   I compared my unit to his, and corrected some small things.

   All routines have expirency dates 29-2-2100 or higher.
   (LeapYr and DayNr are good until the year 4000)
   I don't know exactly how long unixdates based on 32-bits last, but
   it won't be a problem until after 2090.

  Most of the original (16-bit) routines were converted to assembler because
  one of my major application scanned log files with a date on every line.
  On that 386SX20 that WAS a major bottleneck.
  The others to get the unit complete, or as an exercise (in arithmetic on
  32-bits numbers in 16-bit assembler)

  The conversion from 16 to 32 bits assembler code was easier than the writing
  the pascal equivalents :-)

  This module is entirely done in (long)integers, since that SX20 I mention
  above didn't have a coprocessor. It's possible since now almost all computers
  have a coprocessor, a floating point approach is both more fast and more
  elegant in some cases. Let me know, if you know more about this subject.

  I'm no historian, so I didn't implement a full date range procedures, only
  the Pope Gregorius calendar, if you go before the year 1582/83, the procedures
  don't start using the Julian calendar.  }

INTERFACE

{$define useasm}                {Default : Use assembler, comment this off to
                                         compile pascal code
                                Right now assembler is turned off, cause some
                                asm-routines are instable}

{$Ifdef VER0_99_8}
 {$UNDEF Useasm}
{$ENDIF}
{$I syseq.inc}                  { some defines used in XTDLIB}



{$IfDEF FPC}
{$ifdef Useasm}
{$Info EDate: Using assembler routines when possible }
{$else}
{$Info EDate: Using All-Pascal routines as much as possible }
{$endif Useasm}
{$ASMMODE ATT}
{$ELSE}
 TYPE CARDINAL=LONGINT;
{$ENDIF}

{Constants}

TYPE    MonthStr    = ARRAY[1..12] OF String[3];
         DowsStr     = ARRAY[0..6] OF String[3];

        Ceremony =   (AshWednesday,GoodFriday,EasterSunday,
                      EasterMonday,AscensionDay,Whitsunday,
                      WhitsunMonday,CorpusChristiDay,RepentanceDay,
                      FirstAdvent,SecondAdvent,MothersDay);

CONST
   Months:MonthStr=('Jan','Feb','Mar','Apr','May','Jun','Jul',
                     'Aug','Sep','Oct','Nov','Dec');
   Dows  :DowsStr=('Sun','Mon','Tue','Wed','Thu','Fri','Sat');


{Primitives}

{ LeapYr defined as (Year DIV 4=0) AND ((Year DIV 100)<>0 OR (Year DIV 400)=0)
    or somthing like that. This routine has been done in assembler for years}
FUNCTION LeapYr( Year : LONGINT) : BOOLEAN;

{ Calculates daynumber, January 1st =1. Februari 1st=32. Uses LeapYr()

    Remember, a DayNr is NOT corresponding to a date. A year and a DayNr is
    however}
FUNCTION DayNr( Day,Month,Year : LONGINT) : LONGINT;

{ Calculates Day and Month back from DayNrIn and Year}
PROCEDURE DayNrBack( Year,DayNrIn:LONGINT; VAR Month,Day : LONGINT);

{ Real procedures}

{  DOW, Day of Week (a Sunday)=0; DOW( a Monday)=1 etc.
 Fixed for 29-2-2000, which IS a LeapYear. }
FUNCTION DOW( Year,Month,Day:LONGINT): LONGINT;

{ Unix dates, seconds after 1-1-70. Works fine till 29-2-2100
    (everything divisible by 4 is a leapyear, so also 2000),
  REMEMBER THIS VALUE IS BIG ENDIAN, LINUX USERS might have to uncomment
  some code, see implementation, I already prep'ed some code, I suppose
  Linux is bigendian, as it's original platform is intel-based?

Uses DayNr and (via DayNr) LeapYr.}
FUNCTION ToUnix(Year, Month, Day, Hrs, Mins, Secs : LONGINT):CARDINAL;

{ Unixdate (BigEndian) back to YYMMDD HHMMSS, expires at 29-2-2100}
PROCEDURE FromUnix(Unix : CARDINAL;VAR Year,Month,Day,Hour,Min,Sec:LONGINT);

{Calculate Weeknumbers, like shown in agendas or on calendars.}
FUNCTION WeekNr(Year,Mnth,Day:LONGINT):LONGINT;

{ Returns date of easter for given year. 1=March 1st. 32= April 1st}
FUNCTION Easter (Year : LONGINT) : LONGINT;

{ PAS

 DaTiToStr is a complex procedure which translates Date-Time to a string, in
a extemely configurable way. This procedure is the biggest EDate procedure.
The modula2 version is assembler, and that is the biggest assembler
procedure I've ever done.

Basically it copies Format to DateStr replacing
 %H with Hour(e.g. 1 and 11 ) %h with Hour, padded with zero (01 and 11)
 %M  "   Min                  %m  "   Min ,  "       "  "
 %S  "   Sec                  %s  "   Sec ,  "       "  "
 %D  "   Day                  %d  "   Day ,  "       "  "
 %O  "   Month                %o  "  Month,  "       "  "
 %Y  "  Year MOD 100 aligned with zeroes (because of e.g. the year 2001
                                           would become 1 instead of 01)
 %y  "  Year  (4 characters)
 %J (or %j) with 3-character month (Jan,Feb etc)
 %W (or %w) with 3-character dayofweek (Sun,Mon etc)
 %L tests the daynumber, if day=1 it outputs "st", for 2 "nd" etc.
 %i Changes the padding-character to a space, %I back to a zero,
    default is zero-padding.
 %U (or %u) is the same as %H (and %h), but it displays hours in the range 1..12
 %A displays AM or PM depending on the time, and the %a displays am or pm.

Format-string to see all options:

'%H:%M:%S  %h:%m:%s  %U:%m %a  %u:%m %A  %D-%O-%Y %d-%o-%y,'+
  %W %J %D%L, %y';

This string is also useful to check which options are affected
 with padding, prepend %i to above string to see space padding.

If you don't use certain options, you can enter an arbitrary constant for that
parameter. Checks on HIGH(DateStr) not implemented yet, please make sure that
it's big enough. Uses DOW and Length.

The FPK PAS-version is about 2K, the M2 ASM-version is 270 bytes (both
excl constants)

USE this procedure, and make the format string configurable! Users with other
date formats than MMDDYY will love it. Some more OS-dependant routines could
interrogate the OS and load the format string accordingly.}
PROCEDURE DaTiToStr(Hour,Min,Sec,Day,Month,Year:LONGINT;CONST Format:String;
                                           VAR DateStr : String);


{ Return the date for a certain movable ceremony.(see definition Ceremony-type,
most are related to easter.}

PROCEDURE MovableCeremony (    movableCeremon : Ceremony;
                               Year           : LONGINT;
                               VAR Day,Month  : LONGINT);

IMPLEMENTATION

{$IFNDEF UseAsm}

TYPE
   Array12Type = ARRAY [1..12] OF LONGINT;
CONST
   MonthDays : Array12Type=(31,28,31,30,31,30,31,31,30,31,30,31);
   MonthCumm : Array12Type=(0,31,59,90,120,151,181,212,243,273,304,334);

 FUNCTION LeapYr( Year : LONGINT) : BOOLEAN;
 {verified}
 BEGIN
   LeapYr:=(Year MOD 4 = 0) AND ((Year MOD 100 <> 0) OR (Year MOD 400 = 0));
END;

FUNCTION DayNr( Day,Month,Year: LONGINT) : LONGINT;
{verified}
VAR
   i : LONGINT;
BEGIN
   i := MonthCumm[Month]+Day;
   IF (Month > 2) AND LeapYr( Year ) THEN
      INC( i );
   DayNr:=i;
END ;

FUNCTION DOW( Year,Month,Day:LONGINT): LONGINT;

TYPE  Arr12     =       ARRAY[1..12] OF BYTE;
      TabType   =       ARRAY[0..3] OF BYTE;


CONST Tab       :    Arr12    =(0,3,3,6,1,4,6,2,5,0,3,5);
      CorTab    :    TabType  =(0,6,4,2);

VAR     iVAR1,
        iVAR2,A    : LONGINT;

BEGIN
 A:=0;
 iVAR1 := Year MOD 100;
 iVAR2 :=  Day  + iVAR1 + iVAR1 DIV 4;
 IF (Year>1999) AND (month>2) THEN {2000=leapyear correction}
  A:=1;
 iVAR2 := ( A+LONGINT(Tab[Month]) + iVAR2+CorTab[((Year DIV 100)+1) AND 3]) MOD 7;
 DOW:=iVAR2;
END;

FUNCTION ToUnix(Year, Month, Day,Hrs, Mins, Secs : LONGINT):CARDINAL;
{ Good till 1-3-2100}
VAR
   i        : LONGINT;
   UTS      : CARDINAL;
BEGIN
   IF Year < 1970 THEN
    UTS:=0
   ELSE
    BEGIN
      i:=Year-1970;
      I:=I*365 + (I+1) SHR 2+DayNr(Day,Month,Year)-1;
      UTS:=I*86400+Hrs*3600+Mins*60+Secs;
   END;
  ToUnix:=UTS;
END ;

PROCEDURE FromUnix(Unix: CARDINAL;VAR Year, Month, Day,Hour, Min, Sec : LONGINT);
VAR
   days,Temp: LONGINT;
BEGIN
   Sec  := LONGINT(Unix MOD 60);
   Min  := LONGINT(Unix DIV 60 MOD 60);
   Hour := LONGINT(Unix DIV 3600 MOD 24);
   days := LONGINT(Unix DIV 86400);
   Year := 1970;
   Temp := 365;
   WHILE days >= Temp DO
    BEGIN
      DEC( days, Temp );
      INC( Year );
      Temp:=365;
      IF (Year AND 3) =0 THEN
       INC(Temp);
   END;
   INC( days );   { Daynumber now correct }
   Month := 1;
   WHILE days > CARDINAL(MonthDays[Month]) DO
    BEGIN
      DEC( days, CARDINAL(MonthDays[Month]) );
      IF (Month = 2) AND LeapYr( Year ) THEN
         DEC( days );
      INC( Month );
    END;
   IF days = 0 THEN
    BEGIN
      days := 29;
      DEC( Month );
    END;
   Day := LONGINT(days);
END;

PROCEDURE DayNrBack( Year,DayNrIn:LONGINT; VAR Month,Day : LONGINT);
TYPE DyNrB=ARRAY[0..12] OF LONGINT;

CONST MonthDaysBack : DyNrB=(0,31,59,90,120,151,181,212,243,273,304,334,365);

VAR i,j,j2,LCor : LONGINT;

BEGIN
 I:=0;
 J:=MonthDaysBack[I];
 IF LeapYr(Year) THEN
  LCor:=1
 ELSE
  LCor:=0;
 WHILE DaynrIn>J DO
  BEGIN
   INC(I);
   J2:=J;
   J:=MonthDaysBack[I];
   IF (I>1) THEN
    INC(J,LCor);
   END;
 Month:=I; Day:=DayNrIn-J2;
END;

FUNCTION Easter(Year : LONGINT):LONGINT;
   { Easter date calculation, return=1 = 1 march; return=32= 1 april }
VAR
   GoldenNo,
   Sun,
   Century,
   LeapCent,
   LunarCorr,
   Epact,
   FullMoon : INTEGER;
BEGIN
   Century := (Year DIV 100) + 1;
   LeapCent := (3 * Century DIV 4) - 12;
   LunarCorr := ((8 * Century + 5) DIV 25) - 5;
   GoldenNo := (Year MOD 19) + 1;
   Epact := 11 * GoldenNo + 20 + LunarCorr - LeapCent;
   Epact:=ABS(Epact ) MOD 30;
   Sun := (5 * INTEGER(Year) DIV 4) - LeapCent - 10;
   IF ((Epact = 25) AND (GoldenNo > 11)) OR (Epact = 24) THEN
    INC(Epact);
   FullMoon := 44 - Epact;
   IF FullMoon < 21 THEN
    INC(FullMoon, 30);
   Easter:=FullMoon + 7 - ((Sun + FullMoon) MOD 7);
END;

FUNCTION WeekNr( Year, Mnth, Day : LONGINT) : LONGINT;
VAR  Week, Dow2 : LONGINT;
BEGIN
   Dow2 := (DOW( Year, 1, 1 )+6) MOD 7; { Set switchdate to Monday }
   Week := ((DayNr( Day, Mnth, Year)+Dow2-1) DIV 7)-1;
   IF Dow2 < 4 THEN
      INC( Week );
   IF Week = 0 THEN
     Week := WeekNr( Year-1, 12, 31 );
   WeekNr:=Week;
END;

{$ELSE}

FUNCTION LeapYr( Year : LONGINT) : BOOLEAN; ASSEMBLER;

asm
                push    %ebx            // ease calls from other procedures
                push    %ecx
                push    %edx
                movl    Year,%eax
                testl    $3,%eax
                jne     .LLeapEnd
                movl    $100,%ebx
                {$ifdef NOATTCDQ}
                  cltd
                {$else}
                  cdq
                {$endif}
                idivl   %ebx
                testb   %dl,%dl
                jne     .LLeapTrue
                and     $3,%eax
                je      .LLeapTrue
 .LLeapEnd:     setb    %al
                jmp     .LLeapExit
 .LLeapTrue:    movw   $1,%ax
 .LLeapExit:    pop     %edx
                pop     %ecx
                pop     %ebx
end ['EAX'];  // procedure saves %ebx..%edx, for internal calling.

{ Calculates daynumber, January 1st =1. incl leapyear(LeapYr()) correction }
FUNCTION DayNr( Day, Month, Year: LONGINT) : LONGINT; ASSEMBLER;
TYPE DyNr=ARRAY[0..11] OF LONGINT;

CONST MonthDays : DyNr=(0,31,59,90,120,151,181,212,243,273,304,334);

ASM
                        movl    Month,%ebx
                        movl    Day,%ecx
                        cmp     $2,%ebx
                        jbe     .LDayNr2
             .LDayNr1:  push    Year
                        call    LeapYr
                        add     %eax,%ecx
              .LDayNr2: decl    %ebx
                        movl    MonthDays(,%ebx,4),%eax
                        add     %ecx,%eax
END ['EAX','EBX','ECX'];

FUNCTION DOW( Year,Month,Day:LONGINT): LONGINT; ASSEMBLER;
  {  DOW( a Sunday)=0; DOW( a Monday)=1
      Ultrafast since it works with tables. }

TYPE  Arr12     =       ARRAY[0..11] OF BYTE;
      TabType=ARRAY[0..3] OF BYTE;

CONST TabCent   : TabType = (0,6,4,2);          { Century correction}
      TabMonth  : Arr12   = (0,3,3,6,1,4,6,2,5,0,3,5); {Month indices}

     ASM

                xorl   %ecx,%ecx
                mov    Year,%eax
                cmp    $2000,%eax
                jb     .LBelow2000
                cmp    $3,Month
                jb     .LBelow2000
                incl   %ecx

.LBelow2000:    movl   $100,%ebx
                {$ifdef NOATTCDQ}
                  cltd
                {$else}
                  cdq
                {$endif}

                idivl  %ebx
                incl   %eax
                andl   $3,%eax

                movzbl TabCent(,%eax),%ebx
                                 // ebx contains century correction
                mov    Day,%eax
                addl   %eax,%ebx      // Add Day
                addl   %edx,%ebx      // Add Year MOD 100
                shrl   $2,%edx
                addl   %edx,%ebx      //  Add (Year MOD 100) SHR 2
                mov    Month,%eax  // Month
                decl   %eax
                movzbl TabMonth(,%eax),%edx // is index
                addl   %edx,%ebx      //  Add TabMonth[Month]
                movl   $7,%eax
                xchgl  %eax,%ebx
                addl   %ecx,%eax
                incl   %eax
                {$ifdef NOATTCDQ}
                  cltd
                {$else}
                  cdq
                {$endif}

                idivl  %ebx           // Divide total by 7
                movl   %edx,%eax
END ['EAX','EBX','ECX','EDX'];

{ calculates Unixdate(seconds past 1-1-1970) }
FUNCTION ToUnix(Year, Month, Day, Hrs, Mins, Secs : LONGINT):CARDINAL; ASSEMBLER;

ASM
                          push    Year
                          push    Month
                          push    Day
                          call    DayNr
                          decl    %eax
                          movl    %eax,%ebx
                          movl    Year,%eax
                          subl    $1970,%eax
                          jb      .LTrouble            // Date earlier than 1970
                          mov     %eax,%esi
                          incl    %esi
                          shrl    $2,%esi
                          imull   $365,%eax
                          addl    %ebx,%eax
                          addl    %esi,%eax

                          imull   $24,%eax
                          movl    Hrs,%edx
                          addl    %edx,%eax

                          imull   $60,%eax
                          movl    Mins,%edx
                          addl    %edx,%eax

                          imull   $60,%eax
                          movl    Secs,%edx
                          addl    %edx,%eax
 { LINUX might need to insert a BIG to LITTLE endian here.
               486+       bswap32 %eax             //486+ instruction
               386+       movl    %eax,%ebx
                          shrl    %16,%ebx
                          xchgb   %ah,%al
                          xchgb   %bl,%bh
                          shll    %16,%eax
                          addl    %ebx,%eax
 END to Little endian conversion}
                          jmp     .LToUnixExit
           .LTrouble:     xorl    %eax,%eax
           .LToUnixExit:
END ['EAX','EBX','ECX','EDX','ESI'];

PROCEDURE FromUnix(Unix : CARDINAL;VAR Year,Month,Day,Hour,Min,Sec:LONGINT); ASSEMBLER;

TYPE FrU=ARRAY[0..11] OF BYTE;

CONST MonthDays2 : FrU=(31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);


        ASM
                mov   Year,%eax
                { Unix Big To Little Endian change here, see ToUnix}
                mov   $60,%ecx
                xorl  %edx,%edx
                idiv  %ecx
                movl  Sec,%ebx                 // mod 60 is seconds
                movl  %edx,(%ebx)
                xorl  %edx,%edx
                idiv  %ecx
                movl  Min,%ebx                 // again mod 60 is minutes
                movl  %edx,(%ebx)
                movl  $24,%ecx
                xorl  %edx,%edx
                idiv  %ecx
                movl  Hour,%ebx                 // again mod 24 is hours
                movl  %edx,(%ebx)
                                          // eax= total days
                movl  $1970,%ecx
.LFromNextYear: mov   $365,%ebx
                test  $3,%ecx         // LeapYear= Year AND 3=0 valid till 2100
                jne   .LFromNoLeap
                incl   %ebx
.LFromNoLeap:   cmp   %ebx,%eax
                jb    .LGoodYear
                subl  %ebx,%eax
                incl  %ecx
                jmp   .LFromNextYear
.LGoodYear:     movl  Year,%ebx
                movl  %ecx,(%ebx)          // save years
                        {Rest is mere DayNrBack as inline}
                incl %eax
                mov   %eax,%edx
                mov   %ecx,%eax
                xor   %ebx,%ebx
.LLoopie:                                 // ax=days, dx=years, bx is months
                movzbl MonthDays2(,%ebx,1),%ecx // is index
                cmpl  $1,%ebx             // Correct for leapyear
                jne   .LNoFebr
                testb $3,%al
                jne   .LNoFebr
                incl  %ecx
.LNoFebr:       cmpl  %ecx,%edx
                jbe   .LDayOk
                subl  %ecx,%edx
                incl   %ebx
                cmp   $12,%ebx
                jb    .LLoopie
                movl  $11,%ebx          // If it reaches this point, you
                                        // are in trouble
 .LDayOk:       incl  %ebx
                movl  Month,%ecx     // everything ok, save day and month
                movl  %ebx,(%ecx)
                movl  Day,%ebx
                movl  %edx,(%ebx)
end ['EAX','EBX','ECX','EDX'];

PROCEDURE DayNrBack( Year,DayNrIn:LONGINT; VAR Month,Day : LONGINT); ASSEMBLER;
TYPE DyNrB=ARRAY[0..13] OF WORD;

CONST MonthDays : DyNrB=(0,31,59,90,120,151,181,212,243,273,304,334,365,365);

ASM
                movl   Year,%edx
                movl   DayNrIn,%eax
                xorl   %ebx,%ebx
                xorl   %ecx,%ecx
  .LMonthLoop:  movl   %ecx,%esi
                movzbl MONTHDAYS(,%ebx,2),%ecx
                cmpl   $2,%ebx             // Correct for leapyear
                jne    .LNoFebr2
                test   $3,%edx
                jne    .LNoFebr2
                incl   %ecx
  .LNoFebr2:    cmp    $3,%ebx
                jne    .LSufssd
                mov    %ecx,%eax
                jmp    .LPutin
.LSufssd:       incl   %ebx
                cmpl   %ecx,%eax
                ja     .LMonthLoop
                dec    %ebx
                subl   %esi,%eax
.LPutin:        movl   Month,%edx
                movl   %ebx,(%edx)
                movl   Day,%edx
                movl   %eax,(%edx)
end ['EAX','EBX','ECX','EDX','ESI'];

{ Returns date of easter for given year. 1=March 1st. 32= April 1st}
FUNCTION Easter (Year : LONGINT) : LONGINT; ASSEMBLER;

{ Based on a function in swag, the variable names in the comments refer
    to that}
                ASM
                       movl   Year,%esi
                       movl   %esi,%eax
                       movl   $100,%ebx
                       {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}

                       idiv   %ebx
                       inc    %eax
                       mov    %eax,%ecx             // ecx=Century
                       imull   $3,%eax
                       shrl   $2,%eax
                       subl   $12,%eax
                       movl   %eax,%edi             // di=LeapCent
                       mov    %ecx,%eax
                       shll   $3,%eax
                       addl   $5,%eax

                       mov    $25,%ebx
                       {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}

                       idiv    %ebx
                       subl    $5,%eax
                       movl    %eax,%ecx         // ecx=LunarCorr
                       movl    %esi,%eax
                       movl    $19,%ebx
                       {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}
                       idiv    %ebx
                       incl    %edx
                       mov     %edi,%ebx          // LeapCent
                       subl    %ebx,%ecx
                       addl    $20,%ecx
                       pushl   %edx             // GoldenNo on stack
                       imull    $11,%edx
                       addl    %ecx,%edx        // edx is Epact before ABS
                       orl     %edx,%edx
                       jns     .LEasterPositive
                       negl    %edx
.LEasterPositive:      movl   %edx,%eax
                       movl   $30,%ebx
                       {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}

                       idivl  %ebx             //edx is Epact after ABS AND MOD
                       pop    %ecx             //ecx=goldeno
                       cmp   $24,%edx
                       je    .LEasterInc
                       cmp    $25,%edx
                       jne   .LEasterInc2
                       cmp   $11,%ecx
                       jbe   .LEasterInc2
          .LEasterInc: inc   %edx
        .LEasterInc2:  movl $44,%eax
                       subl %edx,%eax         // Fullmoon = 44-epact
                       cmpl $21,%eax
                       jge  .LFullMoonOk
                       addl $30,%eax
.LFullMoonOk:          imull $5,%esi
                       shrl $2,%esi
                       subl %edi,%esi
                       subl $10,%esi         // si=Sun=5*Y SHR 2 -LeapCent-10
                       addl %eax,%esi
                       xchgl %eax,%esi
                       mov  $7,%ebx
                       {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}

                       idivl %ebx
                       mov $7,%eax
                       subl %edx,%eax        // ax=7-(Sun + FullMoon MOD 7)
                       addl %esi,%eax
END ['EAX','EBX','ECX','EDX','ESI','EDI'];


{ Calculate Weeknumbers, like shown in agenda's}
FUNCTION WeekNr(Year,Mnth,Day:LONGINT):LONGINT; ASSEMBLER;

        ASM
                     pushl  $1
                     pushl  $1
                     pushl  Year
                     call   Dow
                     addl   $6,%eax
                     movl   $7,%ebx
                     {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}
                     idivl  %ebx            // edx= (DOW(year,1,1) +6) mod 7
                     pushl  %edx
                     push   Year
                     push   Mnth
                     push   Day
                     call   DayNr
                     pop    %ecx
                     mov    $7,%ebx
                     add    %ecx,%eax
                     decl   %eax
                     {$ifdef NOATTCDQ}
                        cltd
                       {$else}
                        cdq
                       {$endif}

                     idiv   %ebx         //  eax=edx(above)+DayNr(Y,M,D)-1 DIV 7)
                     cmp    $4,%ecx
                     jae    .LWeeknrdontinc
                     incl  %eax
.LWeeknrdontinc:     test   %eax,%eax
                     jne    .Lweeknrexit
                     pushl  $31            // Get last years last week.
                     pushl  $12            // with a single recursion.
                     movl  Year,%eax
                     decl  %eax
                     push  %eax
                     call  WeekNr
.Lweeknrexit:
END ['EAX','EBX','ECX','EDX'];

{$ENDIF}

{ Following procedures don't have the 'other' equivalent yet}


PROCEDURE DaTiToStr(Hour,Min,Sec,Day,Month,Year:LONGINT;CONST Format:String;
                                           VAR DateStr : String);

VAR Len,A,B,C : LONGINT;
    PadChar   : CHAR;

PROCEDURE SlipIn(Value:LONGINT;Pad:BOOLEAN);

BEGIN
 IF Value>9 THEN
  BEGIN
   DateStr[B-2]:=CHR(48+Value DIV 10);
   DateStr[B-1]:=CHR(48+Value MOD 10);
  END
 ELSE
  IF Pad THEN
   BEGIN
    DateStr[B-2]:=PadChar;
    DateStr[B-1]:=CHR(48+Value);
   END
  ELSE
   BEGIN
    DateStr[B-2]:=CHR(48+Value);
    DEC(B);
   END;
END; {SlipIn}

PROCEDURE BinPut(Value:LONGINT);

BEGIN
   DateStr[B-2]:=CHR(Value AND 255);
   DateStr[B-1]:=CHR(Value shr 8);
END; {BinPut}


BEGIN
 PadChar:='0';
 Len:=Length(Format);
 A:=1; B:=1;
 REPEAT
  WHILE (A<=Len) AND (Format[A]<>'%') DO
   BEGIN
    DateStr[B]:=Format[A];
    INC(A);
    INC(B);
   END;
  IF Format[A]='%' THEN
   BEGIN
    INC(A,2);
    INC(B,2);
    CASE Format[A-1] OF
      'H' :  Slipin(Hour,False);
      'h' : Slipin(Hour,TRUE);
      'M' : Slipin(Min,FALSE);
      'm' : Slipin(Min,TRUE);
      'S' : Slipin(Sec,FALSE);
      's' : Slipin(Sec,TRUE);
      'D' : Slipin(Day,FALSE);
      'd' : Slipin(Day,TRUE);
      'O' : Slipin(Month,FALSE);
      'o' : Slipin(Month,TRUE);
      'Y' : Slipin(Year Mod 100,TRUE);
      'y' : BEGIN
             Slipin(Year DIV 100,TRUE);
             INC(B,2);
             Slipin(Year MOD 100,TRUE);
            END;
      'J','j' : BEGIN
                 DateStr[B-2]:=Months[Month][1]; {won't compile here}
                 DateStr[B-1]:=Months[Month][2];
                 DateStr[B  ]:=Months[Month][3];
                 INC(B);
                END;
      'W','w' : BEGIN
                 C:=DOW(Year,Month,Day);
                 DateStr[B-2]:=Dows[C][1];
                 DateStr[B-1]:=Dows[C][2];
                 DateStr[B  ]:=Dows[C][3];
                 INC(B);
                END;
       'i'   : PadChar:=' ';
       'I'   : PadChar:='0';
       'U'   : Slipin(Hour MOD 13,FALSE);
       'u'   : Slipin(Hour MOD 13,TRUE);
       'A'   : IF (Hour div 13)=0 THEN
                BinPut($4D41)
               ELSE
                BinPut($4d50);
        'a'   : IF (Hour div 13)=0 THEN
                BinPut($6D61)
               ELSE
                BinPut($6D70);
       'L'   : CASE Day OF
                1,21,31 : BinPut($7473);            { 'st'}
                2,22    : BinPut($646E);            { 'nd'}
                3,23    : BinPut($6472);            { 'nd'}
               ELSE
                BinPut($6874);                      { 'th'}
                END;
           END;
        END;

  UNTIL (A>=Len);
  DateStr[0]:=CHR(B-1);

END;

PROCEDURE MovableCeremony (    movableCeremon : Ceremony;
                               Year           : LONGINT;
                               VAR Day,Month  : LONGINT);


TYPE DateDiffType=ARRAY[0..7] OF BYTE;

CONST DateDiff: DateDiffType=(0,44,46,47,85,95,96,106);

VAR  DayNrEaster, DiffEaster : LONGINT;
  BEGIN
    IF movableCeremon = RepentanceDay THEN
     BEGIN
      Month:=11;
      Day:=25-(DOW (year, 11, 16)+6) MOD 7;
      IF Day>22 THEN
       DEC(Day,7);
      EXIT;
     END;
    write('1');
    IF (movableCeremon = FirstAdvent) OR
       (movableCeremon = SecondAdvent) THEN
      BEGIN
       DiffEaster:=DayNr(24,12,year);
       IF movableCeremon = FirstAdvent THEN
         DEC(DiffEaster, 22)
       ELSE
         DEC(DiffEaster,15);
       DEC(DiffEaster,(DOW (year, 12, 24)+6) MOD 7);
       DayNrBack (year,DiffEaster , Month,Day);
       EXIT;
      END;
    { AshWednesday (Easter-46) as basis, since it's the earliest
        Ceremony. The corrections are 31 (januari) + 28 (Februari)
        + LeapYr(year) (29th Feb when necessary) -46  }
    DayNrEaster:=Easter(year)+13; {+31+48-46+LeapYr(year)}
    IF LeapYr(year) THEN
     INC(DayNrEaster);
    write('x ',DayNrEaster,' x');
    CASE movableCeremon OF
       MothersDay   : BEGIN
                        DiffEaster:=(DOW(year,5,1)+6) MOD 7;
                        IF DayNr(14-DiffEaster,5,year) = (DayNrEaster+95) THEN
                         DiffEaster := 88
                        ELSE
                         BEGIN
                          Month:=5;
                          Day:= 14-DiffEaster;
                         END
                        END
     ELSE
      { Not really elegant, but decreases size by 25%!!! }
      BEGIN
       DiffEaster:=LONGINT(DateDiff[ORD(movableCeremon)]);
       write('x ',DiffEaster,' x');
       DayNrBack(Year,DayNrEaster+DiffEaster,Month,Day);
      END;
     END;
{Instead of :

| ashWednesday    : DiffEaster := 0;
| goodFriday      : DiffEaster :=  44;
| easterSunday    : DiffEaster :=  46;
| easterMonday    : DiffEaster :=  47;
| ascensionDay    : DiffEaster :=  85;
| whitsunday      : DiffEaster :=  95;
| whitsunMonday   : DiffEaster :=  96;
| corpusChristiDay: DiffEaster := 106;}

  END;
END.

