UNIT EDirGlob;

INTERFACE

{ DirTree, directory search systems.

   This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Free for non-commercial use.

ALL PROGRAMS USING THIS UNIT SHOULD BE COMPILED WITH A $M 65520 0 655360
directive! (BP only)

Module searches a path and below for a file ( * and ? allowed), a
system to read a directory tree into memory (the module name already
refers to it) is under development, but fully working. The place where a
matching file is found is reported to a procedure of the type ReportProc, with
complete (I hope) path and a SearchRec with information like
the filedate and -name etc.

The module is no real speedmonster, but the speed seems satisfactory,
probably because I run HyperDisk cache software :-)
The native Win95 caching will also do. (I've Busmastering installed though
and 48 MB mainmem. Window uses all unused memory but 1 or 2 MBs for caching,
which is faster than HyperDisk ( which only receives 2-4 MB max))

I think you can't get it much faster, not with FindFirst and FindNext at
the basis.

I need it because I made a CD of all my software, and I wanted to create
indexes.

Demoes : TreeDemo, Indexer,CdSize and FileFind. (Some not yet in XTDLIB/FPC)

}
Uses Linux;

TYPE  { first parameter  : path where the file is found,
         second parameter : SearchRec of found file, FPC-help for
                             details}
      FilAttr       = BYTE;

      Fileptr       = ^FilesRec;
      FilesRec       = RECORD
                       Next   : FilePtr;
                       Name   : String;
                       FileStat : Stat;
                       END;


      DirTreePoint  = ^DirTreeRecord;
      DirTreeRecord = RECORD
                       NextDir,
                       SubDirs  : DirTreePoint;
                       Name     : String;
                       DirStat  : Stat
                       Files    : FilePtr;
                       END;
      ReportProc    = PROCEDURE (Path:PathStr;StatRec:Stat);
      FileProc      = PROCEDURE(P:PathStr;StatRec:Stat);
      DirProc       = FileProc;
      DetectProc    = FUNCTION(P:PathStr;StatRec:Stat):BOOLEAN;

VAR
    { Module parameter, should a matching DIRECTORYname be reported?
       Default : NO (FALSE) }

    DirsToo      : BOOLEAN;

{ Statistics over last search are kept here.
    valid for procedures FileScan and SearchForFiles }

    FoundCount   : LONGINT;    { Nr. files found }
    TotalBytes   : LONGINT; { Total bytes in files found }

    { Nr directories found in last scan, FileScan and BuildTree }

    FoundDirs    : LONGINT;    { Directories found (except . and ..)}
    ClusterSize  : LONGINT; { Clustersize used for rounding, default
                                   0= no rounding }

{ Set attributes used for all FindFirst's in EDirGlob. Directory attribute
is added or cleared when necessary(if the program searches for directories).}
PROCEDURE SetFAttr (Attr:FilAttr); {Does nothing yet in EDirGlob}

{ Clears statistics. Also initcode, when using OVL117 mode }
PROCEDURE ClearStat;

(* Filesearch in path RootDir and in its subdirectories, for files matching
    FileName (directories are regarded as files when DirsToo=TRUE).
   Files are reported to the procedure Report, with all information.

To build a directory tree, enter "." as filename, and make DirsToo TRUE.
 A demo of this is Indexer. This system works best when the editing in
 each directory is simple, for more advanced directory playing, use
 BuildTree*)
PROCEDURE FileScan(RootDir,FileName : String;Report:ReportProc);

{ ---- Tree building ---- }

{ These procedures build directory trees in memory, and manipulate them,
the difference is only noticable when doing multiple scans in memory, or
when you have to search for several patterns.
For simple scans this procedure is slower (one read extra to get the
directories). Win95's (and probably Linux' too) caching however smoothes this,
but in can be annoying in dos, unless you load HyperDisk.}

{ Load the directory structure starting with RootDir into memory }
FUNCTION BuildTree(CONST RootDir: PChar):DirTreePoint;

{ Use after a BuildTree; searches all directories in memory for occurance
    of Pattern, and adds those to the tree under the "files" field.
    Pattern is something like "*.txt"
    1.04 : Can be used several times, for more than one extension/pattern.
    }
PROCEDURE SearchForFiles(Root:DirTreePoint;CONST InPattern:PChar;Select:DetectProc);

{ Scan tree after BuildTree (and SearchForFiles), perform DoFile on each
file, and DoDir on each Directory if DirsToo).}
PROCEDURE ScanTree(Root : DirTreePoint;DoFile:FileProc;DoDir : DirProc);

{ Remove a Tree made with BuildTree and/or SearchForFiles from memory }
PROCEDURE KillFileTree(VAR RootIn:DirTreePoint);

IMPLEMENTATION


{$ifdef FPC}
Uses EPasStr,Strings;
{$ELSE}
 Uses Strings;
{$endif FPC}

VAR PrintEntry : ReportProc;
    DirAttr,                            { Set by SetFAttr, for fast usage}
    NoDirAttr  : FilAttr;               { By the routines}

PROCEDURE SetFAttr (Attr:FilAttr);

BEGIN
END; {SetFAttr}

PROCEDURE SearchCurrent(CONST Dir : String;CONST Target : PathStr);

VAR
  G1,G2     : PGlob;

BEGIN
//  Path:=concat(Dir,Target);
  G1:=Glob(Dir+Target);
  IF LinuxError=0 THEN
  WHILE G2DosError=0 DO
   BEGIN
    G2:=G1;
    IF ((FoundFile.attr AND Directory)=0) OR DirsToo THEN
      BEGIN
          INC(FoundCount);
          IF ClusterSize=0 THEN
           INC(TotalBytes, FoundFile.size)
          ELSE
           INC(TotalBytes, ((FoundFile.size DIV ClusterSize)+1)*ClusterSize);
          PrintEntry(Dir,FoundFile);
         END
        ELSE
          IF FoundFile.Name[1] <>'.' THEN
           INC(FoundDirs);
       FindNext(FoundFile);
    END; {read loop}
 {$IFDEF CloseFind}
  FindClose(FoundFile);
 {$ENDIF}
END; {SearchCurrent}

PROCEDURE SearchSubDirs(CONST Dir : PathStr;CONST Target : PathStr);
VAR
  FoundDir  : SearchRec; { Current search data. }
  FileSpec  : String;  { Always contains "DIR" + "*." }
  Path      : String;

BEGIN
 SearchCurrent(Dir,Target);
 FileSpec:=concat(Dir,'*.');
 FindFirst(FileSpec,DirAttr, FoundDir);
 WHILE DosError=0 DO
  BEGIN
     WITH FoundDir DO
      BEGIN
       IF Name[1] <> '.' THEN
        IF ((FoundDir.Attr AND directory)<>0) AND ((FoundDir.Attr AND Hidden)=0) THEN
         BEGIN
         {$ifdef Linux}
          Path:=Dir+FoundDir.Name+'/';
         {$else}
          Path:=Dir+FoundDir.Name+'\';
         {$endif}
          SearchSubDirs(Path ,Target);
        END;
      END; {WITH FoundDir}
     FindNext(FoundDir);
   END; {While loop}
 {$IFDEF CloseFind}
  FindClose(FoundDir);
 {$ENDIF}
END;{ SearchSubDirs}

PROCEDURE FileScan(RootDir,FileName : PChar;Report:ReportProc);

VAR Dir    : PathStr;
    Target : PathStr;

BEGIN
 Dir:=StrPas(RootDir);
 Target:=StrPas(FileName);
 {$ifdef FPC}
   AppendBackSlash(Dir); {procedure is safe for linux}
 {$else}
   {Borland doesn't compile under Linux, other compilers not supported :-)}
 IF Dir[ORD(Dir[0])] <> '\' THEN
  Dir:=Concat(Dir,'\');
 {$endif FPC}
 FoundCount:=0;
 FoundDirs:=0;
 TotalBytes:=0;
 PrintEntry:=Report;
 SearchSubDirs(Dir,Target);
END;


{ Internal procedure, but in fact the main procedure the other part (BuildTree)
 is mere interfacial}
PROCEDURE TreeSearch  (CONST Dir : PathStr;Root:DirTreePoint);
VAR
  FoundDir  : SearchRec; { Current search data. }
  FileSpec  : PathStr;  { Always contains "DIR" + "*." }
  Cur       : DirTreePoint;
  First     : BOOLEAN;
BEGIN
 FileSpec:=concat(Dir,'*.');
 FindFirst(FileSpec,DirAttr,FoundDir);
 First:=TRUE;
 Cur:=Root;
 WHILE DosError=0 DO
   BEGIN
     WITH FoundDir DO
       BEGIN
         IF FoundDir.Name[1] <> '.' THEN
           BEGIN
             IF (FoundDir.Attr and directory)<>0  THEN
               BEGIN
                 INC (FoundDirs);
                 IF First THEN
                   BEGIN
                     NEW(Cur^.SubDirs);
                     Cur:=Cur^.SubDirs;
                     First:=FALSE;
                   END
                 ELSE
                   BEGIN
                    NEW(Cur^.NextDir);
                    Cur:=Cur^.NextDir;
                   END; {IF FIrst}
                 Cur^.SubDirs:=NIL;
                 Cur^.NextDir:=NIL;
                 Cur^.Files  :=NIL;
                 {$Ifdef Linux}
                  Cur^.Name:=Dir+FoundDir.Name+'/';
                 {$else}
                  Cur^.Name:=Dir+FoundDir.Name+'\';
                 {$endif}
                TreeSearch(Cur^.Name ,Cur);
                END; {Attr and directory}
              END; { IF Name<>}
         END; {WITH FoundDir}
      FindNext(FoundDir);
   END; {read loop}
 {$IFDEF CloseFind}
  FindClose(FoundDir);
 {$ENDIF}
END; {TreeSearch}

FUNCTION BuildTree(CONST RootDir:PChar):DirTreePoint;

VAR
    Root   : DirTreePoint;
    DummyStr : PathStr;

BEGIN
 DummyStr:=StrPas(RootDir);
 {$ifdef FPC}
   AppendBackSlash(DummyStr); {procedure is safe for linux}
 {$else}
  {Borland doesn't compile under Linux, other compilers not supported :-)}
  IF DummyStr[ORD(DummyStr[0])] <> '\' THEN
  Dummystr:=Concat(DummyStr,'\');
 {$endif FPC}
 FoundDirs:=0;
 NEW(Root);
 Root^.NextDir:=NIL;
 Root^.SubDirs:=NIL;
 Root^.Files  :=NIL;
 TreeSearch(DummyStr,Root);
 BuildTree:=Root;
END;


PROCEDURE SearchForFiles(Root:DirTreePoint;CONST InPattern:PChar;Select:DetectProc);

VAR Pattern   :  String;

PROCEDURE SearchForFile(DirPtr:DirTreePoint);

VAR D          : SearchRec;
    Current    : FilePtr;
    FileSpec   : PathStr;
    First      : BOOLEAN;

BEGIN
 FileSpec:=DirPtr^.Name+Pattern;
 First:=TRUE;
 FindFirst(FileSpec,DirAttr,D);
 IF DosError=0 THEN
   REPEAT
     IF Select(D,DirPtr^.Name) THEN
       BEGIN
         IF First THEN  {Check for files already appended to the tree}
         BEGIN
           IF DirPtr^.Files = NIL THEN
             BEGIN
               NEW(DirPtr^.Files);
               Current:=DirPtr^.Files;
             END
           ELSE
             BEGIN
               Current:=DirPtr^.Files;
               WHILE Current^.Next <> NIL DO
                 Current:=Current^.Next;
               NEW(Current^.Next);
               Current:=Current^.Next;
             END;
          First:=FALSE;
        END
      ELSE
        BEGIN
          NEW(Current^.Next);
          Current:=Current^.Next;
        END;
     Current^.Next:=NIL;
     Current^.DirE:=D;
     INC(FoundCount);
     INC(TotalBytes,D.size);
   END;
   FindNext(D);
  UNTIL DOSERROR<>0;
 {$IFDEF CloseFind}
  FindClose(D);
 {$ENDIF}
 IF DirPtr^.SubDirs <> NIL THEN
  SearchForFile(DirPtr^.SubDirs);
 IF DirPtr^.NextDir <> NIL THEN
  SearchForFile(DirPtr^.NextDir);
END;

BEGIN
 Pattern:=StrPas(InPattern);
 FoundCount:=0;
 TotalBytes:=0;
 SearchForFile(Root);
END;

PROCEDURE KillFileTree(VAR RootIn:DirTreePoint);

PROCEDURE Kill1Tree(Root : DirTreePoint);

VAR
    CurrentFile,
    CurrentFile2 : FilePtr;

BEGIN
  IF Root^.Files <> NIL THEN
   BEGIN
     CurrentFile:=Root^.Files;
     WHILE(CurrentFile^.Next <> NIL) DO
       BEGIN
         CurrentFile2:=CurrentFile;
         CurrentFile:=CurrentFile^.Next;
         DISPOSE(CurrentFile2);
       END;
     DISPOSE(CurrentFile);
     Root^.Files:=NIL;
   END;
  IF Root^.SubDirs <> NIL THEN
   BEGIN
     Kill1Tree(Root^.SubDirs);
     Root^.SubDirs:=NIL;
   END;
  IF Root^.NextDir <> NIL THEN
   BEGIN
     Kill1Tree(Root^.NextDir);
     Root^.NextDir:=NIL;
   END;
  DISPOSE(Root);
END; { Kill1Tree}


BEGIN
 Kill1Tree(RootIn);
END;{KillFileTree}

PROCEDURE ScanTree(Root : DirTreePoint;DoFile:FileProc;DoDir : DirProc);

PROCEDURE Scan1Tree(Root : DirTreePoint);

VAR
    CurrentFile :     FilePtr;

BEGIN
  IF DirsToo THEN
   DoDir(Root^.Name);
  IF Root^.Files <> NIL THEN
   BEGIN
   CurrentFile:=Root^.Files;
   DoFile(Root^.Name,CurrentFile^.DirE);
   WHILE CurrentFile^.Next <> NIL DO
    BEGIN
      CurrentFile:=CurrentFile^.Next;
      DoFile(Root^.Name,CurrentFile^.DirE);
    END;
   END;
  IF Root^.SubDirs <> NIL THEN
   Scan1Tree(Root^.SubDirs);
  IF Root^.NextDir <> NIL THEN
   Scan1Tree(Root^.NextDir);
END;{ Scan1Tree;}

BEGIN
 Scan1Tree(Root);
END;{ ScanTree}

PROCEDURE ClearStat; { Initcode EDirGlob in procedure form, so it can be used
                        to re-init module}

BEGIN
 FoundCount:=0;         TotalBytes:=0;
 FoundDirs:=0;          DirsToo:=FALSE; { Default : File search only }
 ClusterSize:=0;
 SetFAttr(dos.hidden+dos.readonly+dos.archive+dos.sysfile);
END;

BEGIN
 ClearStat;
END.

