UNIT EDos;
{  Dos functions. Some functions (all up till now) only exist under Dos),
   or do nothing under Linux.
   Clustersize could give the inode size under Linux.

   This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)   }

INTERFACE
{$DEFINE USEASM}                {Intel asssembler is probably a lot faster}
                                { Not used in this module. Dos requires Intel}

{$I syseq.inc}                  { Some conditionals only, no actual code}

{$IFDEF Go32V2}
Uses Dos;
{$ENDIF}

{$IFDEF Turbo}
 TYPE CARDINAL=LONGINT;
{$ENDIF}

{$IFDEF go32v2}

TYPE PagerType = (Res0,SW_NOPAGER,SW_DOSPAGER,SW_IOSPAGER);
     TSwapInfo     = RECORD                     { GetSwapFileInfo record }
                      FileName : String[255]; { gives information about}
                      FileSize : COMP;{ swap/page file }
                      Pager    : PagerType;
                      END;

     CMOSRec= PACKED RECORD { Structure Cmos }
            Sec,               AlSec,
            Min,               AlMin,
            Hr,                AlHr,
            DayOfWeek,
            Day,
            Month,
            Year,
            RTCA,              RTCB,
            RTCC,              RTCD,
            POSTStatus,
            ShutDownStatus,
            DiskType,          Res1,
            HDType,            Res2,

            Equipment        : BYTE;
            BaseMem,
            XTDMem           : WORD;
            HD_C,
            HD_D             : BYTE;
            Res3             : ARRAY[0..18] OF BYTE;
            Checksum,
            XTDMem2          : WORD;
            Century,
            Misc             : BYTE;
            Res4             : ARRAY[0..11] OF BYTE;
            Res5             : ARRAY[0..63] OF BYTE;
            END;

CONST Print   =$001;
      PcLan   =$002;     { PC LAN redirector }
      DosCrit =$005;     { Dos critical errorhandler installed }
      AssignF =$006;
      Driver  =$008;     { Driver.sys }
      Share   =$010;     { Share.exe or Win95 sharing}
      Network =$011;     { Network redirector }
      NlsFunc =$014;     { NlsFunc.com }
      Graphics=$015;     { Graphics.com or CDROM MSCDEX}
      MscDex  =$015;
      Windows =$016;     { Windows, (special returnvalues :
                                  0       -> Not running under Windows
                                  2       -> running under Windows 2.x
                                  else    Win 3.0+
                                   returnvalue =versionnumber.
                                    Win 3.1 returns 0310H
                                    Win 95  returns 0400H }
      ShellB  =$019;
      Ansi    =$01A;
      Xma2Ems =$01B;

      XMSDrv  =$043;     { XMS driver(e.g. HIMEM), returns 80H when installed}
      ScrnSav2=$064;     { ScrnSav2.com tsr }
      NovelIpx=$07A;     { Novel IPX drivers }
      Append  =$0B7;

VAR SubstExpand : BOOLEAN;

{ Get serial number for drive Drv (0=current, 1=A, 2=B etc)}
FUNCTION GetSerial(Drv : LONGINT) : LONGINT;

{ Get volume label for drive Drv (0=current, 1=A, 2=B etc)}
FUNCTION GetVolume(Drv : LONGINT) : String;

{ Get the name of the filesystem for drive Drv (0=current, 1=A, 2=B etc)
 returns FAT12, FAT16 or FAT32 (dos 7.1+), maybe NTFS under NT?}
FUNCTION GetFileSystem(Drv : LONGINT) : String;

{What kind of drive is drive Drv
 Both upper as lower case characters are allowed

Returns
  dtError     = 0;  Drive physically isn't available, or no disk loaded.
  dtRemote    = 1;  Remote (network, Ramdrive) disk drive
  dtFixed     = 2;  Fixed (hard) disk drive
  dtRemovable = 3;  Removable (floppy) disk drive
  dtSubsted   = 4;  Substed drive
  dtCdrom     = 5;  Cdrom}
FUNCTION DriveType(Drv:LONGINT):LONGINT;

{Returns number of floppy drives}
FUNCTION NrFloppies:LONGINT;

{Returns number of logical drives. 32 on my system?!
This used to work before Win95!}
FUNCTION NrDrives:LONGINT;

{Nr of fixed disks in a system}
FUNCTION NrFixedDisks:LONGINT;

{Returns highest valid logical drive}
FUNCTION LastDrv:LONGINT;

{ Clustersize detection. LFN aware. If drv=CDROM then returns 2048.}
FUNCTION ClusterSize(Drv:LONGINT):LONGINT;

{This procedure returns the truename of a path, more like FExpand.
 Normally FExpand is enough, but TRUENAME returns the real path of
 substed drives(if substexpand=TRUE). I use that feature sometimes.}
PROCEDURE TRUENAME(VAR Long : String);

{ Only copy string when lfnsupport:=FALSE}

{ TrueName returning always 8.3 notation. Usable to create a short name from
a longfile to pass to old dos-programs
Expands substed drives(if substexpand=TRUE) }
PROCEDURE GetShortPathName (CONST Long: String; VAR Short: String);

{ TrueName returning always lfn whn existant. Usable to create a short name from
a longfile to pass to old dos-programs
Expands substed drives(if substexpand=TRUE) }
PROCEDURE GetLongPathName (Short: String; VAR Long: String);

{ Generate a short filename for a long filename }
PROCEDURE GenerateShortName(VAR Long: String; VAR Short: String);

{Returns Windowsversion in BCD (hi-byte=major version, lo-byte is minor version)}
FUNCTION WinVer:WORD;

{Try to find some info about the swapfile of Win95}
PROCEDURE GetSwapData(VAR Swp :TSwapInfo);

{Read 64-bytes from CMOS, convert BCD to binary values if necessary
(first 10 fields + Century field)}
PROCEDURE GetCMOS(VAR Cmos : CmosRec);

{ Calls multiplex, <Tsr> : code Tsr to check existence.
   If code <255 then this code is put into AH, and AL is cleared.
    else calls interrupt with ax=unchanged code Tsr

   Returns 255 if Tsr installed,
           0   if not installed,
           1   if not installed and not ok to install
           >255 if error, use (returnvalue AND 255) to obtain errorcode }
FUNCTION  Installed(Tsr:WORD):WORD;

TYPE  UART   = (uNoUART, uBadUART, u8250, u16450, u16550, u16550a);

{
Test if an UART (processor behind the com-port) exists on port "Port", IOW does this port
exist. Windows sometimes blocks some ports. I can't detect COM1 under Windows.
(The mouse is attached to it)}
FUNCTION TestUART( Port : LONGINT):UART;

{ Returns TRUE if named file is actually a device. (Like CON,NUL, etc).
  Can be used to detect a device as input on commandline, or to verify
  existance of a certain devicedriver.}
FUNCTION IsDevice(CONST Fnamex: String): BOOLEAN;


{$ENDIF}

IMPLEMENTATION

{$IFDEF go32v2}
Uses Go32;

Type DiskInfo = packed Record                      { Work area }
          Info_Level  : Word;                      { Always zero }
          Serial_Num  : LongInt;                   { Serial number of the specified disk }
          Volume_Name : ARRAY[0..10] OF CHAR;      { 'NO_NAME' or volume name }
          File_Sys    : ARRAY[0..7] OF CHAR;       { 'FAT12' or 'FAT16' }
     End;

{ Generate a short filename for a long filename }
PROCEDURE GetData(VAR Drv: LONGINT;VAR DI:DiskInfo);

VAR DosRegs : Registers;

BEGIN
 IF LFNSupport AND (tb_size>SIZEOF(DiskInfo)) THEN   { 522 =(260+1)*2 }
  BEGIN
    DosRegs.ds:=tb_segment;
    DosRegs.edx:=tb_offset;
    DosRegs.ax:=$6900;
    DosRegs.bx:=Drv AND 7;
    MsDos(DosRegs);
    copyfromdos(DI,SIZEOF(DiskInfo));
  END;
END;

FUNCTION GetSerial(Drv : LONGINT) : LONGINT;

VAR Di:DiskInfo;

BEGIN
 GetData(Drv,Di);
 GetSerial:=DI.Serial_num;
END;

{FUNCTION GetVolume(Drv : LONGINT) : String;

VAR Di:DiskInfo;
    V : String;

BEGIN
 GetData(Drv,Di);
 Move (DI.Volume_Name,V[1],11);
 V[0]:=#11;
 GetVolume:=V;
END;}

FUNCTION GetVolume(Drv:LONGINT):String;
{ Works on more drive-kinds}
VAR
 Dirinfo:Searchrec;
 Vol,Result:String;
 I:Integer;
BEGIN
 IF Drv=0 THEN
  BEGIN
   GetDir(Drv,Vol);
   Drv:=ORD(Vol[1])-64;
  END;
 vol:=''; result:='';
 findfirst(CHR(drv+64)+':\*.*', $08, dirinfo);
 if doserror=0 then
  vol:=dirinfo.name
 ELSE
  Vol[0]:=#0;
 for i:=1 to length(vol) do if vol[i]<>'.' then result:=result+vol[i];
 while length(result)<11 do result:=result+' ';
 GetVolume:=result;
END;

FUNCTION GetFileSystem(Drv : LONGINT) : String;

VAR Di:DiskInfo;
    V : String;

BEGIN
 GetData(Drv,Di);
 Move (DI.File_Sys,V[1],8);
 V[0]:=#8;
 GetFileSystem:=V;
END;

{$ASMMODE INTEL}
{What kind of drive is drive 'Letter'?
 Both upper as lower case characters are allowed
Returns
  dtError     = 0;  Drive physically isn't available
  dtRemote    = 1;  Remote (network) disk drive
  dtFixed     = 2;  Fixed (hard) disk drive
  dtRemovable = 3;  Removable (floppy) disk drive
  dtSubsted   = 4;  Substed drive}
FUNCTION DTI(Drv:INTEGER):INTEGER; assembler;
asm;
                                mov     cx,word Drv { Get parameter}
                                dec     cx
                                cmp     cl,26
                                ja      @NotValid

                                mov     bx,word Drv
                                mov     ax,$4409  {Get drive parameters}
                                int     $21
                                jc     @NotValid
                 @DrvType2:     cmp    al,1
                                jne    @DrvType3
                               { mov al,1 al is already 1} {dtRemote}
                                jmp    @DrvType5
                 @DrvType3:     and    dh,$090
                                cmp    dh,$080
                                jne    @DrvType6
                                mov al,4           {dtSubsted}
                                jmp @DrvType5
                 @DrvType6:     mov ax,$4408
                                int 21H
                                cmp al,1
                                jne @DrvType4
                                mov al,2
                                jmp @DrvType5
                 @DrvType4:     mov al,3          {dtRemovable}
                 @DrvType5:     xor ah,ah
                                jmp @FullEnd
                 @NotValid:    xor ax,ax
                 @FullEnd:
end;

{ Is drive 'Letter' a cdrom? both upper as lower case characters are allowed}
FUNCTION IsCdRom(Drv:WORD):BOOLEAN; assembler;
asm;
                                mov     cx,Drv { Get parameter}
                                dec     cx
                                cmp     cl,26
                                ja      @NotValid
                                mov     ax,$150B
                                xor     ch,ch
                                int     2FH
                                cmp     bx,$0ADAD
                                je      @fullend
                                mov     al,1
                   @NotValid:  xor      al,al
                   @FullEnd:

END['EAX','ECX','EBX'];


FUNCTION IsDevice(CONST Fnamex: String): boolean;

VAR Regs: Registers;
    F   : FILE;
    FH  : word ABSOLUTE F;

BEGIN
      IsDevice := FALSE;
      Assign (F, Fnamex);
      reset(F, 1);
      IF IOresult <> 0 THEN exit;
      WITH Regs DO
       BEGIN { -- Get information about file: }
       AX := $4400;
       BX := FH;
       MsDos(Regs);
       IF NOT odd(Flags) AND (DL AND $80 <> 0) THEN IsDevice := TRUE
      END;
      close(F);
      IF IOresult <> 0 THEN ;
END;

FUNCTION NrDrives:LONGINT; ASSEMBLER;

ASM
                          mov           ah,$19
                          int           $21
                          and           eax,255
                          mov           edx,eax
                          mov           ah,$0E
                          int           $21
                          and           eax,255
END['EAX','EDX'];

FUNCTION LastDrv:LONGINT; ASSEMBLER;
ASM
                           mov          ah, $19
                           int          $21
                           push         ax        { save default drive }
                           mov          ah, $0E
                           mov          dl, $19
                           int          $21
                           mov          cl, al
                           dec          cx
       @ChkDrv:
                           mov          ah, $0E    { check if drive valid }
                           mov          dl, cl
                           int          $21
                           mov          ah, $19
                           int          $21
                           cmp          cl, al
                           je           @Valid
                           dec          cl         { check next lowest drive number }
                           jmp          @ChkDrv
       @Valid:            pop          dx
                           mov          ah, $0E
                           int          $21        { restore default drive }
                           and          ecx,255

                           mov          eax, ecx
END['EAX','ECX','EDX'];

{$ASMMODE ATT}

FUNCTION NrFloppies:LONGINT;

VAR  Tmp:LONGINT;

BEGIN
 Tmp:=0;
 DosMemGet($40,$10,Tmp,1);
 Tmp:=Tmp SHR 6;
 NrFloppies:=Tmp;
END;


FUNCTION NrFixedDisks:LONGINT;

VAR  Tmp:LONGINT;

BEGIN
 Tmp:=0;
 DosMemGet($40,$75,Tmp,1);
 NrFixedDisks:=Tmp;
END;

{DTI and then CDROM if necessary}
FUNCTION DriveType(Drv:LONGINT):LONGINT;

VAR I : LONGINT;

BEGIN
 CASE Drv OF
   1 :  BEGIN
         IF DiskFree(Drv)=-1 THEN  {A:, Disk Loaded?}
          I:=0
         ELSE
          I:=DTI(Drv);            {Something exists, detect it}
         END;
   2 : BEGIN
        IF NrFloppies>1 THEN      {B: Two drives in BIOS?}
         BEGIN
          IF DiskFree(Drv)=-1 THEN  {Yes, disk loaded?}
           I:=0                     {No disk}
          ELSE
           I:=DTI(Drv);            {Something exists on B:, detect it}
         END
        ELSE
         I:=0;                    {Only one disk. Avoid renaming}
       END
  ELSE {Case}
   BEGIN
    I:=DTI(Drv);                { C...Z, do normal detection}
    IF IsCdRom(Drv) THEN        { Cdrom?}
     I:=5;
    IF (I>0) THEN
     IF DiskFree(Drv) =-1 THEN
      I:=0;
   END;
  END;
  DriveType:=I;
END;

PROCEDURE TRUENAME(VAR Long : String);

VAR
    DosRegs: Registers;


BEGIN
   Long[ORD(Long[0])+1]:=CHR(0);             {Zero terminates}
   CopyToDos(Long[1],ORD(Long[0])+1);        {Omit length}
   IF LfnSupport THEN
    BEGIN
     DosRegs.ax:=$7160;
     IF SubstExpand THEN
      DosRegs.cx:=0                          {Expand substes}
     ELSE
      DosRegs.cx:=$8000;
    END
   ELSE
    DosRegs.ax:=$6000;
   DosRegs.esi:=tb_offset;
   DosRegs.ds:=tb_segment;
   DosRegs.edi:=tb_offset;
   DosRegs.es:=tb_segment;
   MsDos(DosRegs);
   CopyFromDos(Long[1],255);
   Long[0]:=CHR(StrLen(@Long[1]));
   if (dosregs.flags and carryflag) <> 0 then
    doserror:=dosregs.ax
   else
    doserror:=0;
END;

FUNCTION ClusterSize(Drv:LONGINT):LONGINT;

TYPE FreeRecord=RECORD
                  SizRec        : WORD; { Size of this record}
                  StrucVersion  : WORD; { Structure version}
                  SecPerClus    : CARDINAL;
                  BytePerSec    : CARDINAL;
                  AvailClus     : CARDINAL;
                  TotClus       : CARDINAL;
                  PhysSectAvail : CARDINAL;
                  PhysSectTot   : CARDINAL;
                  AllocAvail    : CARDINAL;
                  AllocTot      : CARDINAL;
                  Res1,
                  Res2          : CARDINAL;
                 END;

CONST SecBuf=20;

VAR R : Registers;
    O : LONGINT;
    S : ARRAY[0..5] OF CHAR;
    S2 : String;
    ClusRec : FreeRecord;

BEGIN
 IF LfnSupport THEN
  BEGIN
    IF Drv=0 THEN
     BEGIN
      GetDir(Drv,S2);
      Drv:=ORD(S2[1])-64;
     END;

    S:='C:\'#0;
    S[0]:=CHR(Drv+64);
    CopyToDos(S,8);
    R.AX:=$7303;
    R.CX:=SIZEOF(FreeRecord);
    R.esi:=tb_offset;
    R.ds:=tb_segment;
    R.edi:=tb_offset+SecBuf;
    R.es:=tb_segment;
    MsDos(R);
    seg_move(dosmemselector,transfer_buffer+SecBuf,get_ds,longint(@ClusRec),SIZEOF(FreeRecord));
    O:=ClusRec.BytePerSec*ClusRec.SecPerClus;
    if (R.flags and carryflag) <> 0 then
     doserror:=R.ax
   else
     doserror:=0;

  END
  ELSE
   BEGIN
    R.AX:=$3600;
    R.DX:=Drv;
    MsDos(R);
    IF R.AX=$FFFF THEN
     O:=0
    ELSE
     O:=R.AX
   END;
  IF IsCdRom(Drv) THEN         {Exception, cdroms are always 2048 anyway}
   O:=2048;
  ClusterSize:=O;
END;

PROCEDURE DoInt7160(CONST StrIn: String; VAR StrOut: String;Func:WORD);

VAR DosRegs : Registers;

BEGIN
 IF LFNSupport AND (tb_size>262) THEN
  BEGIN
   PChar(@StrIn[ORD(StrIn[0])+1])[0]:=#0;         {Zero terminates}
   CopyToDos(StrIn[1],ORD(StrIn[0])+1);           {Don't copy length, do copy #0}
   DosRegs.ax:=$7160;
   IF NOT SubstExpand THEN
    INC(Func,$8000);
   DosRegs.cx:=Func;
   DosRegs.esi:=tb_offset;
   DosRegs.ds:=tb_segment;
   DosRegs.edi:=tb_offset;
   DosRegs.es:=tb_segment;
   MsDos(DosRegs);
   CopyFromDos(StrOut[1],255);
   StrOut[255]:=#0;
   StrOut[0]:=CHR(StrLen(@StrOut[1]));
   if (dosregs.flags and carryflag) <> 0 then
    doserror:=dosregs.ax
   else
   doserror:=0;
  END
 ELSE
  StrOut:=StrIn; {Not supported}
END;

{ Get the short path (8.3) of a file or directory, TRUENAME returning 8.3 only }
PROCEDURE GetShortPathName (CONST Long: String; VAR Short: String);

BEGIN
 DoInt7160(Long,Short,1);
END;


{ Get the long path of a file or directory TRUENAME returning lfn when possible}
PROCEDURE GetLongPathName (Short: String; VAR Long: String);

BEGIN
 DoInt7160(Short,Long,2);
END;

{$ASMMODE ATT}
FUNCTION WinVer:WORD; ASSEMBLER;

ASM
             mov  $5642,%ax
             int  $47
             test %al,%al
             je   .LWinVerOk
             xor  %eax,%eax
             jmp  .LWinVerEnd
.LWinVerOk:  mov  %ebx,%eax
.LWinVerEnd:
END['EAX','EBX','ECX'];

PROCEDURE GetSwapData(VAR Swp :TSwapInfo);

VAR R : Registers;

BEGIN
 IF LFNSupport THEN
  BEGIN
   R.AX:=$440D;
   R.CX:=$086E;
   R.edx:=tb_offset;
   R.ds:=tb_segment;
   R.BX:=3;
   R.di:=0;
   R.si:=0;
   MsDos(R);
   if (R.flags and carryflag) <> 0 then
    BEGIN
     doserror:=R.ax;
     FillChar(Swp,SIZEOF(TSwapInfo),0);
    END
   else
    BEGIN
     doserror:=0;
     CopyFromDos(Swp.FileName[1],255);
     Swp.Filename[0]:=CHR(StrLen(@Swp.Filename[1]));
     Swp.FileSize:=R.CX SHL 28 + R.BX SHL 12;
     Swp.Pager:=PagerType(R.AX);
    END;
  END;
END;

{ Generate a short filename for a long filename }
PROCEDURE GenerateShortName(VAR Long: String; VAR Short: String);

VAR A : LONGINT;
    dosRegs : Registers;

BEGIN
 IF LFNSupport AND (tb_size>261) THEN
  BEGIN
   Long[ORD(Long[0])+1]:=CHR(0);             {Zero terminate}
   CopyToDos(Long[1],ORD(Long[0])+1);        {Omit length, include #0}
   DosRegs.ax:=$71A8;
   DosRegs.dx:=$11;
   DosRegs.esi:=tb_offset;
   DosRegs.ds:=tb_segment;
   DosRegs.edi:=tb_offset;
   DosRegs.es:=tb_segment;
   MsDos(DosRegs);
   { 8.3 version (DH=$1) is buggy. Using FCB-method, and convert back to 8.3
    Thanks to Thomas Schatzl, who found out that the 8.3 version was buggy
    under the extender}
   seg_move(dosmemselector,transfer_buffer,get_ds,longint(@short[1]),8);
   A:=8;
   WHILE (Short[A]=#32) AND (A>1) DO
    DEC(A);
   Short[A+1]:='.';
   seg_move(dosmemselector,transfer_buffer+A,get_ds,longint(@short[A+2]),3);
   INC(A,4);
   WHILE (Short[A]=#32) AND (A>1) DO
    DEC(A);
   Short[0]:=CHR(A);
    if (dosregs.flags and carryflag) <> 0 then
  doserror:=dosregs.ax
 else
  doserror:=0;
  END
 ELSE
  Short:=Long; {Not supported}
END;

PROCEDURE DoCmos; ASSEMBLER;            {INTERNAL, register parameters}
{ Reads ECX bytes from CMOS, starting with value %ESI (%ESI is updated with
  bytes read) if %ebx=0, the read values are converted from BCD to binary}

ASM
.LNextClock:   mov $112,%edx
               mov %esi,%eax
               outb %al,%dx
               inc %edx
               xor  %eax,%eax
               inb %dx,%al                       // Read a byte
               test $3,%ebx
               jne   .LNoBCD
               push  %eax
               mov   %eax,%edx
               shr   $4,%eax
               imul  $10,%eax
               pop   %edx
               and  $15,%edx
               addl %edx,%eax                   // Convert BCD to binary
.LNoBCD:       stosb                            // Store
               inc   %esi
               dec   %ecx
               jnz   .LNextClock
END['EAX','ECX','EDX','ESI','EDI']; {EBX is tested but not changed}

PROCEDURE GetCMOS(VAR Cmos : CmosRec); ASSEMBLER;

ASM
               mov Cmos,%edi
               mov $112,%edx                     // Read RTC-B register
               mov $11,%eax
               outb %al,%dx
               inc %edx
               inb %dx,%al                       // Read RTC-B byte
               mov $1,%ebx
               test $4,%eax                      // test bit 2
               jne   .LBCD
               xor %ebx,%ebx                      // Do I have to convert from BCD?
.LBCD:
               xor %esi,%esi                    // Number of byte read.
               mov $10,%ecx                     // Read 10 bytes clock values
               and  $1,%ebx                      // these 10 bytes should be
                                                // converted from BCD if RTC says so.
               call DoCmos
               or  $2,%ebx
               mov $40,%ecx
               call DoCmos                      // Next 40 bytes no BCDconv
               and  $1,%ebx
               mov  $1,%ecx
               call DoCmos                      // 1 byte with BCDconv. (century)
               or  $2,%ebx
               mov $13,%ecx
               call DoCmos                      // 13 bytes no BCDconv
END['EAX','EBX','ECX','EDX','ESI','EDI'];

FUNCTION  Installed(Tsr:WORD):WORD; ASSEMBLER;
ASM
                            push   %ebp
                            push   %ds
                            push   %es
                            mov    Tsr,%ax
                            test   %ah,%ah
                            jne    .LBigCode
                            mov    %al,%ah
                            xor    %al,%al
.LBigCode:                  cmp    $22,%ah
                            je     .LWinTest
                            xor    %ebx,%ebx
                            int    $47
                            js     .LInstFout
.LInstGoed:                 xor    %ah,%ah
                            jmp     .LInstEnd
.LInstFout:                 mov    $-1,%ah
                            jmp     .LInstEnd
.LInstWin2x:                mov    $2,%ax
                            jmp     .LInstEnd
.LWinTest:                  int    $47
                            js     .LInstFout
                            cmp    $128,%al
                            jne    .LInstVerder
                            mov    $0,%al
.LInstVerder:               or     %al,%al
                            je     .LInstGoed
                            cmp    $255,%al
                            je     .LInstWin2x
                            xchg   %al,%ah

.LInstEnd:                  pop    %es             // Multiplexer can change a lot!
                            pop    %ds
                            pop    %ebp
END['EAX','EBX','ECX','EDX','ESI','EDI'];



FUNCTION TestUART( Port : LONGINT):UART; ASSEMBLER;

{          uNoUART    =0;         uBadUART   =1
          u8250      =2;         u16450     =3
          u16550     =4;         u16550a    =5

  MCR             = 4;           MSR             = 6;
  Scratch         = 7;           FCR             = 2;
  IIR             = 2;           LOOPBIT         = 010H
}

ASM;
                        mov    Port,%edx
                        add    $4,%edx
                        in     %dx,%al
                        mov    %dx,%cx
                        mov    %eax,%esi
                        or     $16,%eax
                        outb   %al,%dx
                        incl   %edx
                        incl   %edx
                        in     %dx,%al
                        mov    %eax,%edi
                        mov    %ecx,%edx
                        mov    $26,%eax
                        out    %al,%dx
                        inc    %edx
                        inc    %edx
                        inb    %dx,%al
                        and    $240,%eax
                        mov    %eax,%ebx
                        mov    %edi,%eax
                        out    %al,%dx
                        mov    %esi,%eax
                        out    %al,%dx
                        and    $249,%eax
                        mov    %ecx,%edx
                        out    %al,%dx
                        xor    %ecx,%ecx
                        cmp    $144,%ebx
                        jne    .LCollectUART
                        incl    %edx
                        incl    %edx
                        incl    %edx
                        mov    $170,%eax
                        out    %al,%dx
                        in     %dx,%al
                        mov    $2,%ecx
                        cmp    $170,%eax
                        jne    .LCollectUART
                        sub    $5,%edx
                        mov    $1,%eax
                        out    %al,%dx
                        in     %dx,%al
                        and    $192,%eax
                        mov    $5,%ecx
                        cmp    $192,%eax
                        je     .LCollectUART2
                        decl   %ecx
                        cmp    $128,%eax
                        je     .LCollectUART2
                        decl   %ecx
                        test   %eax,%eax
                        je     .LCollectUART2
                        dec    %ecx
                        dec    %ecx
.LCollectUART2:         xor    %eax,%eax
                        out    %al,%dx

.LCollectUART:          mov    %ecx,%eax
END ['EAX','EBX','ECX','EDX','ESI','EDI'];



{$ENDIF}


BEGIN
 {$IFDEF Go32V2}
 SubstExpand:=FALSE;
 {$ENDIF}
End.

