UNIT EFIO;
{  Some routines used while handling files.
   This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)   }

INTERFACE
{$DEFINE USEASM}    {Standard header. There is no assembler in this unit (yet)}

{$I syseq.inc}		  { Some conditionals only, no actual code}

Uses Dos;

{$IFDEF Turbo}
 TYPE CARDINAL=LONGINT;
{$ENDIF}

TYPE
      {ARC, PAK, A7P, HYP are unimplemented. (I have some routines for these
			     packers, but I have to decrypt the assembler
      DWC		 No algoritm present.}

   ArchiveType = ( none,SQZ, ZIP, HPK, ZOO, LZH,
		   ARJ, DWC, ARC, PAK, A7P, HYP,
		   RAR,   Q, UC2, Gif, LBM, PCX,
           WAV, BMP, MP3);                  {MP3 not ready yet, see help}


{$IFNDEF Turbo}
FUNCTION ArchiveMethod( FileName : String) : ArchiveType;
{$ENDIF}

{ Boolean function that returns True if the file exists;otherwise,
 it returns False. Closes the file if it exists, copied from the
 BP7 help.
 Seems not to work for a directory. }
FUNCTION FileExists(FileName: String): Boolean;

{ Output hexrepresentation for a WORD or CARDINAL, overloaded versions for
output to texttyped file,
the WrHex procedures always output 4 chars, the WrLngHex always 8, both pad
with character zero.}

PROCEDURE WrHex     (InValue:WORD);
PROCEDURE WrLngHex  (InValue:CARDINAL);

{$IFDEF Turbo} {Overloading fails in BP}
 PROCEDURE WrHexFile  (VAR F : Text;InValue:WORD);
 PROCEDURE WrLngHexFile  (VAR F:Text;InValue:CARDINAL);
{$ELSE}
 PROCEDURE WrHex      (VAR F : Text;InValue:WORD);
 PROCEDURE WrLngHex  (VAR F:Text;InValue:CARDINAL);
{$ENDIF}

{Same for Octal. WrOct outputs 6 digits, WrLngOct 11}
PROCEDURE WrOct     (InValue:WORD);
PROCEDURE WrLngOct  (InValue:CARDINAL);

{$IFDEF Turbo}
 PROCEDURE WrOctFile     (VAR F : Text;InValue:WORD);
 PROCEDURE WrLngOctFile  (VAR F:Text;InValue:CARDINAL);
{$ELSE}
 PROCEDURE WrOct     (VAR F : Text;InValue:WORD);
 PROCEDURE WrLngOct  (VAR F:Text;InValue:CARDINAL);
{$ENDIF}

{Same for Binary. WrBinary outputs 16 digits, WrLngBinary 32}
PROCEDURE WrBinary    (InValue:WORD);
PROCEDURE WrLngBinary (InValue:CARDINAL);

{$IFDEF Turbo}
 PROCEDURE WrBinaryFile     (VAR F : Text;InValue:WORD);
 PROCEDURE WrLngBinaryFile  (VAR F:Text;InValue:CARDINAL);
{$ELSE}
 PROCEDURE WrBinary     (VAR F : Text;InValue:WORD);
 PROCEDURE WrLngBinary  (VAR F:Text;InValue:CARDINAL);
{$ENDIF}

{ WrStrAdj(S,L) Is the same as write(S:L), however WrStrAdj(S:-L) pads
on the other side}
PROCEDURE WrStrAdj(CONST InS: String;L : LONGINT);
PROCEDURE WrStrAdj(VAR F: Text;CONST InS: String;L : LONGINT);

{ Operations on PATHS,
position of extension in string}
FUNCTION ExtensionPos ( CONST s : String) : WORD;

{Add extension (only if extension does not exists)}
PROCEDURE AddExtension ( VAR s : String; CONST Extension : String) ;

{Assign (F,S); Append(F); which also works if file doesn't exist}
PROCEDURE FileAppend(VAR F :Text; CONST FileName:String);

{Remove an extension}
PROCEDURE RemoveExtension ( VAR s : String) ;

{Remove, then Add a new extension}
PROCEDURE ChangeExtension ( VAR s : String; CONST Extension : String) ;

{Caution! Dangerous command, specially under Linux as ROOT Deltree alike (rm -rf
 if it runs under Linux) Deletes all files in given directory. Including hidden!
Runtime error when it encounters files locked by Win95 (and NT I presume)}
PROCEDURE Deldir(Dir : PathStr);

{Touch command (Set filetimedate to current datetime). Directories and
wildcards supported. (e.g. Touch ('c:\windows\*.*'); )}
PROCEDURE Touch(Const FileName:String);

{Create a directory, but also works for 'd:\prog\ts\pp' if d:\prog
    doesn't exist). The drive(if specified) must exist though. LFN working if
    RTL is LFN (=standard on Go32V2)}
PROCEDURE MkFullDir(CONST InPath:String);

{For use with Msg2Inc, or Data2Inc. Suppose you have a string constant >255
 chars. You can't declare it normally in FPC, so I put them in a file to be
 compiled with MSG2INC or Data2Inc to a constant array of char. This procedure
 does nothing but write the array of char designated by Data to screen while
 issuing a writeln for each #13 or #0. Call with WrArrChar(@data,SIZEOF(Data));
 (This has been fixed in 0.99.12+)
 }
PROCEDURE WrArrChar(Data:PCHAR; sizedata: LONGINT);

{Reads a variable size (version independant) record from an untyped file. Simply
a record with a longint for size prefixed and possible extra data AFTER the known
fields, which may be ignored. Idea copied from Folkert Wijnstra's FMAIL
structures.
Returns #valid bytes read.}
FUNCTION ReadVarRecord(VAR F:File;VAR Rec;SizeRec:LONGINT): LONGINT;

{Blockwrites Rec, but prepends 4 bytes size of rec (in bytes)}
FUNCTION WriteVarRecord(VAR F:File;VAR Rec;SizeRec:LONGINT): LONGINT;

IMPLEMENTATION

Uses EPasStr;

FUNCTION FileExists(FileName: String): Boolean;
{ Boolean function that returns True if the file exists;otherwise,
 it returns False. Closes the file if it exists, copied from the
 BP7 help. }
var
 F: file;
begin
 {$I-}
 Assign(F, FileName);
 FileMode := 0;  {Set file access to read only }
 Reset(F);
 Close(F);
 {$I+}
 FileMode:=2;
 FileExists := (IOResult = 0) and (FileName <> '');
end;  { FileExists }

PROCEDURE FileAppend(VAR F :text;CONST FileName:String);

BEGIN
 Assign (F,FileName);
 {$I-}
 Append(F);
 {$I+}
 IF IOResult=6 THEN
  BEGIN
   Close(F);
   Assign(F,FileName);
   Rewrite(F);
  END;
END;

CONST HexTab : String[16] ='0123456789ABCDEF';

PROCEDURE WrHex(InValue:WORD);

VAR S : String[10];
    A : WORD;

BEGIN
 FOR A :=0 TO 3 DO		      { 4 nibbles in a word}
  BEGIN
   S[4-A]:=HexTab[1+(InValue AND 15)];
   InValue:=InValue SHR 4;
  END;
 S[0]:=CHR(4);
 write(S);
END;

{$IFDEF Turbo} {Overloading fails in BP}
 PROCEDURE WrHexFile  (VAR F : Text;InValue:WORD);
{$ELSE}
 PROCEDURE WrHex      (VAR F : Text;InValue:WORD);
{$ENDIF}

VAR S : String[10];
    A : WORD;

BEGIN
 FOR A :=0 TO 3 DO		      { 4 nibbles in a word}
  BEGIN
   S[4-A]:=HexTab[1+(InValue AND 15)];
   InValue:=InValue SHR 4;
  END;
 S[0]:=CHR(4);
 write(F,S);
END;

PROCEDURE WrLngHex(InValue:CARDINAL);


VAR S : String[10];
    A : WORD;

BEGIN
 FOR A :=0 TO 7 DO		      { 8 nibbles in a CARDINAL}
  BEGIN
   S[8-A]:=HexTab[1+(InValue AND 15)];
   InValue:=InValue SHR 4;
  END;
 S[0]:=CHR(8);
 write(S);
END;


{$IFDEF Turbo} {Overloading fails in BP}
 PROCEDURE WrLngHexFile  (VAR F:Text;InValue:CARDINAL);
{$ELSE}
 PROCEDURE WrLngHex  (VAR F:Text;InValue:CARDINAL);
{$ENDIF}

VAR S : String[10];
    A : WORD;

BEGIN
 FOR A :=0 TO 7 DO		      { 8 nibbles in a CARDINAL}
  BEGIN
   S[8-A]:=HexTab[1+(InValue AND 15)];
   InValue:=InValue SHR 4;
  END;
 S[0]:=CHR(8);
 write(F,S);
END;

{CONST HexTab : String[16] ='0123456789ABCDEF';}

PROCEDURE WrOct(InValue:WORD);

VAR S : String[10];
    A : WORD;

BEGIN
 FOR A :=0 TO 5 DO		      { 16/3=5.3 rounded up digits in a word}
  BEGIN
   S[6-A]:=CHR(48+(InValue AND 7));
   InValue:=InValue SHR 3;
  END;
 S[0]:=CHR(6);
 write(S);
END;

{$IFDEF Turbo}
 PROCEDURE WrOctFile     (VAR F : Text;InValue:WORD);
{$ELSE}
 PROCEDURE WrOct     (VAR F : Text;InValue:WORD);
{$ENDIF}

VAR S : String[10];
    A : WORD;

BEGIN
 FOR A :=0 TO 5 DO		   { 16/3=5.3 rounded up digits in a word}
  BEGIN
   S[6-A]:=CHR(48+(InValue AND 7));
   InValue:=InValue SHR 3;
  END;
 S[0]:=CHR(6);
 write(F,S);
END;

PROCEDURE WrLngOct(InValue:CARDINAL);


VAR S : String[12];
    A : WORD;

BEGIN
 FOR A :=0 TO 10 DO		   { 32/3=10.6, rounded up, digits in a word}
  BEGIN
   S[11-A]:=CHR(48+(InValue AND 7));
   InValue:=InValue SHR 3;
  END;
 S[0]:=CHR(11);
 write(S);
END;

{$IFDEF Turbo}
 PROCEDURE WrLngOctFile  (VAR F:Text;InValue:CARDINAL);
{$ELSE}
 PROCEDURE WrLngOct  (VAR F:Text;InValue:CARDINAL);
{$ENDIF}

VAR S : String[12];
    A : WORD;

BEGIN
 FOR A :=0 TO 10 DO		   { 32/3=10.6, rounded up, digits in a word}
  BEGIN
   S[11-A]:=CHR(48+(InValue AND 7));
   InValue:=InValue SHR 3;
  END;
 S[0]:=CHR(11);
 write(F,S);
END;

PROCEDURE WrBinary(InValue:WORD);

VAR S : String[18];
    A : WORD;

BEGIN
 FOR A :=0 TO 15 DO		      { 16-bit rounded up digits in a word}
  BEGIN
   S[16-A]:=CHR(48+(InValue AND 1));
   InValue:=InValue SHR 1;
  END;
 S[0]:=CHR(16);
 write(S);
END;

{$IFDEF Turbo}
 PROCEDURE WrBinaryFile     (VAR F : Text;InValue:WORD);
{$ELSE}
 PROCEDURE WrBinary     (VAR F : Text;InValue:WORD);
{$ENDIF}

VAR S : String[18];
    A : WORD;

BEGIN
 FOR A :=0 TO 15 DO		      { 16-bit rounded up digits in a word}
  BEGIN
   S[16-A]:=CHR(48+(InValue AND 1));
   InValue:=InValue SHR 1;
  END;
 S[0]:=CHR(16);
 write(F,S);
END;

PROCEDURE WrLngBinary(InValue:CARDINAL);


VAR S : String[33];
    A : WORD;

BEGIN
 FOR A :=0 TO 31 DO		   { 32, rounded up, digits in a word}
  BEGIN
   S[32-A]:=CHR(48+(InValue AND 1));
   InValue:=InValue SHR 1;
  END;
 S[0]:=CHR(32);
 write(S);
END;

{$IFDEF Turbo}
 PROCEDURE WrLngBinaryFile  (VAR F:Text;InValue:CARDINAL);
{$ELSE}
 PROCEDURE WrLngBinary  (VAR F:Text;InValue:CARDINAL);
{$ENDIF}

VAR S : String[33];
    A : WORD;

BEGIN
 FOR A :=0 TO 31 DO		   { 32/3=10.6, rounded up, digits in a word}
  BEGIN
   S[32-A]:=CHR(48+(InValue AND 1));
   InValue:=InValue SHR 1;
  END;
 S[0]:=CHR(32);
 write(F,S);
END;

FUNCTION IAbs(I:LONGINT):LONGINT;

BEGIN
 IF I>=0 THEN
  IAbs:=I
 ELSE
  IAbs:=-I;
END;

PROCEDURE WrStrAdj(VAR F: Text;CONST InS: String;L : LONGINT);
																															       
VAR Len,I : LONGINT;																													   
																																																																																			    
BEGIN																																																										   
 Len:=Length(InS);																																																							
 IF Len>=ABS(L) THEN																																																						      
  Write(F,InS)																																																								   
 ELSE																																																										   
  IF L<0 THEN																																																								    
   BEGIN																																																									 
    Write(F,inS);																																																							 
    FOR I:= 1 TO (ABS(L)-Len) DO Write(F,' ');																																																       
   END																																																										  
 ELSE																																																										   
  BEGIN																																																										 
   FOR I:= 1 TO (ABS(L)-Len) DO Write(F,' ');																																																	
    Write(F,inS);																																																							 
  END;																															
END;																																																										    

PROCEDURE WrStrAdj(CONST InS: String;L : LONGINT);

VAR Len : LONGINT;

BEGIN
 Len:=Length(InS);
 IF Len>=IABS(L) THEN
  Write(InS)
 ELSE
  IF L<0 THEN
   Write(InS,' ':(IABS(L)-Len))
 ELSE
   Write(' ':(L-Len),inS);
END;



{ Test a file for a archivetype, DWC is not implemented yet, but the entry
    in ArchiveType is for compability reasons with TLib (MISC directory),
    A7P is the compression algoritm better known as ARC+ (plus!)
    SQZ HPK ZOO (DWC, see above) PAK A7P and HYP not tested yet, but will
    probably work, I'm not sure how versionindependant the RAR/Q/UC2
    detections are, since I only own one version of those. (Rar appears to be
    working with Rar 2.0x)
}

{$IFNDEF Turbo}
FUNCTION ArchiveMethod( FileName : String) : ArchiveType;

CONST AutoDetect=8;

TYPE WordArrTable=ARRAY[0..AutoDetect] OF ArchiveType;
TYPE CardArrTable=ARRAY[0..AutoDetect] OF CARDINAL;

CONST HdlTable :WordArrTable =(SQZ,ZIP,HPK,RAR,UC2,Q,Gif,LBM,PCX);
      IdentTable : CardArrTable=($51534C48,$04034B50,$4B415048,
				 $21726152,$1A324355,$60005344,
				 $38464947,$4D524F46,$0801050A);
			       { SQZ	 ZIP      HAPK
				 RAR	 UC2       Q,
				 Gif	 LBM       PCX,
				 Wav}


VAR F	    : File;
    CardPtr      : ^CARDINAL;
    WordPtr      : ^WORD;
    Buf	  : ARRAY[0..25] OF BYTE;
    Found	: ArchiveType;
    A	    : WORD;

BEGIN
 IF NOT FileExists(Filename) THEN {Standard FileExists like in BP-help}
  BEGIN
   ArchiveMethod:=None;
   EXIT
  END;
 FileMode:=0;	   {Read only}
 Assign(F,FileName);
 Reset(F,1);
 BlockRead(F,Buf,25,A);
 IF A<>25 THEN
  BEGIN
   ArchiveMethod:=None;
   close(F);
   EXIT;
  END;
 CardPtr:=@Buf;
 WordPtr:=@Buf;
 Found:=None;
 FOR A :=0 TO AutoDetect DO
  IF CardPtr[0]=IdentTable[A] THEN
   Found:=HdlTable[A];		      { SQZ,ZIP,HAPK,RAR,UC2,Q}
 IF WordPtr[0]=$EA60 THEN
  Found:=ARJ;
 IF WordPtr[0]=$4D42 THEN
  Found:=BMP;
 IF CardPtr[2]=$45564157 THEN
  Found:=WAV;
 IF Found= None THEN
  BEGIN
   CardPtr:=@Buf[1];
   IF CardPtr[5]=$FDC4A7DC THEN
    Found:=ZOO
   ELSE
    BEGIN
     IF (WordPtr[1]=$6C2D)  AND (Buf[6]=$2D) THEN
      Found:=LZH
     ELSE
      BEGIN
       IF Buf[0]=$1A THEN {Arc-ish type?}
	BEGIN
	 {Arc and clones detection, (Arc..Hyp), not implemented yet}
	 BlockRead(F,Buf,25,A);
	  IF A<>25 THEN
	   BEGIN
	     Seek(F,FileSize(F)-128);
	     BlockRead(F,Buf,3,A);
	     Buf[4]:=0;
	     CardPtr:=@buf;
	     IF CardPtr^=$474154 THEN
	      Found:=MP3
	     ELSE
	      Found:=NONE;
	   END;
	END;
      END;
    END;
  END;
 Close(F);
 ArchiveMethod:=Found;
END;
{$ENDIF Turbo}


FUNCTION ExtensionPos ( CONST s : String) : WORD;
VAR
  c   : CHAR ;
  l,i : WORD;
BEGIN
  l := Length(s) ;
  IF L>0 THEN
   BEGIN
    i := l ; C:=S[i];
    WHILE (i<>0) AND NOT (c IN ['/','\',':','.']) DO
     BEGIN
      c := s[i] ;
      DEC(i) ;
     END;
   IF (S[I+1]='.') AND (I<>0) THEN
    ExtensionPos:=i+1
   ELSE
    ExtensionPos:=65535;
   END
  ELSE
   ExtensionPos:=65535;
END;

PROCEDURE AddExtension ( VAR s : String; CONST Extension : String) ;
BEGIN
 {$IFDEF Go32V2}
  IF (ExtensionPos(s)=65535) OR LFNSUPPORT THEN
   BEGIN
 {$ENDIF}
    INC(S[0]);
    S[ORD(S[0])]:='.';
    IF (Extension [1]>' ') THEN	{Small safety check}
     s:=S+Extension;
 {$IFDEF Go32V2}
  END;
  {$ENDIF}
END;

PROCEDURE RemoveExtension ( VAR s : String) ;
VAR
  p : WORD;
BEGIN
  p := ExtensionPos(s) ;
  IF p<>65535 THEN
   S[0]:=CHR(P-1);
END;

PROCEDURE ChangeExtension ( VAR s : String; CONST Extension : String) ;
BEGIN
  RemoveExtension(s) ;
  AddExtension(s,Extension) ;
END;


PROCEDURE DelDirSubDirs(CONST Dir : PathStr);
VAR
  FoundDir  : SearchRec; { Current search data. }
  F	 : File;

BEGIN
 FindFirst(Dir+'*.',AnyFile, FoundDir);
 WHILE DosError=0 DO
  BEGIN
      IF (FoundDir.Name[1] <> '.') AND ((FoundDir.Attr AND directory)>0) THEN
	 BEGIN
	 {$ifdef Linux}
	  DeldirSubDirs(Dir+FoundDir.Name+'/');
	 {$else}
	  DeldirSubDirs(Dir+FoundDir.Name+'\');
	 {$endif}
	END;
     FindNext(FoundDir);
   END; {While loop}
 {$IFDEF Closefind}
  Findclose(FoundDir);
 {$ENDIF}
  FindFirst(Dir+'*.*',AnyFile, FoundDir);
  WHILE DosError=0 DO
   BEGIN
      IF ((FoundDir.attr AND Directory)=0) THEN
	BEGIN
	  Assign(F,Dir+FoundDir.Name);
	  SetFAttr(F,FoundDir.Attr and 56);
	  Erase(F);
	 END
	ELSE
	 IF (FoundDir.Name[1]<>'.') THEN
	  RMDIR(Dir+FoundDir.Name);
       FindNext(FoundDir);
    END; {read loop}
 {$IFDEF CloseFind}
  FindClose(FoundDir);
 {$ENDIF}
END;{ SearchSubDirs}

PROCEDURE Deldir(Dir : PathStr);

BEGIN
 {$ifdef FPC}
   AppendBackSlash(Dir); {procedure is safe for linux}
 {$else}
   {Borland doesn't compile under Linux, other compilers not supported :-)}
 IF Dir[ORD(Dir[0])] <> '\' THEN
  Dir:=Concat(Dir,'\');
 {$endif FPK}
 DeldirSubDirs(Dir);
 DEC(Dir[0]);	   {remove backslash/slash, RMDIR will fail otherwise}
 IF Dir[1]<>'.' THEN
  rmdir (Dir);		  { turning IO-checks off would be better I think}
END;

PROCEDURE MkFullDir(CONST InPath:String);
{ Create a directory, but also works for paths like 'd:\prog\ts\pp' if d:\prog
    doesn't exist). The drive(if specified) must exist though.
  Use FExpand to make sure you pass a full path to this procedure. Or add a
  FExpand on the first parameter}

VAR S,
    S2	  : String;
    A,C	  : LONGINT;
    TestRec      : SearchRec;

BEGIN
 S:=InPath;
 {$IFDEF Linux}
  ReplaceChar(S,'/','\');
 {$ELSE}
  ReplaceChar(S,'\','/');
 {$ENDIF}
 S:=FExpand(S);
 A:=3;
 REPEAT
  {$IFDEF Linux}
   A:=NextCharPos(S,'\',A+1);
  {$ELSE}
   A:=NextCharPos(S,'/',A+1);
  {$ENDIF}
  C:=A;
  IF C=65535 THEN
   C:=Length(S[0])+1;
  S2:=Copy(S,1,C-1);
  FindFirst(S2,Directory,TestRec);
  IF DosError<>0 THEN Mkdir(S2);
 UNTIL A=65535;
END;


PROCEDURE Touch(Const FileName:String);

Var  DirE	: SearchRec;
     F	   : Text;
     V1,V2,V3,V4 : Word;
     FTime       : LongInt;
     DT	  : DateTime;
     {$IFDEF Turbo}
     Path	: DirStr;
     Name	: NameStr;
     Ext	 : ExtStr;
     {$ELSE}
     Path,Name,
     Ext	 : String;
     {$ENDIF}

Begin
 FSplit(Filename,Path,Name,Ext);
 {$IFNDEF Turbo}
 IF Length(Path[0])=0 THEN
  AppendBackslash(Path);
 {$ENDIF}
 GetTime(V1,V2,V3,V4);
 DT.Hour:=V1;  DT.Min:=V2;      DT.Sec:=V3;
 GetDate(V1,V2,V3,V4);
 DT.Year:=V1;  DT.Month:=V2;    DT.Day:=V3;
 PackTime(DT,FTime);
 FindFirst(FileName,anyfile,DirE);
 WHILE DosError=0 DO
  BEGIN
   IF Length(Path[0])=0 THEN
    Assign(F,DirE.Name)
   ELSE
    BEGIN
     Name:=Path+DirE.Name;
     Assign(F,Name);
    END;
   Reset(F);			{Open read-write}
   SetFTime(F,FTime);
   Close(F);
   FindNext(DirE);
 End;
 {$IFDEF CloseFind}
  FindClose(DirE);
 {$ENDIF}
END;

{---------}

PROCEDURE WrArrChar(Data : PChar; sizedata: LONGINT);

VAR I : LONGINT;

BEGIN
 IF SizeData>0 THEN
 FOR I :=0 TO SizeData-1 DO
  IF (Data[I]=#13) OR (Data[I]=#0) THEN
   WriteLn
  ELSE
   Write(Data[I]);
END;

FUNCTION ReadVarRecord(VAR F:File;VAR Rec;SizeRec:LONGINT): LONGINT;
{Reads a variable size (version independant) record from an untyped file. Simply
a record with a longint for size prefixed and possible extra data AFTER the known
fields, which may be ignored. Idea copied from Folkert Wijnstra's FMAIL
structures.
Returns no valid bytes read.}

CONST ReadBufSize=400;

VAR Buffer : ARRAY[0..ReadBufSize-1] OF BYTE;
    I,J,K  : LONGINT;

BEGIN
 BlockRead(F,I,4,J);         {Read size on disk}
 IF J<>4 THEN
  EXIT(0);
 FillChar(Rec,0,SizeRec);
 IF I<=SizeRec THEN
  BEGIN
   BlockRead(F,Rec,I,K);            {Current version, or older. Newer fields are zero}
   IF I<>K THEN
    EXIT(0);
   ReadVarRecord:=K;
  END
 ELSE
  BEGIN
   {Record is newer version. Read valid fields, and ignore the others (but
   read them to a dump buffer to advance the filepos)}
   BlockRead(F,Rec,sizerec,K);
   IF K<>SizeRec THEN
    EXIT(0);
   DEC(I,sizerec);          {Calc bytes left to read}
   IF (I DIV ReadBufSize)>0 THEN
    FOR J:=1 TO I DIV ReadBufSize DO
     BEGIN
      BlockRead(F,Buffer,ReadBufSize ,K);
      IF K<>ReadBufSize THEN
       EXIT(0); {Should have been more bytes, no valid bytes}
     END;
   I:=I MOD 400;
   IF I<>0 THEN
    BEGIN
     BlockRead(F,Buffer,I,K);
     IF I<>J THEN
      EXIT(0);
    END;
   ReadVarRecord:=SizeRec;
  END;
END;


FUNCTION WriteVarRecord(VAR F:File;VAR Rec;SizeRec:LONGINT): LONGINT;

VAR J:LONGINT;

BEGIN
 BlockWrite(F,SizeRec,4,J);
 IF J<>4 THEN
   EXIT(0);
 BlockWrite(F,Rec,SizeRec,J);
 IF SizeRec<>J THEN
  EXIT(0);
 WriteVarRecord:=SizeRec;
END;

END.

