UNIT EStrings;
{Extra Strings (actually, PChar's) routines, not found in Strings.

   This is a unit of the XTDLIB/FPK package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Free for non-commercial use.

 Some are basic (like the LTrim, RTrim, Kill* procedure and the simpler
  CharPos procedures). Some are really weird, like StripDoubleChar and
  GetBetween(the last one not ported yet)

 If you wonder what all these weird string routines are for; I used to
  maintain an util which scanned several logfiles. These logfiles used
  certain characters (like +-!@#$%^&*()) to indicate a certain event or type
  of logline. Using these characters sped up the process, (a CharPos is
  faster than a Pos) and also made it
  more version independant (because it didn't rely on fixed positions, but
  on positions relative to a certain signal-character).

}
INTERFACE


{$define useasm}                {Default : Use assembler, comment this off to
                                         compile pascal code}
{$I syseq.inc}                  { defines processor used for XTDLIB}

{$IFDEF FPC}
{$ifdef Useasm}
{$Info EStrings: Using assembler routines when possible }
{$else}
{$Info EStrings: Using All-Pascal routines }
{$endif Useasm}
{$ENDIF}
{$IFDEF Turbo}
 TYPE CARDINAL=LONGINT;
{$ENDIF}

TYPE CHARSET = SET OF CHAR;

{ Strip character from the left of the string}
PROCEDURE PLTrim (P : PChar;Ch:CHAR);

{ Strip character from the right of the string}
PROCEDURE PRTrim (P : PChar;Ch:CHAR);

{ Same as LTrim, but for a whole set of chars. PKillChar(B,['A','B']) turns
   'ABABABABFlup' into 'Flup'. Slow (relative to LTrim) though.
 About 2 LTrims for a KillPChar I estimate}
PROCEDURE PKillChar(P : PChar;CONST CSet:CHARSET);

{ KillChar at the right of the string. KillBPChar(B,['A','B']) turns
   'FLUPABABABAB' into 'Flup'.}
PROCEDURE PKillBChar(P:PChar;const CSet:CHARSET);

{ Appends a backslash to the PChar, if it doesn't already have one.
   Primitive for dirscanning structures.
   Contains some conditional code to change a \ to a / for Linux}
PROCEDURE PAppendBackslash(P:PChar);

(* Replace in string S the character "ReplaceMe" with "RepWith".*)
PROCEDURE PReplaceChar(P : PChar;ReplaceMe,RepWith:CHAR);

{Remove C entirely from P, removing #0 will leave string untouched}
PROCEDURE PStripChar(P:PChar;C:CHAR);

{StripChar for a CHARSET; Remove any char in C entirely from P}
PROCEDURE PKillChrTot(P:PChar;CONST CSet:CHARSET);

{ Pos for one char only. Faster than an ordinary Pos, -1 when not found.

I yesterday found out that the Pos function (at least FPK's) supports
character-only-Pos by procedure overloading. That one isn't assembler, this
one is however}
FUNCTION PCharPos(P:PChar;Ch:Char):WORD;

{ Like CharPos but starts searching on character number Count (zero based),
    returnvalue is zero based and absolute (to beginning of string)}
FUNCTION PNextCharPos(P:PChar;Ch:Char;Count:WORD):WORD;

{ CharPos, but starting on the right of the string, the index is 0 based
    however}
FUNCTION PRCharPos(P:PChar;Ch:Char):WORD;

{ Next variant (starts searching on a different point) of RCharPos}
FUNCTION PNextRCharPos(P:PChar;Ch:Char;Count:WORD):WORD;

{ A bit more complex CharPos, search first occurance of character in set C}
FUNCTION PCharPosSet(P:PChar;CONST CSet :CHARSET):WORD;

{ Cleans a string of double(of more) sequences of char C.
 e.g. ' 1 2  3   4    5' becomes ' 1 2 3 4 5' (Stripping for space).

Used to make mail from Fido and Newsgroup newbies readable (with
a StripDoubleChar for '.', '!' and space) }
PROCEDURE PStripDoubleChar(P:PChar;Ch:Char);

{Like Delete for pascal-strings, but for PChars}
PROCEDURE StrDelete(P:PChar;Position,Count:WORD);

{All (normal, not international ones) characters uppercase}
PROCEDURE PUpperCase(P:PChar);

{All (normal, not international ones) characters lowercase}
PROCEDURE PLowerCase(P:PChar);

{ Get first "bits" bits or digits from S (like S:='0101010'),
and return their binary value.
Allowable range PStrToBinary parameter Bits  range is 0..32,
                PStrToOct    parameter Digit range is 0..11,
                PStrToHex    parameter Digit range is 0..8,
                though 0 is useless (returns 0)}

FUNCTION PStrToBinary(P : PChar;Bits  : CARDINAL):CARDINAL;
FUNCTION PStrToOct   (P : PChar;Digits: CARDINAL):CARDINAL;
FUNCTION PStrToHex   (P : PChar;Digits: CARDINAL):CARDINAL;

{  -------------------------------------------------------------------

   Other than the above procedures, the procedures below (can) enlarge the
   string.  This could cause problems, I quickly checked strings.pp, but I
   can't see  any extra allocation for strings that get larger (e.g. strcat)

 If anyone could explain me the internals of the PChar in an E-mail, then I
 would be very gratefull

Strings seem to suppose that the PChars are always allocated big enough,
so so will I }

{ Make PChar P Count characters big. Pad right (at end of string) with C}
PROCEDURE PRGrow(P:PChar;Ch:CHAR;Count:WORD);

{ Make PChar P Count characters big. Pad left (at beginning of string) with C}
PROCEDURE PLGrow(P:PChar;Ch:CHAR;Count:WORD);

{ Old Basic procedure String$, fill string with chars.}
PROCEDURE PStrStr(P:PChar;Ch:Char;Count:LONGINT);

{ Procedure Insert for PChar's}
PROCEDURE StrInsert(Tobeins:PCHAR;Dest : PCHAR ;Position:WORD);

{ Inserts character on each 3rd spot starting from the back.
(2123456789 -> 2,123,456,789}
PROCEDURE PCommastr(P : PChar;sep:CHAR);

{ PAS,OVERLOADED
    Returns Nth substring of S in R (first is 0), separator is character T
One of my favourite string procedures. Multiple separators between char's
allowed}
{$IFDEF TURBO}
PROCEDURE PItemC(Dest,Source:PChar;T: CHAR; N: WORD);
{$ELSE}
PROCEDURE PItem(Dest,Source:PChar;T: CHAR; N: WORD);
{$ENDIF}

{ PAS,OVERLOADED
     Returns Nth substring of S in R (first is 0), separators are all
characters in S. One of my favourite string procedures.
(Same as Item, but with more possible separators)}
PROCEDURE PItem(Dest,Source:PChar;CONST T: CHARSET; N: WORD);

{ PAS
  Returns Nth substring of S in R (first is 0), separator are characters in
    string T.
 Interface procedure, which converts string T to a CHARSET and then calls Item.}
PROCEDURE PItemS(Dest,Source,T:PChar; N: WORD);

{
    Kind of Copy(the one for Strings) for PChars
    Copy Num chars from Source, starting at position Start to Dest}
PROCEDURE PCopy(Source,Dest:PChar;Start,Num: LONGINT);

{PAS,

Copy chars between first occurance of C1 and C2 to Dest, return status
C1=C2 is allowed}
FUNCTION PGetBetween (Source,Dest:PCHAR;C1,C2:CHAR):BOOLEAN;

{Expands (hard) tabs to spaces}
PROCEDURE PExpandTabs(P,P2:PChar;Tabsize:WORD);

{PAS, compress tabs to spaces, with variable tabsize,
 Doesn't function well with hardtabs in the Pchar Source.
 Run PExpandTabs first in that case.}
PROCEDURE PCompresstabs(Source,Dest:PChar;Tabsize:LONGINT);

{Convert Number(Value) to string(S) in binary, octal or hexadecimal
representation, with a configurable number of bits or digits,
Range of Bits  in PBinaryStr  0..32,
Range of Digit in POctToStr   0..11
Range of Digit in PHexToStr   0..8
Though value zero has no real meaning.}

PROCEDURE PBinaryToStr(P : PChar;Value,Bits   : CARDINAL);
PROCEDURE PHexToStr   (P : PChar;Value,Digits : CARDINAL);
PROCEDURE POctToStr   (P : PChar;Value,Digits : CARDINAL);


IMPLEMENTATION

Uses Strings;

{Pascal procedures}

FUNCTION PGetBetween (Source,Dest:PCHAR;C1,C2:CHAR):BOOLEAN;

VAR C,B : CARDINAL;

BEGIN
 Dest[0]:=CHR(0);
 IF StrLen(Source)=0 THEN
  BEGIN
   PGetBetween:=FALSE;
   EXIT
  END;
 B:=PCharPos(Source,C1);
 IF (B=(StrLen(Source)-1)) OR (B=65535) THEN
  BEGIN
   PGetBetween:=FALSE;
   EXIT
  END;
 C:=PNextCharPos(Source,C2,B+1);      { NextCharPos in case C1=C2}
 IF (C=65535) THEN
  BEGIN
   PGetBetween:=FALSE;
   EXIT
  END;
 IF C=(B+1) THEN
  BEGIN
   PGetBetween:=FALSE;
   EXIT
  END;
 PCopy(Source,Dest,B+1,C-B-1);
 PGetBetween:=TRUE;
END;

PROCEDURE PItem(Dest,Source:PChar; CONST T: CHARSET; N: WORD);
{ Returns Nth substring of S in R (first is 0), separators are all
characters in S. One of my favourite string procedures.}

VAR
  I,J,
  L : WORD;
  U : BOOLEAN;

BEGIN
  I := 0;
  L := StrLen(Source)+1;
  REPEAT
   WHILE (I <= L) AND (Source[I] IN T) DO INC(I); { Skip separators }
   IF (N = 0) OR (I = L) THEN   { N is number of separator-aggregates to
                                    skip, or end of string (number
                                    substring to high}
    U:=TRUE
   ELSE
    BEGIN
     U:=FALSE;
     DEC(N);                    { One Item/separator-aggregate less to go}
    WHILE (I < L) AND NOT (Source[I] IN T) DO INC(I);  { Skip item }
    END;
  UNTIL U;
  Dest[0]:=CHR(0);
  IF (N=0) THEN
   IF I<L THEN
  BEGIN
    J  := 0;
{  HR := HIGH(R);               Testing for High not possible}
  WHILE (I <= (L)) AND NOT (Source[I] IN T) DO
  BEGIN
    Dest[J] := Source[I];
    INC(I);
    INC(J);
  END;
  IF I>L THEN
   DEC(J);
  Dest[J]:=CHR(0);
  END
END;

{ Returns Nth substring of S in R (first is 0), separator are characters in
    string T
One of my favourite string procedures.}
PROCEDURE PItemS(Dest,Source,T:PChar; N: WORD);

VAR Ses : CHARSET;
    A   : WORD;

BEGIN
 Ses:=[];
 A:=StrLen(T);
 IF A>0 THEN
  BEGIN
   FOR A :=0 TO A-1 DO Ses:=Ses+[Source[A]];  { faster equivalent:
                                         ELib.INCL256(Ses,ORD(S[A]));,
                                        but that procedure is still experimental}
   PItem(Dest,Source,Ses,N);
  END;
END;

{OVERLOADED}
{$IFDEF TURBO}
PROCEDURE PItemC(Dest,Source:PChar;T: CHAR; N: WORD);
{$ELSE}
PROCEDURE PItem(Dest,Source:PChar;T: CHAR; N: WORD);
{$ENDIF}

{Returns Nth substring of S in R (first is 0), separator is character T
One of my favourite string procedures.}

VAR A,B,C,D,E : WORD;

BEGIN
 A:=StrLen(Source)+1; B:=0; C:=0; D:=0;
 IF Source[0]=T THEN
  BEGIN
   B:=1;
   D:=1;
  END;

 WHILE (B<A) AND (C<(N+1)) DO
  BEGIN
   IF Source[B]=T THEN
    BEGIN
     IF Source[B-1] <> T THEN
      BEGIN
       INC(C);
       E:=D;
       D:=B;
      END;
    END;
   INC(B);
  END;
 IF (B=A) THEN
BEGIN
  E:=D;
  D:=B+1;
 END
ELSE
 DEC(D);
 IF N=0 THEN
  DEC(E);
 IF (C=0) OR (C<N) THEN
  Dest[0]:=CHR(0)
 ELSE
  BEGIN
   INC(E);
   WHILE Source[E]=T DO INC(E);
   PCopy(Source,Dest,E,D-E);
  END;
END;

PROCEDURE PCompresstabs(Source,Dest:PChar;Tabsize:LONGINT);

VAR I,B,C,D,RestSpac,
    L1,L2 : LONGINT;

BEGIN
 I:=0; B:=0;
 WHILE Source[I]<>CHR(0) DO
  BEGIN
   D:=0;
   WHILE (Source[I+D]=CHR(32)) AND (Source[D+I]<>CHR(0)) DO { Count spaces}
    INC(D);
   IF D<>0 THEN
    BEGIN
     IF I=0 THEN
      L1:=I DIV Tabsize
     ELSE
      L1:=(I-1) DIV Tabsize;
     L2:=(I+D-1) DIV Tabsize;
     IF L1<L2 THEN               {Tab border crossed?}
      BEGIN
       FOR C:= L1 TO L2-1 DO     {Insert all tabs needed}
        BEGIN
         Dest[B]:=CHR(9);
         INC(B);
        END;
       RestSpac:=(I+D-1) MOD Tabsize;   {Insert rest-spaces}
       IF RestSpac>0 THEN
        FOR C:=1 TO RestSpac DO
         BEGIN
          Dest[B]:=CHR(32);
          INC(B);
         END;
        INC(I,D);                       {Advance over spaces}
      END
     ELSE
      BEGIN
       FOR C:=0 TO D-1 DO Dest[B+C]:=Source[I+C];
       INC(B,D); INC(I,D);
      END;
     END
     ELSE
      BEGIN
       Dest[B]:=Source[I];
       INC(B); INC(I);
      END;
  END;
 Dest[B]:=CHR(0);
END;


{$IFNDEF Useasm}

PROCEDURE PCommastr(P : PChar;sep:CHAR);

VAR L,I ,J   : WORD;
    SepStr   : WORD;
    SepPChar : PCHar;

BEGIN
 SepStr:=ORD(Sep);
 SepPChar:=@SepStr;
 L:=StrLen(P);
 I:=L DIV 3;
 IF ((L MOD 3)=0) AND (I>0) THEN
  DEC(I);
 IF I>0 THEN
  FOR J:= 1 TO I DO
    StrInsert(SepPChar,P,L-J*3);
END;


PROCEDURE StrInsert(Tobeins:PCHAR;Dest : PCHAR ;Position:WORD);

VAR L,TobeMove,I : WORD;

BEGIN
 TobeMove:=StrLen(ToBeIns);
 IF TobeMove>0 THEN              {Do we have to insert?}
  BEGIN
   L:=StrLen(Dest);
   IF Position>L THEN
    Position:=L;
   FOR I:=L+1 DOWNTO Position DO  {+1= copy CHR(0)}
    Dest[I+TobeMove]:=Dest[I];
   FOR I:=0 TO Tobemove-1 DO
    Dest[Position+I]:=Tobeins[I];
   END;
END;

CONST MaxWord=65535;

PROCEDURE PLTrim(P :PChar;Ch:Char);

VAR I,J : WORD;

BEGIN
  J:=0;
  WHILE (P[J]=Ch) AND (P[J]<>CHR(0)) DO INC(J);
   IF J>1 THEN
    StrDelete(P,0,J);
END;

PROCEDURE PRTrim(P : PChar;Ch:Char);

VAR I,J : WORD;

BEGIN
 I:=StrLen(P);
 IF (I>0) THEN
  BEGIN
   J:=I-1;
   WHILE (P[J]=Ch) AND (J>0) DO DEC(J);
   IF (J=0) AND (P[0]=Ch) THEN
    P[J]:=CHR(0)
   ELSE
    P[J+1]:=CHR(0);
   END;
END;

PROCEDURE PRGrow(P : PChar;Ch:CHAR;Count:WORD);

VAR I,J :WORD;

BEGIN
 I:=StrLen(P);
 IF I<Count THEN
  BEGIN
   FOR J:=I TO Count-1 DO
    P[J]:=Ch;
   P[Count]:=CHR(0);
  END;
END;

PROCEDURE PLGrow(P:PChar;Ch:CHAR;Count:WORD);

VAR  I,J :WORD;

BEGIN
 I:=StrLen(P);
 IF I<Count THEN
  BEGIN
   IF I<>0 THEN
    Move(P[0],P[Count-I],I);
   FOR J:=0 TO Count-I-1 DO
    P[J]:=Ch;
   P[Count]:=CHR(0);
  END;
END;

PROCEDURE PKillChar(P : PChar;CONST CSet:CHARSET);

VAR J : WORD;

BEGIN
  J:=0;
  WHILE (P[J] IN CSet) AND (P[J]<>CHR(0)) DO INC(J);
   IF J>1 THEN
    StrDelete(P,0,J);
END;

PROCEDURE PKillBChar(P : PChar;CONST CSet:CHARSET);

VAR I,J : WORD;

BEGIN
 I:=StrLen(P);
 IF (I>0) THEN
  BEGIN
   J:=I-1;
   WHILE (P[J] IN CSet) AND (J>0) DO DEC(J);
   IF (J=0) AND (P[0] IN CSet) THEN
    P[J]:=CHR(0)
   ELSE
    P[J+1]:=CHR(0);
   END;
END;

FUNCTION PCharPos(P:PChar;Ch:CHAR):WORD;

VAR J : WORD;

BEGIN
 J:=0;
 WHILE (P[J]<>Ch) AND (P[J]<>CHR(0)) DO INC(J);
 IF P[J]=CHR(0) THEN
  J:=MaxWord;
 PCharPos:=J;
END;

FUNCTION PNextCharPos(P:PChar;Ch:CHAR;Count:WORD):WORD;

VAR I,J:WORD;

BEGIN
 I:=StrLen(P);  {Now I have to check two chars, just to make sure that Count<Length}
 J:=Count;
 IF J>=I THEN
  J:=I-1;
 WHILE (P[J]<>Ch) AND (P[J]<>CHR(0)) DO INC(J);
 IF P[J]=CHR(0) THEN
  J:=MaxWord;
 PNextCharPos:=J;
END;

FUNCTION PRCharPos(P:PChar;Ch:CHAR):WORD;

VAR J:WORD;

BEGIN
 J:=StrLen(P);
 IF J>0 THEN
  BEGIN
   WHILE (P[J]<>Ch) AND (J>0) DO DEC(J);
   IF J=0 THEN
     IF P[0]=Ch THEN
      PRCharPos:=0
     ELSE
      PRCharPos:=MaxWord
   ELSE
    PRCharPos:=J
  END
 ELSE
  PRCharPos:=MaxWord;
END;

FUNCTION PNextRCharPos(P:PChar;Ch:CHAR;Count:WORD):WORD;

VAR J:WORD;

BEGIN
 J:=StrLen(P);
 IF J>0 THEN
  BEGIN
   DEC(J);
   IF Count<J THEN
    J:=Count;
   WHILE (P[J]<>Ch) AND (J>0) DO DEC(J);
   IF J=0 THEN
     IF P[0]=Ch THEN
      PNextRCharPos:=0
     ELSE
      PNextRCharPos:=MaxWord
   ELSE
    PNextRCharPos:=J
  END
 ELSE
  PNextRCharPos:=MaxWord;
END;


FUNCTION PCharPosSet(P : PChar;CONST CSet:CHARSET):WORD;


VAR J : WORD;

BEGIN
 J:=0;
 WHILE NOT (P[J] IN CSet) AND (P[J]<>CHR(0)) DO INC(J);
 IF P[J]=CHR(0) THEN
  J:=MaxWord;
 PCharPosSet:=J;
END;

PROCEDURE PReplaceChar(P :PChar;ReplaceMe,RepWith:CHAR);

VAR I,J : WORD;

BEGIN
 I:=StrLen(P);
 IF I>0 THEN
  FOR J:=0 TO I-1 DO IF P[J]=ReplaceMe THEN P[J]:=RepWith;
END;

PROCEDURE PStrStr(P:PChar;Ch:Char;Count:LONGINT);

VAR I :WORD;

BEGIN
 IF Count>0 THEN
  FOR I:=0 TO Count-1 DO
   P[I]:=Ch;
 P[Count]:=CHR(0);
END;

PROCEDURE PStripChar(P : PChar;C:CHAR);

VAR I,J,L : WORD;

BEGIN
 L:=StrLen(P);
 IF L>0 THEN
  BEGIN
   J:=0;
   FOR I:=0 TO L-1 DO
    IF P[I]<>C THEN
     BEGIN
      P[J]:=P[I];
      INC(J);
     END;
   P[J]:=CHR(0);
  END;
END;

PROCEDURE PKillChrTot(P:PChar;CONST CSet:CHARSET);

VAR I,J,L : WORD;

BEGIN
 L:=StrLen(P);
 IF L>0 THEN
  BEGIN
   J:=0;
   FOR I:=0 TO L-1 DO
    IF NOT(P[I] IN CSet) THEN
     BEGIN
      P[J]:=P[I];
      INC(J);
     END;
   P[J]:=CHR(0);
  END;
END;

PROCEDURE PAppendBackSlash(P : PChar);

VAR I : WORD;

BEGIN
 I:=StrLen(P);
 IF I>0 THEN
  IF P[I-1]<>'\' THEN
   BEGIN
    P[I]:='\';
    P[I+1]:=CHR(0);
   END;
END;

PROCEDURE PStripDoubleChar(P : PChar;Ch:CHAR);

VAR I,J,L : WORD;

BEGIN
 L:=StrLen(P);
 IF L>1 THEN  {No double chars to strip when Length(S)=1}
  BEGIN
   J:=1;
   FOR I:=1 TO L-1 DO
    IF NOT((P[I]=Ch) AND (P[I]=P[I-1])) THEN
     BEGIN
      P[J]:=P[I];
      INC(J);
     END;
   P[j]:=CHR(0);
  END;
END;

{Like Delete for pascal-strings, but for PChars}
PROCEDURE StrDelete(P:PChar;Position,Count:WORD);

VAR L,J :WORD;

BEGIN
 L:=StrLen(P);
 IF Position<L THEN
  IF (Count+Position)>L THEN
   P[Position]:=CHR(0)
  ELSE
   BEGIN
    L:=Position+Count;
    J:=Position;
    WHILE P[L]<>CHR(0) DO
     BEGIN
      P[J]:=P[L];
      INC(J);
      INC(L);
     END;
    P[J]:=CHR(0);
   END;
END;

{
    Kind of Copy(the one for Strings) for PChars
    Copy Num chars from Source, starting at position Start to Dest}
PROCEDURE PCopy(Source,Dest:PChar;Start,Num: WORD);

VAR L,I:WORD;

BEGIN
 L:=StrLen(Source);
 IF Start<L THEN
  BEGIN
   I:=Start;
   L:=0;
   WHILE (Source[I]<>CHR(0)) AND (L<Num) DO
    BEGIN
     Dest[L]:=Source[I];
     INC(I);
     INC(L);
    END;
   Dest[L]:=CHR(0);
   END
  ELSE
   Dest[0]:=CHR(0);
END;

PROCEDURE PUpperCase(P:PChar);

VAR I : WORD;

BEGIN
 I:=0;
 WHILE P[I]<>CHR(0) DO
  BEGIN
   IF (P[I]>CHR(96)) AND (P[I]<CHR(123)) THEN
    P[I]:=CHR(ORD(P[I])-32);
   INC(I);
  END;
END;

PROCEDURE PLowerCase(P:PChar);

VAR L,I : WORD;

BEGIN
 I:=0;
 WHILE P[I]<>CHR(0) DO
  BEGIN
   IF (P[I]>CHR(64)) AND (P[I]<CHR(91)) THEN
    P[I]:=CHR(ORD(P[I])+32);
   INC(I);
  END;
END;

PROCEDURE PExpandTabs(P,P2:PChar;Tabsize:WORD);

VAR I,B,C,D : WORD;

BEGIN
 I:=0; B:=0;
 WHILE P[I]<>CHR(0) DO
  BEGIN
   IF P[I]=CHR(9) THEN
    BEGIN
     C:=Tabsize-(B MOD TabSize);
     FOR D:=0 TO C-1 DO
      P2[B+D]:=' ';
     B:=B+C-1;
    END
   ELSE
    P2[B]:=P[I];
   INC(B); INC(I);
  END;
 P2[B]:=CHR(0);
END;

PROCEDURE PBinaryToStr(P : PChar ;Value,Bits : CARDINAL);

VAR I,J : WORD;

BEGIN
 IF Bits>0 THEN
  BEGIN
   IF Bits<>32 THEN
    Value:=Value SHL (32-bits);
   FOR I:=0 TO Bits-1 DO
    BEGIN
     IF (Value AND (1 SHL 31))<>0 THEN
      P[I]:='1'
     ELSE
      P[I]:='0';
    Value:=Value shl 1;
    END;
   END;
 P[Bits]:=CHR(0);
END;

FUNCTION PStrToBinary(P : PCHAR;Bits : CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 I:=0;
 IF Bits>0 THEN
  BEGIN
   FOR A :=0 TO Bits-1 DO
    BEGIN
     I:=I SHL 1;
     I:=I OR (ORD(P[A]) AND 1);
    END;
  END;
 PStrToBinary:=I;
END;

PROCEDURE PHexToStr(P : PChar;Value,Digits : CARDINAL);

VAR I : WORD;
    O : CARDINAL;

BEGIN
 IF Digits >0 THEN
  BEGIN
   IF Digits<>32 THEN
    Value:=Value SHL (32-(Digits shl 2));
   FOR I:=0 TO Digits-1 DO
    BEGIN
     O:=(Value SHR 28) +48;
     IF O>57 THEN
      INC(O,7);
     P[I]:=CHR(O);
    Value:=Value shl 4;
    END;
   END;
  P[Digits]:=CHR(0);
END;


PROCEDURE POctToStr(P : PChar;Value,Digits : CARDINAL);

VAR I : WORD;
    O : CARDINAL;

BEGIN
 IF Digits >0 THEN
  BEGIN
   IF Digits<>32 THEN
    Value:=Value SHL (32-(Digits shl 1 )-Digits);
   FOR I:=0 TO Digits-1 DO
    BEGIN
     O:=(Value SHR 29) +48;
     P[I]:=CHR(O);
    Value:=Value shl 3;
    END;
   END;
  P[Digits]:=CHR(0);
END;


FUNCTION PStrToHex(P : PChar;Digits: CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 I:=0;
 IF Digits>0 THEN
  BEGIN
   FOR A :=0 TO Digits-1 DO
    BEGIN
     I:=I SHL 4;
     IF P[A]>CHR(58) THEN
      I:=I+   ((ORD(P[A])-55) AND 15)
     ELSE
      I:=I + ((ORD(P[A])-48) AND 15);
    END;
  END;
 PStrToHex:=I;
END;

FUNCTION PStrToOct(P : PChar;Digits: CARDINAL):CARDINAL;

VAR I : CARDINAL;
    A : CARDINAL;

BEGIN
 I:=0;
 IF Digits>0 THEN
  BEGIN
   FOR A :=0 TO Digits-1 DO
     I:=(I SHL 3) +  ((ORD(P[A])-48) AND 7)
  END;
 PStrToOct:=I;
END;


{$ELSE}

{FUNCTION PStrToBinary(P : PCHAR ;Bits : CARDINAL):CARDINAL; ASSEMBLER;
{Same as StrTobinary except for one INC %esi}

ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    P,%esi
                            movl    bits,%ecx
                            jecxz   .LPStrToBinNothing
.LPStrToBinNextbit:         shl     $1,%ebx
                            lodsb
                            and     $1,%eax       // 48 and 1=0 | 49 and 1=1
                            or      %eax,%ebx     // move into low bit
                            loop    .LPStrToBinNextbit
.LPStrToBinNothing:         mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];
}

PROCEDURE PCopy(Source,Dest:PChar;Start,Num: LONGINT); ASSEMBLER;
ASM
                   movl Source,%edi       // Load source PChar
                   movl %edi,%esi
                   movl $0xffffffff,%ecx   // StrLength(P) as inline
                   xorb %al,%al
                   repne
                   scasb
                   not  %ecx
                   decl %ecx
                   movl Source,%esi        // Load source PChar
                   movl Dest,%edi       // Load target PChar
                   movl Start,%ebx     // Load start
                   xor %eax,%eax
                   cmp %ecx,%ebx            // Start>Length?
                   ja .LPCopyExit
                   addl %ebx,%esi
                   movl Num,%eax
                   addl %eax,%ebx           // ebx=Start+Num
                   cmp  %ecx,%ebx
                   jb   .LPCopyOK
                   movl Start,%ebx
                   subl  %ecx,%ebx
                   mov  %ecx,%eax

     .LPCopyOK:    cld
                   mov %eax,%ecx
                   rep
                    movsb
                   xor %eax,%eax
    .LPCopyExit:  stosb

END['EAX','EBX','ECX','ESI','EDI'];

PROCEDURE PLTrim(P : PChar;Ch:CHAR); ASSEMBLER;
  asm
                   movl P,%edi       // Load PChar
                   movl %edi,%esi
                   movl $0xffffffff,%ecx   // Length(P) as inline
                   xorb %al,%al
                   repne
                   scasb
                   not  %ecx
                   movl %ecx,%ebx
                   movl %esi,%edi
                   movb Ch,%al      // Load character
                   cld
                   repe
                   scasb                // Search for NOT equal, no longer
                                         // then string length
                   dec %edi
                   cmpl %edi,%esi        // Chars found?
                   je   .LNothingToDo
                   movl %esi,%ecx        // ecx=esi-edi Result is negative
                   subl %edi,%ecx        // add ebx
                   addl %ebx,%ecx        // Chars to be copied calculated
                   xchgl %esi,%edi
  .LTRIM1:         lodsb
                   stosb
                   decl  %ecx
                   jz .LTRIM2      // 0 reached ends
                   orb %al,%al
                   jnz .LTRIM1
 .LNothingToDo:    movl P,%eax  // Returnvalue, not yet in def
                   leave
                   ret $6
  .LTRIM2:         xorb %al,%al        // If cutted
                   stosb               // add a #0
                   movl P,%eax
end ['EAX','EBX','ECX','ESI','EDI'];

PROCEDURE PKillChar(P : PChar;CONST CSet:CHARSET); ASSEMBLER;
asm
                   movl P,%edi       // Load PChar
                   movl %edi,%esi
                   cld
                   movl $0xffffffff,%ecx   // Length(P) as inline
                   xorb %al,%al
                   repne
                    scasb
                   not   %ecx
                   movl  %ecx,%ebx
                   movl  %esi,%edi
                   mov   CSet,%edi
 .LKILLCHAR4:      test  %ecx,%ecx
                   je     .LNothingToDo2
                   lodsb
                   mov   %eax,%edx
                   andl  $0xf8,%edx
                   shrl  $3,%edx
                   andl  $7,%eax
                   decl  %ecx
                   btl   %eax,(%edi,%edx)
                   jc   .LKILLCHAR4
     .LKILLCHAR3:  dec  %esi
                   movl P,%edi       // Load PChar
                   cmpl %edi,%esi        // Chars found?
                   je   .LNothingToDo2
                   movl %edi,%ecx        // ecx=esi-edi Result is negative
                   subl %esi,%ecx        // add ebx
                   addl %ebx,%ecx        // Chars to be copied calculated
 .LKILLCHAR1:      lodsb
                   stosb
                   decl  %ecx
                   jz .LKILLCHAR2      // 0 reached ends
                   orb %al,%al
                   jnz .LKILLCHAR1
.LNothingToDo2:    movl P,%eax  // Returnvalue, not yet in def
  .Ldumptst:       leave
                   ret $8
 .LKILLCHAR2:      xorb %al,%al        // If cutted
                   stosb               // add a #0
                   movl P,%eax
end ['EAX','EBX','ECX','EDX','ESI','EDI'];

{ Old Basic procedure String$, fill string with chars.}
PROCEDURE PStrStr(P:PChar;Ch:Char;Count:LONGINT); ASSEMBLER;

asm
                   mov   P,%edi
                   cld
                   movl  Count,%ecx
                   movb  Ch,%al
                   movb  %al,%ah
                   movl  %eax,%ebx
                   shll  $16,%eax
                   movw  %bx,%ax
                   movl  %ecx,%ebx
                   andl  $3,%ebx
                   shrl  $2,%ecx
                   rep
                    stosl
                   movl  %ebx,%ecx
                   rep
                    stosb
                   xorb %al,%al
                   stosb
end ['EAX','EBX','ECX','EDI'];

PROCEDURE PAppendBackslash(P:PChar); ASSEMBLER;
ASM
                         movl P,%edi
                         cld
                         movl $0xffffffff,%ecx   // Length(P) as inline
                         xorb %al,%al
                         repne
                          scasb
                         decl %edi
{$ifdef linux}
                        cmpb $47,-1(%edi)
                         je .Lappendbackslashready
                         movw $47,(%edi)      // append backslash and new zero

{$else}
                          // Now we are pointing to the zero.
                         cmpb $92,-1(%edi)
                         je .Lappendbackslashready
                         movw $92,(%edi)      // append backslash and new zero
{$endif}
.Lappendbackslashready:
END ['EAX','EDI'];

PROCEDURE PRTrim(P:PChar;Ch:Char); ASSEMBLER;
  asm
                   movl P,%edi       // Load PChar
                   cld
                   movl $0xffffffff,%ecx   // Length(P) as inline
                   xorb %al,%al
                   repne
                   scasb
                   not   %ecx
                   decl %ecx
                   movl %edi,%esi
                   test %ecx,%ecx
                   je  .LRTrimStringEmpty
                   movb  Ch,%al
                   decl %edi
     .LRTrimloop:  cmpb %al,-1(%edi)
                   jne .LRTrimTrimit
                   decl %edi
                   loop .LRTrimloop
  .LRTrimTrimit:   cmpl  %edi,%esi
                   je .LRTrimStringEmpty
                   movb $0,(%edi)

.LRTrimStringEmpty: movl P,%eax
end ['EAX','ECX','ESI','EDI'];

PROCEDURE PKillBChar(P:PChar;const CSet:CHARSET); ASSEMBLER;
  asm
                   movl P,%edi       // Load PChar
                   cld
                   movl $0xffffffff,%ecx   // Length(P) as inline
                   xorb %al,%al
                   repne
                   scasb
                   not  %ecx
                   decl %ecx
                   test %ecx,%ecx
                   je  .LKillBPCharStringEmpty
                   addl $2,%ecx
                   movl  CSet,%esi
                   decl %edi                     // EDI points to last char.
                                                 // and string isn't empty
                   movl  %edi,%ebx
.LKillBPCharLoop:   decl  %edi
                   test  %ecx,%ecx
                   je    .LKillBPCharTrimit
                   movb  (%edi),%al
                   movl  %eax,%edx            // And clears the high bytes
                   andl  $0xf8,%edx
                   shrl  $3,%edx
                   andl  $7,%eax
                   decl  %ecx
                   je    .LKillBPCharTrimit

                   btl   %eax,(%esi,%edx)
                   jc   .LKillBPCharLoop

.LKillBPCharTrimit:
                   mov   P,%esi
                   cmpl  %edi,%ebx
                   je .LKillBPCharStringEmpty  // no characters deleted.
                   movb $0,1(%edi)
.LKillBPCharStringEmpty:
                       movl P,%eax
end ['EAX','EBX','ECX','EDX','ESI','EDI'];

(* Replace in string S the character "ReplaceMe" with "RepWith".*)
PROCEDURE PReplaceChar(P : PChar;ReplaceMe,RepWith:CHAR); ASSEMBLER;

ASM
                      movl P,%edi       // Load PChar
                      cld
                      movl $0xffffffff,%ecx   // Length(P) as inline
                      xorb %al,%al
                        repne
                        scasb
                        not    %ecx
                        decl   %ecx
                        test   %ecx,%ecx
                        je     .LReplaceCharStringEmpty
                        movl   P,%edi
                        movb   ReplaceMe,%bl
                        movb   RepWith,%al
.LReplaceCharLoop:      cmpb   %bl,(%edi)
                        jne    .LReplaceCharNextChar
                        movb   %al,(%edi)
.LReplaceCharNextChar:  incl   %edi
                        loop   .LReplaceCharLoop
.LReplaceCharStringEmpty:
end ['EAX','EBX','ECX','EDI'];

PROCEDURE PRGrow(P:PChar;Ch:CHAR;Count:WORD); ASSEMBLER;

                ASM
                      movl P,%edi       // Load PChar
                      cld
                      movl $0xffffffff,%ecx   // Length(P) as inline
                      xorb %al,%al
                      repne
                       scasb
                      not    %ecx
                      decl   %ecx
                      movzwl Count,%eax         // load count
                      cmp  %eax,%ecx               // subtract size
                      ja    .LRGrowStringBigEnough
                      sub  %ecx,%eax
                      mov  %eax,%ecx
                      movb Ch,%al
                      decl  %edi                   // point to zero char
                      rep
                       stosb                      //  store padding
                      xorb %al,%al                // store terminating zero
                      stosb
.LRGrowStringBigEnough:
end ['EAX','ECX','EDI'];

PROCEDURE PLGrow(P:PChar;Ch:CHAR;Count:WORD); ASSEMBLER;

ASM
                      movl P,%edi       // Load PChar
                      cld
                      movl $0xffffffff,%ecx   // Length(P) as inline
                      xorb %al,%al
                      repne
                       scasb
                      not    %ecx
                      decl   %ecx
                      movl   %ecx,%ebx

    

                      movzwl Count,%eax     // load count
                      cmp  %eax,%ebx           // subtract size
                      ja    .LGrowStringBigEnough
                      sub  %ebx,%eax
                      mov  %eax,%ebx           // bx is nr pads, cx is length
                      decl %edi
                      movl %edi,%esi
                      addl %ebx,%edi           // moving up, std required,
                                               // calculating si
                      incl %ecx                // also move zero byte
                      std
                      rep                      // move string, make room
                       movsb                   //  for padding
                      cld
                      movb Ch,%al
                      movl P,%edi        //  reload Pchar and pad-char
                      movl %ebx,%ecx           //  count padding chars
                      rep
                      stosb
                      mov  %ebx,%eax
.LGrowStringBigEnough:

end ['EAX','EBX','ECX','ESI','EDI'];

{Remove C entirely from P}
PROCEDURE PStripChar(P:PChar;C:CHAR); ASSEMBLER;
   ASM
                  movl  P,%esi
                  movl  %esi,%edi
                  movb  C,%bl
.LStripCharLoop:  lodsb
                  cmpb  %al,%bl
                  je    .LStripCharSkip
                  stosb
.LStripCharSkip:  test %al,%al
                  jne  .LStripCharLoop
END ['EAX','EBX','ESI','EDI'];


{StripChar for a CHARSET; Remove any char in C entirely from P}
PROCEDURE PKillChrTot(P:PChar;CONST CSet: CHARSET); ASSEMBLER;
   ASM
                  movl  P,%esi
                  movl  %esi,%edi
                  movl  CSet,%ebx

.LKillChrTotLoop: lodsb
                  mov %eax,%ecx
                  test %al,%al
                  je  .LKillChrTotSkip
                  mov   %eax,%edx
                  andl  $0xf8,%edx
                  shrl  $3,%edx
                  andl  $7,%eax
                  btl   %eax,(%ebx,%edx)
                  jc    .LKillChrTotLoop
                  mov   %ecx,%eax
                  stosb
                  jmp .LKillChrTotLoop
.LKillChrTotSkip: stosb
END ['EAX','EBX','ECX','EDX','ESI','EDI'];


{ Pos for one char only. Faster than an ordinary Pos}
FUNCTION PCharPos(P:PChar;C:Char):WORD;   ASSEMBLER;
    ASM  // Can't use scasb, or will have to do it twice, think this
         //   is faster.
                      movl  P,%esi       // Load PChar
                      movb  C,%bl
                      xor   %ecx,%ecx
.LCharPosLoop:        lodsb
                      cmpb  %al,%bl
                      je    .LCharPosCharFound
                      incl   %ecx
                      testb  %al,%al
                      jne   .LCharPosLoop
                      mov  $-1,%ecx
.LCharPosCharFound:   mov  %ecx,%eax
END ['EAX','EBX','ECX','ESI'];

{ Like CharPos but starts searching on character number Count (zero based),
    returnvalue is zero based and absolute (to beginning of string)}
FUNCTION PNextCharPos(P:PChar;Ch:Char;Count:WORD):WORD; ASSEMBLER;

        ASM
                          movl   P,%esi       // Load PChar
                          movb   Ch,%bl
                          movzwl Count,%edx
                          add    %edx,%esi
                          xor    %ecx,%ecx
  .LNextCharPosLoop:      lodsb
                          cmpb   %al,%bl
                          je     .LNextCharPosChar
                          inc    %ecx
                          testb  %al,%al
                          jne    .LNextCharPosLoop
                          mov    $-1,%ecx
.LNextCharPosChar:        mov    %ecx,%eax
                          addl   %edx,%eax
   END ['EAX','EBX','ECX','EDX','EDI'];

PROCEDURE PStripDoubleChar(P:PChar;C:Char); ASSEMBLER;

ASM
                          movl   P,%esi       // Load PChar
                          movl   %esi,%edi
                          movb   C,%bl
  .LStripDoubleLoop:      lodsb
                          testb  %al,%al
                          je    .LStripDoubleExit
                          cmp    %al,%bl          // Check for right char
                          jne    .LStripDoubleCheck
                          cmpb   %al,(%esi)       // Equal to next, skip
                          je     .LStripDoubleLoop
.LStripDoubleCheck:       stosb
                          jmp    .LStripDoubleLoop
.LStripDoubleExit:        stosb
END ['EAX','EBX','ESI','EDI'];

FUNCTION PCharPosSet(P:PChar;CONST CSet:CHARSET):WORD; ASSEMBLER;

ASM
                      movl  P,%esi       // Load PChar
                      movl  CSet,%ebx
                      xor   %ecx,%ecx
.LCharPosSetLoop:       lodsb
                      test  %al,%al
                      je    .LCharPosSetEOS
                      incl  %ecx
                      movl  %eax,%edx
                      andl  $0xf8,%edx
                      shrl  $3,%edx
                      andl  $7,%eax
                      btl   %eax,(%ebx,%edx)
                      jnc   .LCharPosSetLoop
                      decl  %ecx
                      mov   %ecx,%eax
                      leave
                      ret $8
.LCharPosSetEOS:        mov  $-1,%eax
END ['EAX','EBX','ECX','EDX','ESI'];

FUNCTION PRCharPos(P:PChar;C:Char):WORD; ASSEMBLER;

ASM
                      movl P,%edi       // Load PChar
                      cld
                      movl $0xffffffff,%ecx   // Length(P) as inline
                      xorb %al,%al
                      repne
                       scasb
                      not    %ecx
                      decl %edi
                      incl %ecx
                      movb  C,%al
                      std
                      repne
                       scasb
                      cld
                      mov %ecx,%eax
                      decl %eax
END ['EAX','ECX','EDI'];

FUNCTION PNextRCharPos(P:PChar;Ch:Char;Count:WORD):WORD; ASSEMBLER;

    ASM
                      movl   P,%edi       // Load PChar
                      movzwl Count,%ecx
                      addl   %ecx,%edi
                      decl   %edi
                      incl   %ecx
                      movb   Ch,%al
                      std
                      repne
                       scasb
                      cld
                      decl %ecx
                      mov %ecx,%eax
END ['EAX','ECX','EDI'];

PROCEDURE StrDelete(P:PChar;Position,Count:WORD); ASSEMBLER;
ASM
            movl P,%edi       // Load PChar
            cld
            movl $0xffffffff,%ecx   // Length(P) as inline
            xorb %al,%al
            repne
            scasb
            not  %ecx
            movzwl Position,%ebx
            cmp %ebx,%ecx
            jb .LStrDelete          //Noting to do, position >length

            movzwl Count,%eax
            movl P,%esi
            addl %ebx,%esi
            mov  %esi,%edi
            subl %ebx,%ecx      //%eax=Length-Position, =max bytes to copy
            cmp  %eax,%ecx
            jbe  .LDeleteRest
            movl  %ecx,%ebx
            subl %eax,%ecx      //Bytes to copy=restlength-Count
            addl %eax,%esi
            rep
             movsb
.LDeleteRest: xor %eax,%eax
              movb %al,(%edi)
.LStrDelete:
END['EAX','EBX','ECX','ESI','EDI'];


PROCEDURE PUpperCase(P:PChar); ASSEMBLER;

ASM
                        movl   P,%esi
                        mov    %esi,%edi
.LUpperLoop:            lodsb
                        test   %al,%al
                        je     .LUpperEnd
                        cmp  $97,%al
                        jb  .LUpperNotthischar
                        cmp  $122,%al
                        ja  .LUpperNotthischar
                        subb $32,%al
    .LUpperNotthischar: stosb
                        loop .LUpperLoop
.LUpperEnd:
end ['EAX','ESI','EDI'];

PROCEDURE PLowerCase(P:PChar); ASSEMBLER;

ASM
                        movl   P,%esi
                        mov    %esi,%edi
.LLowerLoop:            lodsb
                        test   %al,%al
                        je     .LLowerEnd
                        cmp  $65,%al
                        jb  .LLowerNotthischar
                        cmp  $90,%al
                        ja  .LLowerNotthischar
                        addb $32,%al
    .LLowerNotthischar: stosb
                        loop .LLowerLoop
.LLowerEnd:
end ['EAX','ESI','EDI'];

PROCEDURE StrInsert(Tobeins:PCHAR;Dest : PCHAR ;Position:WORD); ASSEMBLER;

ASM
                   movl ToBeIns,%edi       // Load source PChar
                   movl $0xffffffff,%ecx   // StrLength(P) as inline
                   xorb %al,%al
                   repne
                   scasb
                   not  %ecx
                   dec  %ecx
                   jecxz .LNothingToInsert
                   mov  %ecx,%ebx
                   movl Dest,%edi       // Load source PChar
                   movl %edi,%esi
                   movl $0xffffffff,%ecx   // StrLength(P) as inline
                   xorb %al,%al
                   repne
                    scasb
                   not  %ecx
                   dec %ecx
                   add  %ecx,%esi
                   mov  %esi,%edi
                   movzwl Position,%edx
                   cmp  %ecx,%edx
                   jb  .LPositionOK
                   mov  %ecx,%edx
.LPositionOK:      incl %ecx
                   addl %ebx,%edi
                   subl %edx,%ecx
                   std
                   rep
                    movsb
                   cld
                   mov ToBeIns,%esi
                   mov Dest,%edi
                   mov %ebx,%ecx
                   add %edx,%edi
                   rep
                    movsb
.LNothingToInsert:
END['EAX','EBX','ECX','EDX','ESI','EDI'];

PROCEDURE PCommastr(P:PChar;sep:CHAR); ASSEMBLER;

ASM
                   mov   P,%edi
                   movl  %edi,%esi
                   movl $0xffffffff,%ecx   // StrLength(P) as inline
                   xorl %eax,%eax
                   repne
                    scasb
                   not  %ecx
                   dec  %ecx
                   mov   %ecx,%eax
                   mov   %eax,%ebx
                   mov   $3,%ecx
//                   cwde
                  cltd
                   idiv   %ecx
                   je   .LLessthan4
                   test  %edx,%edx
                   jne   .LNotDiv3
                   dec %eax
                   je  .LLessthan4
 .LNotDiv3:
                   add   %ebx,%esi
                   mov   %esi,%edi
                   add   %eax,%edi
                   mov   %eax,%ecx
                   movl  sep,%eax
                   std
                   movsb   // copy zero-byte
.LNext3:
                   movsb
                   movsb
                   movsb
                   stosb
                   loop .LNext3
                   cld
.LLessthan4:
END['EAX','EBX','ECX','EDX','ESI','EDI'];

PROCEDURE PExpandTabs(P,P2:PChar;Tabsize:WORD); assembler;
ASM
            movl P,%esi       // Load PChar
            movl P2,%edi
            cld
            xor %ebx,%ebx
.LPExpandLoop:lodsb
            testb %al,%al
            je .LPExpandReady
            cmp $9,%al
            je .LPExpandThisTab
            inc %ebx
            stosb
            jmp .LPExpandLoop
.LPExpandThisTab:
            movzwl Tabsize,%ecx
            mov %ebx,%eax
            // cwde
            cltd
            idiv %ecx
            subl %edx,%ecx
            addl %ecx,%ebx
.Lnodebug:
            movb $32,%al
            rep
             stosb
            jmp .LPExpandLoop
.LPExpandReady:
              stosb
.LQuickexit:
end['EAX','EBX','ECX','EDX','ESI','EDI'];

PROCEDURE PBinaryToStr(P : PChar;Value,Bits : CARDINAL);ASSEMBLER;

ASM
                            movl    P    ,%edi
                            movl    Value,%ebx
                            movl    Bits ,%edx
                            mov     $32,%ecx
                            sub     %edx,%ecx        // calc unwanted bits
                            jecxz   .LPBinaryNoShl   // 32-Bits, test for 0
                            shll    %cl,%ebx         // Shift left to remove
                                                     // unwanted bits.
.LPBinaryNoShl:             mov     %edx,%ecx        // Number of bits to do
                            jecxz   .LPBinaryNothing // test for no-bits.

.LPBinaryLoop:              xor     %eax,%eax        // Empty eax
                            shl     $1,%ebx          // highest bit to carry
                            adc     $48,%eax         // add carry to ascii 0
                            stosb                    // save in string
                            loop    .LPBinaryLoop
.LPBinaryNothing:           xor     %eax,%eax        // terminating CHR(0)
                            stosb
end ['EAX','EBX','ECX','EDX','EDI'];

FUNCTION PStrToBinary(P : PCHAR ;Bits : CARDINAL):CARDINAL; ASSEMBLER;
{Same as StrTobinary except for one INC %esi}

ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    P,%esi
                            movl    bits,%ecx
                            jecxz   .LPStrToBinNothing
.LPStrToBinNextbit:         shl     $1,%ebx
                            lodsb
                            and     $1,%eax       // 48 and 1=0 | 49 and 1=1
                            or      %eax,%ebx     // move into low bit
                            loop    .LPStrToBinNextbit
.LPStrToBinNothing:         mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];

FUNCTION PStrToHex(P : PChar;Digits : CARDINAL):CARDINAL; ASSEMBLER;

ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    P,%esi
                            movl    Digits,%ecx
                            jecxz   .LPStrToHexNothing

.LPStrToHexNextbit:         shl     $4,%ebx
                            xor     %eax,%eax
                            lodsb
                            subl    $48,%eax

                            cmp     $10,%eax
                            jb      .LPHexlow
                            subl    $7,%eax
                            and     $15,%eax        // safety.
        .LPHexlow:          addl    %eax,%ebx
                            loop    .LPStrToHexNextbit
.LPStrToHexNothing:          mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];

FUNCTION PStrToOct(P : PChar;Digits : CARDINAL):CARDINAL; ASSEMBLER;

ASM
                            xor     %ebx,%ebx     // start with empty value.
                            movl    P,%esi
                            movl    Digits,%ecx
                            jecxz   .LPStrToOctNothing

.LPStrToOctNextbit:         shl     $3,%ebx
                            xor     %eax,%eax
                            lodsb
                            subl    $48,%eax
                            and     $7,%eax        // safety.
        .LPOctlow:          addl    %eax,%ebx
                            loop    .LPStrToOctNextbit
.LPStrToOctNothing:          mov     %ebx,%eax     // set als return value.
end['EAX','EBX','ECX','ESI'];

PROCEDURE PHexToStr(P : PChar;Value,Digits : CARDINAL);ASSEMBLER;

ASM
                            movl    P,%edi

                            movl    Digits,%edx
                            mov     %edx,%ebx
                            shll    $2,%ebx
                            mov     $32,%ecx
                            sub     %ebx,%ecx        // calc unwanted Digits
                            movl    Value,%ebx
                            jecxz   .LPHexNoShl      // 32-Digits, test for 0
                            shll    %cl,%ebx         // Shift left to remove
                                                     // unwanted Digits.
.LPHexNoShl:                mov     %edx,%ecx
.LPHexLoop:                 mov     %ebx,%eax
                            shr     $28,%eax
                            addl    $48,%eax
                            cmp     $58,%eax
                            jb      .LPHexStore
                            addl    $7,%eax
.LPHexStore:                stosb
                            shl     $4,%ebx
                            loop    .LPHexLoop
.LPHexNothing:              xor     %eax,%eax
                            stosb
end ['EAX','EBX','ECX','EDX','EDI'];


PROCEDURE POctToStr(P : PChar;Value,Digits : CARDINAL);ASSEMBLER;

ASM
                            movl    P,%edi

                            movl    Digits,%edx
                            mov     %edx,%ebx
                            shll    $1,%ebx          // Digits*3
                            addl    %edx,%ebx
                            mov     $32,%ecx
                            sub     %ebx,%ecx        // calc unwanted Digits
                            movl    Value,%ebx
                            jecxz   .LPOctNoShl       // 32-Digits, test for 0
                            shll    %cl,%ebx         // Shift left to remove
                                                     // unwanted Digits.
.LPOctNoShl:                mov     %edx,%ecx
.LPOctLoop:                 mov     %ebx,%eax
                            shr     $29,%eax
                            addl    $48,%eax
                            stosb
                            shl     $3,%ebx
                            loop    .LPOctLoop
.LPOctNothing:              xor     %eax,%eax
                            stosb
end ['EAX','EBX','ECX','EDX','EDI'];
{$Endif Useasm}

end.

