{  This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Free for non-commercial use.

  Original header:
  -----------------
  Tools Library for DOS
  Minimal List Handling Unit
  BitSoft's Containers Library is a much more advanced solution.
  ----------------

  Originally by Brad Williams. Donated to XTDFPC/toolkit in August.
  I didn't know how to use objects at the time. Now I'm learning, and I liked
  the simple but effective approach of this unit the minute I saw it.
  Bitsoft is located at www.bitsoft.com, if you want the original package.

  }

UNIT Lists;


INTERFACE

USES Objects;

TYPE
          PQueue = ^TQueue;
          TQueue = OBJECT(TCollection)
                    CONSTRUCTOR Init;
                    PROCEDURE Clear; VIRTUAL;
                    PROCEDURE Error (Code, Info : Integer); VIRTUAL;
                    FUNCTION  Empty : Boolean;
                    PROCEDURE Enqueue (Item : Pointer); VIRTUAL;
                    FUNCTION  Remove : Pointer; VIRTUAL;
                   END; { of TQueue }

          PStack = ^TStack;
          TStack = OBJECT(TCollection)
                    CONSTRUCTOR Init;
                    PROCEDURE Clear; VIRTUAL;
                    FUNCTION  Empty : Boolean;
                    PROCEDURE Error (Code, Info : Integer); VIRTUAL;
                    FUNCTION  Pop : Pointer; VIRTUAL;
                    PROCEDURE Push (Item : Pointer); VIRTUAL;
                   END; { of TStack }

       { Integer pointers for putting an integer on a collection}
  PShortInt = ^ShortInt;
  PInteger  = ^Integer;
  PLongInt  = ^LongInt;
  PByte     = ^Byte;
  PWord     = ^Word;

       { Non-integers pointers for putting a non-integers on a collection }
  PReal     = ^Real;
  PSingle   = ^Single;
  PDouble   = ^Double;
  PExtended = ^Extended;
  PComp     = ^Comp;

{ Allocates a pointer on the stack, and fills it with the parameter value}

FUNCTION IntPtr (AnInteger : Integer) : PInteger;
FUNCTION LongIntPtr (ALongInt : LongInt) : PLongInt;
FUNCTION RealPtr (AReal : Real) : PReal;


IMPLEMENTATION

CONSTRUCTOR TQueue.Init;
  { The queue is initially empty and will grow by one item each time an }
  { item is added.                                                      }
BEGIN
  TCollection.Init(0,1);
END;

PROCEDURE TQueue.Clear;
BEGIN
  FreeAll;
  Pack;
END;

FUNCTION TQueue.Empty : Boolean;
BEGIN
  Empty := (Count = 0);
END;

PROCEDURE TQueue.Enqueue (Item : Pointer);
BEGIN
  Insert(Item);
END;

PROCEDURE TQueue.Error (Code, Info : Integer);
  { Only needs to respond to the coIndexError and coOverflow }
BEGIN
  IF Code = coOverflow
     THEN WriteLn('Insufficient Memory.  WVS queue error.')
     ELSE WriteLn('WVS queue range error.');
  Halt(1);
END;

FUNCTION TQueue.Remove : Pointer;
VAR Item : Pointer;
BEGIN
  Item := At(0);
  AtDelete(0);
  Pack;
  Remove := Item;
END;

CONSTRUCTOR TStack.Init;
BEGIN
  TCollection.Init(0,1);
END;

PROCEDURE TStack.Clear;
BEGIN
  FreeAll;
  Pack;
END;

FUNCTION TStack.Empty : Boolean;
BEGIN
  Empty := (Count = 0);
END;

PROCEDURE TStack.Error (Code, Info : Integer);
  { Only needs to respond to the coIndexError and coOverflow }
BEGIN
  IF Code = coOverflow
     THEN WriteLn('Insufficient Memory.  WVS stack overflow.')
     ELSE WriteLn('WVS stack range error.');
  Halt(1);
END;

FUNCTION TStack.Pop : Pointer;
VAR Item : Pointer;
BEGIN
  IF Count = 0
     THEN Error(coIndexError,0)
     ELSE BEGIN
            Item := At(Count-1);
            AtDelete(Count-1);
            Pack;
            Pop := Item;
          END;
END;

PROCEDURE TStack.Push (Item : Pointer);
BEGIN
  Insert(Item);
END;

FUNCTION IntPtr (AnInteger : Integer) : PInteger;
VAR NewPtr : PInteger;
BEGIN
  GetMem(NewPtr,SizeOf(Integer));
  NewPtr^ := AnInteger;
  IntPtr := NewPtr;
END;

FUNCTION LongIntPtr (ALongInt : LongInt) : PLongInt;
VAR NewPtr : PLongInt;
BEGIN
  GetMem(NewPtr,SizeOf(LongInt));
  NewPtr^ := ALongInt;
  LongIntPtr := NewPtr;
END;

FUNCTION RealPtr (AReal : Real) : PReal;
VAR NewPtr : PReal;
BEGIN
  NewPtr := New(PReal);
  NewPtr^ := AReal;
  RealPtr := NewPtr;
END;

END.
