Unit LZH;

{$i syseq.inc}

{$IFDEF FPC}
{$Asmmode ATT}
{$ENDIF}
{
 * LZHUF.C English version 1.0
 * Based on Japanese version 29-NOV-1988
 * LZSS coded by Haruhiko OKUMURA
 * Adaptive Huffman Coding coded by Haruyasu YOSHIZAKI
 * Edited and translated to English by Kenji RIKITAKE
 * Translated from C to Turbo Pascal by Douglas Webb   2/18/91
 *    Update and bug correction of TP version 4/29/91 (Sorry!!)
 * FPC Marco van de Voort Behaved badly under FPC Pascal. Turned out to be a
                          faulty implementation of Pred(x) and Succ(x) in
                          some versions of FPC (fixdate around Sept 1st '98)
                          Replaced succ and Pred by x-1 and x+1. Still TP
                          compatible I hope.
 * FPC MvdV               Working on a 32-bits InsertNode for speed, FPC global
                           var -Ratt/-Rdirect problem has to be fixed first.
 * FPC MvdV               Added some routines for in mem compression
     This Unit allows the user to commpress data using a combination of
   LZSS Compression and adaptive Huffman coding, or conversely to deCompress
   data that was previously Compressed by this Unit.

     There are a number of options as to where the data being Compressed/
   deCompressed is coming from/going to.

    In fact it requires that you pass the "LZHPack" Procedure 2 procedural
  parameter of Type 'GetProcType' and 'PutProcType' (declared below) which
  will accept 3 parameters and act in every way like a 'BlockRead'/'BlockWrite'
  Procedure call. Your 'GetProcType' Procedure should return the data
  to be Compressed, and Your 'PutProcType' Procedure should do something with
  the Compressed data (ie., put it in a File).  In Case you need to know (and
  you do if you want to deCompress this data again) the number of Bytes in the
  Compressed data (original, not Compressed size) is returned in 'Bytes_Written'.

  GetBytesProc = Procedure(Var DTA; NBytes:Word; Var Bytes_Got : Word);
  
  DTA is the start of a memory location where the inFormation returned should
  be.  NBytes is the number of Bytes requested.  The actual number of Bytes
  returned must be passed in Bytes_Got (if there is no more data then 0
  should be returned).

  PutBytesProc = Procedure(Var DTA; NBytes:Word; Var Bytes_Got : Word);

  As above except instead of asking For data the Procedure is dumping out
  Compressed data, do somthing With it.


    "LZHUnPack" is basically the same thing in reverse.  It requires
  procedural parameters of Type 'PutProcType'/'GetProcType' which
  will act as above.  'GetProcType' must retrieve data Compressed using
  "LZHPack" (above) and feed it to the unpacking routine as requested.
  'PutProcType' must accept the deCompressed data and do something
  withit.  You must also pass in the original size of the deCompressed data,
  failure to do so will have adverse results.


     Don't Forget that as procedural parameters the 'GetProcType'/'PutProcType'
  Procedures must be Compiled in the 'F+' state to avoid a catastrophe.

  note: All the large data structures For these routines are allocated when
  needed from the heap, and deallocated when finished.  So when not in use
  memory requirements are minimal.  However, this Unit Uses about 34K of
  heap space, and 400 Bytes of stack when in use. }

Interface

Type


  PutBytesProc = Procedure(Var DTA; NBytes : Word; Var Bytes_Put : Word);
  GetBytesProc = Procedure(Var DTA; NBytes : Word; Var Bytes_Got : Word);



Procedure LZHPack(Var Bytes_Written : LongInt;
                      GetBytes : GetBytesProc;
                      PutBytes : PutBytesProc);


Procedure LZHUnpack(TextSize : LongInt;
                    GetBytes : GetBytesProc;
                    PutBytes : PutBytesProc);

{ MvdV extension, predefined procedures which use above two procedures to
   (un)pack uncomprbyte uncompressed bytes from datain to dataout.

Seem to be working now.
}
FUNCTION Packer(VAR DataIn; VAR DataOut;Uncomprbyte:LONGINT):LONGINT;
PROCEDURE UnPacker(VAR DataIn; VAR DataOut;Uncomprbyte:LONGINT);

Implementation

Const
  Exit_OK = 0;
  Exit_FAILED = 1;

  { LZSS Parameters }
  N = 4096;                            { Size of String buffer }
  NPlusone=N+1;
  F = 60;                              { Size of look-ahead buffer }
  THRESHOLD = 2;
  NUL = N;                             { end of tree's node  }

  { Huffman coding parameters }
  N_Char = (256 - THRESHOLD + F);

  { Character code (:= 0..N_Char-1) }
  T = (N_Char * 2 - 1);                { Size of table }
  R = (T - 1);                         { root position }

  { update when cumulative frequency }
  { reaches to this value }
  MAX_FREQ = $8000;

{
 * Tables For encoding/decoding upper 6 bits of
 * sliding dictionary Pointer
 }

  { encoder table }
  p_len : Array[0..63] of Byte =
  ($03, $04, $04, $04, $05, $05, $05, $05,
   $05, $05, $05, $05, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08,
   $08, $08, $08, $08, $08, $08, $08, $08);

  p_code : Array[0..63] of Byte =
  ($00, $20, $30, $40, $50, $58, $60, $68,
   $70, $78, $80, $88, $90, $94, $98, $9C,
   $A0, $A4, $A8, $AC, $B0, $B4, $B8, $BC,
   $C0, $C2, $C4, $C6, $C8, $CA, $CC, $CE,
   $D0, $D2, $D4, $D6, $D8, $DA, $DC, $DE,
   $E0, $E2, $E4, $E6, $E8, $EA, $EC, $EE,
   $F0, $F1, $F2, $F3, $F4, $F5, $F6, $F7,
   $F8, $F9, $FA, $FB, $FC, $FD, $FE, $FF);

  { decoder table }
  d_code : Array[0..255] of Byte =
  ($00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00,
   $00, $00, $00, $00, $00, $00, $00, $00,
   $01, $01, $01, $01, $01, $01, $01, $01,
   $01, $01, $01, $01, $01, $01, $01, $01,
   $02, $02, $02, $02, $02, $02, $02, $02,
   $02, $02, $02, $02, $02, $02, $02, $02,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08,
   $09, $09, $09, $09, $09, $09, $09, $09,
   $0A, $0A, $0A, $0A, $0A, $0A, $0A, $0A,
   $0B, $0B, $0B, $0B, $0B, $0B, $0B, $0B,
   $0C, $0C, $0C, $0C, $0D, $0D, $0D, $0D,
   $0E, $0E, $0E, $0E, $0F, $0F, $0F, $0F,
   $10, $10, $10, $10, $11, $11, $11, $11,
   $12, $12, $12, $12, $13, $13, $13, $13,
   $14, $14, $14, $14, $15, $15, $15, $15,
   $16, $16, $16, $16, $17, $17, $17, $17,
   $18, $18, $19, $19, $1A, $1A, $1B, $1B,
   $1C, $1C, $1D, $1D, $1E, $1E, $1F, $1F,
   $20, $20, $21, $21, $22, $22, $23, $23,
   $24, $24, $25, $25, $26, $26, $27, $27,
   $28, $28, $29, $29, $2A, $2A, $2B, $2B,
   $2C, $2C, $2D, $2D, $2E, $2E, $2F, $2F,
   $30, $31, $32, $33, $34, $35, $36, $37,
   $38, $39, $3A, $3B, $3C, $3D, $3E, $3F);

  d_len : Array[0..255] of Byte =
  ($03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $03, $03, $03, $03, $03, $03, $03, $03,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $04, $04, $04, $04, $04, $04, $04, $04,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $05, $05, $05, $05, $05, $05, $05, $05,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $06, $06, $06, $06, $06, $06, $06, $06,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $07, $07, $07, $07, $07, $07, $07, $07,
   $08, $08, $08, $08, $08, $08, $08, $08,
   $08, $08, $08, $08, $08, $08, $08, $08);

  getbuf : Word = 0;
  getlen : Byte = 0;
  putlen : Byte = 0;
  putbuf : Word = 0;
  TextSize : LongInt = 0;
  codesize : LongInt = 0;
  printcount : LongInt = 0;
  match_position : Integer = 0;
  match_length : Integer = 0;


Type
  FreqType = Array[0..T] of Word; 
  FreqPtr = ^FreqType;
  PntrType = Array[0..(T + N_Char-1)] of Integer;
  pntrPtr = ^PntrType;
  SonType = Array[0..(T-1)] of Integer;
  SonPtr = ^SonType;
  TextBufType = Array[0..N + F - 2] of Byte;
  TBufPtr = ^TextBufType;
  WordRay = Array[0..N] of Integer;
  WordRayPtr = ^WordRay;
  BWordRay = Array[0..N + 256] of Integer;
  BWordRayPtr = ^BWordRay;

Var
  Text_buf : TBufPtr;
  lson, dad : WordRayPtr;
  rson : BWordRayPtr;
  freq : FreqPtr;                      { cumulative freq table }

{
 * pointing parent nodes.
 * area [T..(T + N_Char - 1)] are Pointers For leaves
 }
  prnt : pntrPtr;

  { pointing children nodes (son[], son[] + 1)}
  son : SonPtr;


  Procedure InitTree;                  { Initializing tree }
  Var
    i : Integer;
  begin
    For i := N + 1 to N + 256 do
      rson^[i] := NUL;                 { root }
    For i := 0 to N do
      dad^[i] := NUL;                  { node }
  end;

{ When -Ratt parser and/or the -Rdirect parsers don't have a problem with
   accessing global variables anymore, I want to develop some assembler
   versions of critical routines. Can somebody profile a program using
   this unit to show which other procedures are bottlenecks?}

{
PROCEDURE InsertNode2(r:word); ASSEMBLER;
ASM
                        {$IFDEF Framestack}
                          push   %ebp
                          mov    %esp,%ebp
                         {$ENDIF}
                         cld
                         subl   $8,%esp
                         lea    TEXT_BUF,%esi
                         mov    %esi,%edi
                         incl   %edi
                         movzwl 8(%ebp),%ebx
                         addl   %ebx,%esi
                         xor    %edx,%edx             // cmp:=0
                         xor    %eax,%eax             // convert byte to word (together with lodsb)
                                                      //  p := N + 1 + INTEGER(textbuf[r])

                         movb  (%esi),%al             // get textbuf[r], incrementing SI
                         add    $NPlusone,%eax             // Does this work, create something like CONST NPLUSONE=N+1
                         mov    %eax,-4(%ebp)         // save P
                         mov    rson,%eax
                         movw   $NUL,(%eax,%ebx,2)    // rson[r] := NUL
                         mov    lson,%eax
                         movw   $NUL,(%eax,%ebx,2)    // lson[r] := NUL
                         andw   $0,MATCH_LENGTH

                                                      //     top of main loop

.LInsLoop0:              movzwl 8(%ebp),%ebx          // maybe this can be removed
                         movl   -4(%ebp),%ecx
                         testb  $128,%dh                  // IF cmp>=0 THEN
                         jnz    .LInsLoop2
                         movl   rson,%eax
                         movzwl (%eax,%ecx,2),%eax      // IF rson[p]<>NILPTR THEN
                         cmp    $4096,%eax
                         jz     .LInsLoop1
                         movl   %eax,-4(%ebp)            // p := rson[p]
                         jmp    .LInsLoop4
.LInsLoop1:              lea    rson,%eax
                         jmp    .LInsLoop3Entry
.LInsLoop2:              mov    lson,%eax
                         movzwl (%eax,%ecx,2),%eax
                         cmp    $4096,%eax
                         jz     .LInsLoop3
                         movl   %eax,-4(%ebp)            // p := lson[p]
                         jmp    .LInsLoop4
.LInsLoop3:              mov    lson,%eax
.LInsLoop3Entry:         movw   %bx,(%eax,%ecx,2)        // lson[p] := r
                         mov    dad,%eax
                         movw   %cx,(%eax,%ebx,2)        // dad[r] := p
                         jmp    .LInsexit0               // another goto statement.
.LInsLoop4:              mov    $F-1,%ecx
                                                         // p already in %eax
                         pushl  %edi
                         addl   %eax,%edi                // DI := offset of textbuf[p+1]
                         popl   %eax
.LInsLoop5:              repe
                          cmpsb
                         movl   %eax,%edi
                         lahf
                         mov    %ah,%dh
                         je     .LInsLoop6
                         incl   %ecx
.LInsLoop6:              mov    $F,%eax
                         subl   %ecx,%eax
                         mov    %eax,%ecx
                         subl   %ecx,%esi
                         cmp    $THRESHOLD,%ecx
                         jbe    .LInsLoop8
                         movzwl MATCH_LENGTH,%ebx
                         cmp    %ebx,%ecx
                         jb     .LInsLoop8
                         cmp    %ebx,%ecx
                         jz     .LInsLoop7


.LInsLoop7:
.LInsLoop8:
.LInsexit0:                {$ifdef framestack}
                          leave
                         {$ENDIF}
END['EAX','ESI'];
}

Procedure InsertNode(R : Integer);   { Inserting node to the tree }
  Var
    tmp, i, p, cmp : Integer;
    key : TBufPtr;
    c : Word;
  begin
    cmp := 1;
    key := @Text_buf^[R];
    p := (N+1) + key^[0];
    rson^[R] := NUL;
    lson^[R] := NUL;
    match_length := 0;
    While match_length < F do
      begin
        if (cmp >= 0) then
          begin
            if (rson^[p] <> NUL) then
              p := rson^[p]
            else
              begin
                rson^[p] := R;
                dad^[R] := p;
                Exit;
              end;
          end
        else
          begin
            if (lson^[p] <> NUL) then
              p := lson^[p]
            else
              begin
                lson^[p] := R;
                dad^[R] := p;
                Exit;
              end;
          end;
        i := 0;
        cmp := 0;
        While (i < F) and (cmp = 0) do
          begin
            inc(i);
            cmp := key^[i] - Text_buf^[p + i];
          end;
        if (i > THRESHOLD) then
          begin
            tmp := ((R - p) and (N-1))-1;
            if (i > match_length) then
              begin
                match_position := tmp;
                match_length := i;
              end;
            if (match_length < F) and (i = match_length) then
              begin
                c := tmp;
                if (c < match_position) then
                  match_position := c;
              end;
          end;
      end;                             { While True do }
    dad^[R] := dad^[p];
    lson^[R] := lson^[p];
    rson^[R] := rson^[p];
    dad^[lson^[p]] := R;
    dad^[rson^[p]] := R;
    if (rson^[dad^[p]] = p) then
      rson^[dad^[p]] := R
    else
      lson^[dad^[p]] := R;
    dad^[p] := NUL;                    { remove p }
  end;


  Procedure DeleteNode(p : Integer);   { Deleting node from the tree }
  Var
    q : Integer;
  begin
    if (dad^[p] = NUL) then
      Exit;                            { unregistered }
    if (rson^[p] = NUL) then
      q := lson^[p]
    else if (lson^[p] = NUL) then
      q := rson^[p]
    else
      begin
        q := lson^[p];
        if (rson^[q] <> NUL) then
          begin
            Repeat
              q := rson^[q];
            Until (rson^[q] = NUL);
            rson^[dad^[q]] := lson^[q];
            dad^[lson^[q]] := dad^[q];
            lson^[q] := lson^[p];
            dad^[lson^[p]] := q;
          end;
        rson^[q] := rson^[p];
        dad^[rson^[p]] := q;
      end;
    dad^[q] := dad^[p];
    if (rson^[dad^[p]] = p) then
      rson^[dad^[p]] := q
    else
      lson^[dad^[p]] := q;
    dad^[p] := NUL;
  end;

  { Huffman coding parameters }

  Function GetBit(GetBytes : GetBytesProc) : Integer; { get one bit }
  Var
    i : Byte;
    i2 : Integer;
    oresult : Word;
  begin
    While (getlen <= 8) do
      begin
        GetBytes(i, 1, oresult);
        if oresult = 1 then
          i2 := i
        else i2 := 0;
        getbuf := getbuf or (i2 shl (8 - getlen));
        inc(getlen, 8);
      end;
    i2 := getbuf;
    getbuf := getbuf shl 1;
    dec(getlen);
    GetBit := Integer((i2 < 0));
  end;


  Function GetByte(GetBytes : GetBytesProc) : Integer; { get a Byte }
  Var
    j : Byte;
    i, oresult : Word;
  begin
    While (getlen <= 8) do
      begin
        GetBytes(j, 1, oresult);
        if oresult = 1 then
          i := j
        else
          i := 0;
        getbuf := getbuf or (i shl (8 - getlen));
        inc(getlen, 8);
      end;
    i := getbuf;
    getbuf := getbuf shl 8;
    dec(getlen, 8);
    GetByte := Integer(i shr 8);
  end;


  Procedure Putcode(l : Integer; c : Word;
                    PutBytes : PutBytesProc); { output c bits }
  Var
    Temp : Byte;
    Got : Word;
  begin
    putbuf := putbuf or (c shr putlen);
    inc(putlen, l);
    if (putlen >= 8) then
      begin
        Temp := putbuf shr 8;
        PutBytes(Temp, 1, Got);
        dec(putlen, 8);
        if (putlen >= 8) then
          begin
            Temp := lo(putbuf);
            PutBytes(Temp, 1, Got);
            inc(codesize, 2);
            dec(putlen, 8);
            putbuf := c shl (l - putlen);
          end
        else
          begin
            putbuf := putbuf shl 8;
            inc(codesize);
          end;
      end;
  end;


  { initialize freq tree }

  Procedure StartHuff;
  Var
    i, j : Integer;
  begin
    For i := 0 to (N_Char-1) do
      begin
        freq^[i] := 1;
        son^[i] := i + T;
        prnt^[i + T] := i;
      end;
    i := 0;
    j := N_Char;
    While (j <= R) do
      begin
        freq^[j] := freq^[i] + freq^[i + 1];
        son^[j] := i;
        prnt^[i] := j;
        prnt^[i + 1] := j;
        inc(i, 2);
        inc(j);
      end;
    freq^[T] := $ffff;
    prnt^[R] := 0;
  end;


  { reConstruct freq tree }

  Procedure reConst;
  Var
    i, j, k, tmp : Integer;
    F, l : Word;
  begin
    { halven cumulative freq For leaf nodes }
    j := 0;
    For i := 0 to (T-1) do
      begin
        if (son^[i] >= T) then
          begin
            freq^[j] := (1+freq^[i]) div 2; {@@ Bug Fix MOD -> div @@}
            son^[j] := son^[i];
            inc(j);
          end;
      end;
    { make a tree : first, connect children nodes }
    i := 0;
    j := N_Char;
    While (j < T) do
      begin
        k := (i+1);
        F := freq^[i] + freq^[k];
        freq^[j] := F;
        k := (j-1);
        While F < freq^[k] do
          dec(k);
        inc(k);
        l := (j - k) shl 1;
        tmp := (k+1);
        move(freq^[k], freq^[tmp], l);
        freq^[k] := F;
        move(son^[k], son^[tmp], l);
        son^[k] := i;
        inc(i, 2);
        inc(j);
      end;
    { connect parent nodes }
    For i := 0 to (T-1) do
      begin
        k := son^[i];
        if (k >= T) then
          begin
            prnt^[k] := i;
          end
        else
          begin
            prnt^[k] := i;
            prnt^[(k+1)] := i;
          end;
      end;
  end;


  { update freq tree }

  Procedure update(c : Integer);
  Var
    i, j, k, l : Integer;
  begin
    if (freq^[R] = MAX_FREQ) then
      begin
        reConst;
      end;
    c := prnt^[c + T];
    Repeat
      inc(freq^[c]);
      k := freq^[c];
      { swap nodes to keep the tree freq-ordered }
      l := (c+1);
      if (k > freq^[l]) then
        begin
          While (k > freq^[l]) do
            inc(l);
          dec(l);
          freq^[c] := freq^[l];
          freq^[l] := k;
          i := son^[c];
          prnt^[i] := l;
          if (i < T) then prnt^[(i+1)] := l;
          j := son^[l];
          son^[l] := i;
          prnt^[j] := c;
          if (j < T) then prnt^[j+1] := c;
          son^[c] := j;
          c := l;
        end;
      c := prnt^[c];
    Until (c = 0);                     { Repeat it Until reaching the root }
  end;


Var
  code, len : Word;

  Procedure EncodeChar(c : Word; PutBytes : PutBytesProc);
  Var
    i : Word;
    j, k : Integer;
  begin
    i := 0;
    j := 0;
    k := prnt^[c + T];
    { search connections from leaf node to the root }
    Repeat
      i := i shr 1;
 {      if node's address is odd, output 1
        else output 0
        }
      if Boolean(k and 1) then inc(i, $8000);
      inc(j);
      k := prnt^[k];
    Until (k = R);
    Putcode(j, i, PutBytes);
    code := i;
    len := j;
    update(c);
  end;


  Procedure EncodePosition(c : Word; PutBytes : PutBytesProc);
  Var
    i, j : Word;
  begin
    { output upper 6 bits With encoding }
    i := c shr 6;
    j := p_code[i];
    Putcode(p_len[i], j shl 8, PutBytes);
    { output lower 6 bits directly }
    Putcode(6, (c and $3f) shl 10, PutBytes);
  end;


  Procedure Encodeend(PutBytes : PutBytesProc);
  Var
    Temp : Byte;
    Got : Word;
  begin
    if Boolean(putlen) then
      begin
        Temp := lo(putbuf shr 8);
        PutBytes(Temp, 1, Got);
        inc(codesize);
      end;
  end;


  Function DecodeChar(GetBytes : GetBytesProc) : Integer;
  Var
    c : Word;
  begin
    c := son^[R];
    {
     * start searching tree from the root to leaves.
     * choose node #(son[]) if input bit = 0
     * else choose #(son[]+1) (input bit = 1)
    }
    While (c < T) do
      begin
        c := c + GetBit(GetBytes);
        c := son^[c];
      end;
    c := c - T;
    update(c);
    DecodeChar := Integer(c);
  end;


  Function DecodePosition(GetBytes : GetBytesProc) : Word;
  Var
    i, j, c : Word;
  begin
    { decode upper 6 bits from given table }
    i := GetByte(GetBytes);
    c := Word(d_code[i] shl 6);
    j := d_len[i];
    { input lower 6 bits directly }
    dec(j, 2);
    While j <> 0 do
      begin
        i := (i shl 1) + GetBit(GetBytes);
        dec(j);
      end;
    DecodePosition := c or i and $3f;
  end;


  { Compression }

  Procedure InitLZH;
  begin
    getbuf := 0;
    getlen := 0;
    putlen := 0;
    putbuf := 0;
    TextSize := 0;
    codesize := 0;
    printcount := 0;
    match_position := 0;
    match_length := 0;
    new(lson);
    new(dad);
    new(rson);
    new(Text_buf);
    new(freq);
    new(prnt);
    new(son);
  end;


  Procedure endLZH;
  begin
    dispose(son);
    dispose(prnt);
    dispose(freq);
    dispose(Text_buf);
    dispose(rson);
    dispose(dad);
    dispose(lson);
  end;


  Procedure LZHPack(Var Bytes_Written : LongInt;
                        GetBytes : GetBytesProc;
                        PutBytes : PutBytesProc);
  Var
    ct : Byte;
    i, len, R, s, last_match_length : Integer;
    Got : Word;
  begin
    InitLZH;
    TextSize := 0;                     { rewind and rescan }
    StartHuff;
    InitTree;
    s := 0;
    R := N - F;
    fillChar(Text_buf^[0], R, ' ');
    len := 0;
    Got := 1;
    While (len < F) and (Got <> 0) do
      begin
        GetBytes(ct, 1, Got);
        if Got <> 0 then
          begin
            Text_buf^[R + len] := ct;
            inc(len);
          end;
      end;
    TextSize := len;
    For i := 1 to F do
      InsertNode(R - i);
    InsertNode(R);
    Repeat
      if (match_length > len) then
        match_length := len;
      if (match_length <= THRESHOLD) then
        begin
          match_length := 1;
          EncodeChar(Text_buf^[R], PutBytes);
        end
      else
        begin
          EncodeChar(255 - THRESHOLD + match_length, PutBytes);
          EncodePosition(match_position, PutBytes);
        end;
      last_match_length := match_length;
      i := 0;
      Got := 1;
      While (i < last_match_length) and (Got <> 0) do
        begin
          GetBytes(ct, 1, Got);
          if Got <> 0 then
            begin
              DeleteNode(s);
              Text_buf^[s] := ct;
              if (s < (F-1)) then
                Text_buf^[s + N] := ct;
              s := (s+1) and (N-1);
              R := (R+1) and (N-1);
              InsertNode(R);
              inc(i);
            end;
        end;
      inc(TextSize, i);
      While (i < last_match_length) do
        begin
          inc(i);
          DeleteNode(s);
          s := (s+1) and (N-1);
          R := (R+1) and (N-1);
          dec(len);
          if Boolean(len) then InsertNode(R);
        end;
    Until (len <= 0);
    Encodeend(PutBytes);
    endLZH;
    Bytes_Written := TextSize;
  end;


  Procedure LZHUnpack(TextSize : LongInt;
                      GetBytes : GetBytesProc;
                      PutBytes : PutBytesProc);
  Var
    c, i, j, k, R : Integer;
    c2    : Byte;
    count : LongInt;
    Put : Word;
  begin
    InitLZH;
    StartHuff;
    R := N - F;
    fillChar(Text_buf^[0], R, ' ');
    count := 0;
    While count < TextSize do
      begin
        c := DecodeChar(GetBytes);
        if (c < 256) then
          begin
            c2 := lo(c);
            PutBytes(c2, 1, Put);
            Text_buf^[R] := c;
            inc(R);
            R := R and (N-1);
            inc(count);
          end
        else
          begin
            i := (R - (1+DecodePosition(GetBytes))) and (N-1);
            j := c - 255 + THRESHOLD;
            For k := 0 to (j-1) do
              begin
                c := Text_buf^[(i + k) and (N-1)];
                c2 := lo(c);
                PutBytes(c2, 1, Put);
                Text_buf^[R] := c;
                inc(R);
                R := R and (N-1);
                inc(count);
              end;
          end;
      end;
    endLZH;
  end;

{----------- Addition by Marco van de Voort, predefined functions for
             in-memory compression -----------}

{$IFDEF FPC}
TYPE BytePtr=^BYTE;
{$ELSE}
TYPE   ByteAr=ARRAY[0..65000] OF BYTE;
TYPE BytePtr=^ByteAr;
{$ENDIF}

VAR InData,OutData : BytePtr;
    ToRead,
    ReadPnt,WritePnt : CARDINAL;

Procedure GetBlock(Var Target; NoBytes : Word; Var Actual_Bytes : Word); FAR;

Var
    BytesLeft: LONGINT;

BEGIN
   BytesLeft:=ToRead-ReadPnt;
   IF Longint(Nobytes)>=BytesLeft THEN
    BEGIN
     Actual_Bytes:=BytesLeft;
     IF Actual_Bytes>0 THEN
      BEGIN
       {$IFDEF FPC}
        Move(InData[ReadPnt], Target, Actual_Bytes);
       {$ELSE}
        Move(InData^[ReadPnt], Target, Actual_Bytes);
       {$ENDIF}
       ReadPnt:=ToRead;
      END;
    END
   ELSE
    BEGIN
       {$IFDEF FPC}
        Move(InData[ReadPnt], Target, NoBytes);
       {$ELSE}
        Move(InData^[ReadPnt], Target, NoBytes);
       {$ENDIF}
     Actual_Bytes:=Nobytes;
     Inc(ReadPnt,Nobytes);
    END;
end;

PROCEDURE PutBlock(Var Source; NoBytes : Word; Var Actual_Bytes : Word); FAR;

BEGIN
  {$IFDEF Turbo}
   Move(Source,Outdata^[Writepnt],nobytes);
  {$ELSE}
   Move(Source,Outdata[Writepnt],nobytes);
  {$ENDIF}
  Actual_bytes:=Nobytes;
  inc(Writepnt,Nobytes);
end;

FUNCTION Packer(VAR DataIn; VAR DataOut;Uncomprbyte:LONGINT):LONGINT;

BEGIN
 InData:=@DataIn;
 OutData:=@DataOut;
 ReadPnt:=0;
 ToRead:=Uncomprbyte;
 WritePnt:=0;
 LZHPack(Uncomprbyte, @GetBlock, @PutBlock);
 Packer:=WritePnt;
END;

PROCEDURE UnPacker(VAR DataIn; VAR DataOut;Uncomprbyte:LONGINT);

BEGIN
 InData:=@DataIn;
 OutData:=@DataOut;
 ReadPnt:=0;
 ToRead:=uncomprbyte;
 WritePnt:=0;
 LZHUnPack(Uncomprbyte, @GetBlock, @PutBlock);
END;

end.

