UNIT ODirTree;

INTERFACE

{ ODirTree, directory search systems, OOP version. NOT BP COMPATIBLE (CLASSES)

   This is a unit of the XTDLIB/FPC package
   COPYRIGHT (C) 1993..1998 by Marco van de Voort,  All Rights Reserved
   Free for non-commercial use. See license.txt.
   (www.stack.nl/~marcov/xtdlib.htm)
   Free for non-commercial use.

Module searches a path and below for a file ( * and ? allowed),
extended wilcards only when LFN is true (Linux or LFN under Win32)
a system to read a directory tree into memory (the module name already
refers to it) is under development, but fully working. (NOT in OOP version yet.

The module is no real speedmonster, but the speed seems satisfactory. I think
you can't get it much faster, not with FindFirst and FindNext at the basis.

I need it because I made a CD of all my software, and I wanted to create
indexes.

Demoes : TreeDemo, Indexer,CdSize and FileFind. (Some not yet in XTDLIB/FPC)

}

{$I syseq.inc}
Uses Dos;
{ Dos.PathStr and Dos.SearchRec are used in interface}



TYPE

{----------------------    TFileScan class(es) --------------------------}

  TFileScan   = CLASS
		  CONSTRUCTOR FileScanInit;
		 PRIVATE
		  NoDirAttr  : BYTE;    {Parameters to FindFirst/Next pairs AND NOT Directory}
		  DirAttr    : BYTE;    {Parameters to FindFirst/Next pairs OR Directory}
		  Level      : LONGINT; { Recursion level is readonly via GetLevel}
		  PROCEDURE SearchCurrent(CONST Dir : String;Target : PathStr);
		  PROCEDURE SearchSubDirs(CONST Dir : PathStr;CONST Target : PathStr);
		 PUBLIC
		  DirsToo    : BOOLEAN; { Object parameter, should a matching DIRECTORYname be reported to FileFound?
						Default : NO (FALSE) }
		  { Statistics over last search with the current object are kept here.}
		  FoundCount   : LONGINT; { Nr. files found }
		  TotalBytes   : LONGINT; { Total bytes in files found }
		  FoundDirs    : LONGINT; { Directories found}
		  ClusterSize  : LONGINT; { Clustersize used for rounding,
					   default 0= no rounding }

		  PROCEDURE ClearStat; { Clears statistics, all above vars to initial value}
		  PROCEDURE DirEnter  (Path : PathStr); VIRTUAL; ABSTRACT;
		  PROCEDURE DirExit   (Path : PathStr); VIRTUAL; ABSTRACT;
		  FUNCTION  Detect    (Search:SearchRec;P:PathStr):BOOLEAN; VIRTUAL; ABSTRACT;
		  PROCEDURE FileFound (Path : PathStr;Search : SearchRec); VIRTUAL; ABSTRACT;
		  PROCEDURE FileScan  (RootDir,FileName : STRING);
		  PROCEDURE SetAttr   (Attr : BYTE);
		  FUNCTION  GetLevel : LONGINT;	{get recursionlevel}
		  END;

{----------------------    TTree class(es) generic types ------------------------}

      FilAttr       = BYTE;

      Fileptr       = ^FilesRec;
      FilesRec       = RECORD
		       Next     : FilePtr;
		       DirE     : SearchRec;
		       UserData : POINTER;
		       END;


      PDirTree  = ^DirTreeRecord;
      DirTreeRecord = RECORD
		       NextDir,
		       SubDirs  : PDirTree;
		       Name     : PathStr;
		       DirE     : SearchRec;
		       Files    : FilePtr;
		       UserData : POINTER;
		       END;


{----------------------    Basic TTreeBuild Class     ------------------------}


      TTreeBuild    = CLASS
		       CONSTRUCTOR TreeSearchInit;
		       DESTRUCTOR  TreeSearchFree;     { override class.destroy?}
		      PRIVATE
		       NoDirAttr  : BYTE;    {Parameters to FindFirst/Next pairs AND NOT Directory}
		       DirAttr    : BYTE;    {Parameters to FindFirst/Next pairs OR Directory}
		       Level      : LONGINT; { Recursion level is readonly via GetLevel}
		       Root       : PDirTree;
		       PROCEDURE    TreeSearch  (CONST Dir : PathStr;NRoot:PDirTree);

		       PROCEDURE    KillFileTree(VAR RootIn:PDirTree);
		      PUBLIC
		       DirsToo    : BOOLEAN; { Object parameter, should a matching DIRECTORYname be reported to FileFound?
						Default : NO (FALSE) }
		     { Statistics over last search with the current object are kept here.}

		       FoundFiles   : LONGINT; { Nr. files found }
		       TotalBytes   : LONGINT; { Total bytes in files found }
		       FoundDirs    : LONGINT; { Directories found}
		       ClusterSize  : LONGINT; { Clustersize used for rounding,
						    default 0= no rounding }

		       PROCEDURE SetAttr   (Attr : BYTE); {Set attributes for FindFirst}
		       FUNCTION  GetLevel : LONGINT;	{get recursionlevel}

		       PROCEDURE ClearStat; { Clears above statistics variables}
		       PROCEDURE ResetClass; { Class init code. Resets all variables to default}

		       FUNCTION  CreateObject       (Path : PathStr;Search: SearchRec):POINTER; VIRTUAL; ABSTRACT;
		       PROCEDURE FreeObject       (VAR Obj : POINTER); VIRTUAL; ABSTRACT;

		       {Creates an tree. Uses Detect to select which files are added.
			Runs SearchDirEnter on every entery of the dir, and SearchDirExit
			 on every exit.}
		       PROCEDURE BuildTree (RootDir: String);
		       PROCEDURE BuildDirEnter  (Path : PathStr); VIRTUAL; ABSTRACT;
		       FUNCTION  BuildDetect    (Search:SearchRec;P:PathStr):BOOLEAN; VIRTUAL; ABSTRACT;
		       PROCEDURE BuildDirExit   (Path : PathStr); VIRTUAL; ABSTRACT;
		      END;


{----------------------    Derivative TTreeScan Class     ----------------------}


    TTreeScan= CLASS(TTreeBuild)
		  PRIVATE
		   PROCEDURE Scan1Tree(LocRoot : PDirTree);
		  PUBLIC
		   ReportSubDir : BOOLEAN;
		   PROCEDURE ScanTree;
		   { Methods to use when scanning a built tree}

		   PROCEDURE ScanDirEnter  (Path : PathStr); VIRTUAL; ABSTRACT;
		   PROCEDURE FileFound (Path : PathStr;VAR Search : SearchRec;VAR UserPtr:POINTER); VIRTUAL; ABSTRACT;
		   PROCEDURE ScanDirExit   (Path : PathStr); VIRTUAL; ABSTRACT;
		  END;


{----------------------    Derivative TTreeAddFiles Class     -------------------}

{ Still Under (De)construction}

   TTreeAddFiles   = CLASS(TTreeScan)       { TTreeBuild also possible if you
					     don't use predef'ed scan methods}

		      PUBLIC
		       PROCEDURE AddDirEnter  (Path : PathStr); VIRTUAL; ABSTRACT;
		       FUNCTION  AddDetect    (Search:SearchRec;P:PathStr):BOOLEAN; VIRTUAL; ABSTRACT;
		       PROCEDURE AddDirExit   (Path : PathStr); VIRTUAL; ABSTRACT;

		       PROCEDURE AddFiles(Pattern :String);
		     END;



IMPLEMENTATION


{$ifdef FPC}
Uses EPasStr,Strings;
{$ELSE}
 Uses Strings;
{$endif FPC}

PROCEDURE TFileScan.SetAttr (Attr:BYTE);

BEGIN
 DirAttr:=Attr OR Directory;	   {Matching directories}
 NoDirAttr:=Attr AND (NOT Directory);  {Matching files}
END; {SetAttr}


PROCEDURE TFileScan.ClearStat;

BEGIN
 SetAttr(AnyFile-Volumeid);
 DirsToo:=FALSE;
 FoundCount:=0;
 TotalBytes:=0;
 FoundDirs :=0;
 ClusterSize:=0;
END;

CONSTRUCTOR TFileScan.FileScanInit;

BEGIN
 ClearStat;
END;

PROCEDURE TFileScan.SearchCurrent(CONST Dir : String;Target : PathStr);

VAR

  FoundFile : SearchRec;

BEGIN
  FindFirst(Dir+Target,DirAttr, FoundFile);
  WHILE DosError=0 DO
   BEGIN
      IF ((FoundFile.attr AND Directory)=0) OR DirsToo THEN
	BEGIN
	  INC(FoundCount);
	  IF ClusterSize=0 THEN
	   INC(TotalBytes, FoundFile.size)
	  ELSE
	   INC(TotalBytes, ((FoundFile.size DIV ClusterSize)+1)*ClusterSize);
	   FileFound(Dir,FoundFile);
	 END
	ELSE
	  IF FoundFile.Name[1] <>'.' THEN
	   INC(FoundDirs);
       FindNext(FoundFile);
    END; {read loop}
 {$IFDEF CloseFind}
  FindClose(FoundFile);
 {$ENDIF}
END; {SearchCurrent}

PROCEDURE TFileScan.SearchSubDirs(CONST Dir : PathStr;CONST Target : PathStr);
VAR
  FoundDir  : SearchRec; { Current search data. }
  FileSpec  : String;  { Always contains "DIR" + "*." }
  Path      : String;

BEGIN
 DirEnter(Dir);
 SearchCurrent(Dir,Target);
 FileSpec:=concat(Dir,'*');
 FindFirst(FileSpec,DirAttr, FoundDir);
 WHILE DosError=0 DO
  BEGIN
     WITH FoundDir DO
      BEGIN
       IF Name[1] <> '.' THEN
    IF ((FoundDir.Attr AND directory)<>0) {AND ((FoundDir.Attr AND Hidden)=0)} THEN
	 BEGIN
	 {$ifdef Linux}
	  Path:=Dir+FoundDir.Name+'/';
	 {$else}
	  Path:=Dir+FoundDir.Name+'\';
	 {$endif}
      INC(Level);
	  SearchSubDirs(Path ,Target);
      DEC(Level);
	END;
      END; {WITH FoundDir}
     FindNext(FoundDir);
   END; {While loop}
 {$IFDEF CloseFind}
  FindClose(FoundDir);
 {$ENDIF}
 DirExit(Dir);
END;{ SearchSubDirs}

PROCEDURE TFileScan.FileScan(RootDir,FileName : String);

VAR Dir    : PathStr;
    Target : PathStr;
    SRec   : SearchRec;
    IsRoot : BOOLEAN;

BEGIN
 Dir:=RootDir;
 Target:=FileName;
 {$ifdef FPC}
   AppendBackSlash(Dir); {procedure is safe for linux}
 {$else}
   {Borland doesn't compile for Linux, so a '\' is no problem,
     other compilers not supported :-)}
 IF Dir[ORD(Dir[0])] <> '\' THEN
  Dir:=Concat(Dir,'\');
 {$endif FPC}
 Level:=0;
 IF Target='.' THEN
  BEGIN
   IsRoot:=FALSE;
   FindFirst(Dir+'.',anyfile, SRec);
   IsRoot:=(Doserror<>0);
   {$IFDEF CloseFind}
    FindClose(SRec);
   {$ENDIF}
   IF IsRoot THEN
    BEGIN
     FillChar(SRec,SIZEOF(SearchRec),0);
     SRec.Name:='';
     FileFound(Dir,SRec);
    END;
  END;
 SearchSubDirs(Dir,Target);
END;

FUNCTION TFileScan.GetLevel:LONGINT;

BEGIN
 GetLevel:=Level;
END;

{----------------------    TTreeBuild	   --------------------------}

PROCEDURE TTreeBuild.SetAttr (Attr:BYTE);

BEGIN
 DirAttr:=Attr OR Directory;	   {Matching directories}
 NoDirAttr:=Attr AND (NOT Directory);  {Matching files}
END; {SetAttr}


{ Internal procedure, but in fact the main procedure the other part (BuildTree)
 is mere interfacial}
PROCEDURE TTreeBuild.TreeSearch  (CONST Dir : PathStr;NRoot:PDirTree);
VAR
  FoundDir  : SearchRec; { Current search data. }
  FileSpec  : PathStr;  { Always contains "DIR" + "*." }
  Cur       : PDirTree;
  FCur      : FilePtr;
  FirstDir  : BOOLEAN;
  FirstFil  : BOOLEAN;
BEGIN
 BuildDirEnter(Dir);
 FileSpec:=concat(Dir,'*');
 FindFirst(FileSpec,DirAttr,FoundDir);
 FirstDir:=TRUE;
 FirstFil:=TRUE;
 Cur:=NRoot;
 WHILE DosError=0 DO
   BEGIN
    IF (FoundDir.Attr and directory)=0  THEN
     BEGIN
      IF BuildDetect(FoundDir,Dir) THEN
       BEGIN
	IF FirstFil THEN
	 BEGIN
	  NEW(NRoot^.Files);
	  FCur:=NRoot^.Files;
	  FirstFil:=FALSE;
	 END
	ELSE
	 BEGIN
	  NEW(FCur^.Next);
	  FCur:=FCur^.Next;
	 END;
	FCur^.Next:=NIL;
	FCur^.DirE:=FoundDir;
	FCur^.UserData:=CreateObject(Cur^.Name,FoundDir);
	INC(FoundFiles);
	INC(TotalBytes,FoundDir.size);
       END;
     END
    ELSE
     IF (FoundDir.Name <>'.') AND (FoundDir.Name<>'..') THEN
      BEGIN
       INC (FoundDirs);
       IF FirstDir THEN
	BEGIN
	 NEW(Cur^.SubDirs);
	 Cur:=Cur^.SubDirs;
	 FirstDir:=FALSE;
	END
       ELSE
	BEGIN
	 NEW(Cur^.NextDir);
	 Cur:=Cur^.NextDir;
	END; {IF FIrst}
       Cur^.SubDirs:=NIL;
       Cur^.NextDir:=NIL;
       Cur^.Files  :=NIL;
       Cur^.DirE:=FoundDir;
       Cur^.Name:=Dir+FoundDir.Name+'/';
       Cur^.UserData:=CreateObject(Cur^.Name,FoundDir);
       INC(Level);
       TreeSearch(Cur^.Name ,Cur);
       DEC(Level);
      END; {Attr and directory}
    FindNext(FoundDir);
   END; {read loop}
 {$IFDEF CloseFind}
  FindClose(FoundDir);
 {$ENDIF}
 BuildDirExit(Dir);
END; {TreeSearch}

PROCEDURE TTreeBuild.BuildTree(RootDir:String);

VAR Dummy : String;

BEGIN
 Dummy:=RootDir;	{ RootDir is passed to a CONST procedure!}
 AppendBackSlash(Dummy); {procedure is safe for linux}
 FoundDirs:=0;
 FoundFiles:=0;
 IF Root <> NIL THEN
  KillFileTree(Root);
 NEW(Root);
 FillChar(Root^,SIZEOF(DirTreeRecord),0);
 Root^.Name:=Dummy;
 Level:=0;
 TreeSearch(Dummy,Root);
END;

PROCEDURE TTreeBuild.KillFileTree(VAR RootIn:PDirTree);

PROCEDURE Kill1Tree(Root : PDirTree);

VAR
    CurrentFile,
    CurrentFile2 : FilePtr;

BEGIN
  IF Root^.Files <> NIL THEN
   BEGIN
     CurrentFile:=Root^.Files;
     WHILE(CurrentFile^.Next <> NIL) DO
       BEGIN
	 CurrentFile2:=CurrentFile;
	 CurrentFile:=CurrentFile^.Next;
	 IF CurrentFile2^.UserData<>NIL THEN
	  FreeObject(CurrentFile2^.UserData);
	 DISPOSE(CurrentFile2);
       END;
     IF CurrentFile^.UserData<>NIL THEN
      FreeObject(CurrentFile^.UserData);
     DISPOSE(CurrentFile);
     Root^.Files:=NIL;
   END;
  IF Root^.SubDirs <> NIL THEN
   BEGIN
     Kill1Tree(Root^.SubDirs);
     Root^.SubDirs:=NIL;
   END;
  IF Root^.NextDir <> NIL THEN
   BEGIN
     Kill1Tree(Root^.NextDir);
     Root^.NextDir:=NIL;
   END;
  IF Root^.UserData<>NIL THEN
      FreeObject(Root^.UserData);
  DISPOSE(Root);
END; { Kill1Tree}


BEGIN
 Kill1Tree(RootIn);
END;{KillFileTree}

PROCEDURE TTreeBuild.ClearStat; { Initcode EDirTree in procedure form, so it can be used
			to re-init module}

BEGIN
 FoundFiles:=0;	 TotalBytes:=0;
 FoundDirs:=0;
 Level:=0;
END;

PROCEDURE TTreeBuild.ResetClass; { Initcode EDirTree in procedure form, so it can be used
			to re-init module}

BEGIN
 ClearStat;
 ClusterSize:=0;
 DirsToo:=FALSE; { Default : File search only }
 SetAttr(anyfile-volumeid);
 IF Root<>NIL THEN
  KillFileTree(Root);
END;

CONSTRUCTOR TTreeBuild.TreeSearchInit;

BEGIN
 Root:=NIL;
 ResetClass;
END;

DESTRUCTOR TTreeBuild.TreeSearchFree;

BEGIN
 IF Root<>NIL THEN
  KillFileTree(Root);
END;

FUNCTION  TTreeBuild.GetLevel : LONGINT;	{get recursionlevel}

BEGIN
 GetLevel:=Level;
END;

{----------------------    TTreeScan	    --------------------------}

PROCEDURE TTreeScan.Scan1Tree(LocRoot : PDirTree);

VAR
    CurrentFile :     FilePtr;
    CurrentDir	:     PDirTree;

BEGIN
  ScanDirEnter(LocRoot^.Name);
  IF DirsToo THEN
   BEGIN
    IF ReportSubDir THEN
     BEGIN
      IF (LocRoot^.SubDirs<>NIL) THEN
       BEGIN
	CurrentDir:=LocRoot^.SubDirs;
	REPEAT
	 FileFound(CurrentDir^.Name,CurrentDir^.DirE,CurrentDir^.UserData);
	 CurrentDir:=CurrentDir^.NextDir;
	UNTIL CurrentDir=NIL;
      END;
     END
    ELSE
     FileFound (LocRoot^.Name,LocRoot^.DirE,LocRoot^.UserData);
   END;
  IF LocRoot^.Files <> NIL THEN
   BEGIN
    CurrentFile:=LocRoot^.Files;
    FileFound (LocRoot^.Name,CurrentFile^.DirE,CurrentFile^.UserData);
    WHILE CurrentFile^.Next <> NIL DO
     BEGIN
      CurrentFile:=CurrentFile^.Next;
      FileFound (LocRoot^.Name,CurrentFile^.DirE,CurrentFile^.UserData);
     END;
   END;
  IF LocRoot^.SubDirs <> NIL THEN
   BEGIN
    INC(Level);
    Scan1Tree(LocRoot^.SubDirs);
    DEC(Level);
   END;
  ScanDirExit(LocRoot^.Name);
  IF LocRoot^.NextDir <> NIL THEN
    Scan1Tree(LocRoot^.NextDir);

END;{ Scan1Tree;}


PROCEDURE TTreeScan.ScanTree;

BEGIN
 Level:=0;
 Scan1Tree(Root);
END;{ ScanTree}

{----------------------    TTreeAdded	   --------------------------}

PROCEDURE TTreeAddFiles.AddFiles(Pattern :String);

PROCEDURE SearchForFile(DirPtr:PDirTree);

VAR D  : SearchRec;
    Current    : FilePtr;
    FileSpec   : PathStr;
    First      : BOOLEAN;

BEGIN
 AddDirEnter(DirPtr^.Name);
 FileSpec:=DirPtr^.Name+Pattern;
 First:=TRUE;
 FindFirst(FileSpec,DirAttr,D);
 IF DosError=0 THEN
   REPEAT
     IF AddDetect(D,DirPtr^.Name) THEN
       BEGIN
	IF First THEN  (* Check for files already appended to the tree *)
	 BEGIN
	   IF DirPtr^.Files = NIL THEN
	     BEGIN
	       NEW(DirPtr^.Files);
	       Current:=DirPtr^.Files;
	     END
	   ELSE
	     BEGIN
	       Current:=DirPtr^.Files;
	       WHILE Current^.Next <> NIL DO
		 Current:=Current^.Next;
	       NEW(Current^.Next);
	       Current:=Current^.Next;
	     END;
	  First:=FALSE;
	END
      ELSE
	BEGIN
	  NEW(Current^.Next);
	  Current:=Current^.Next;
	END;
     Current^.Next:=NIL;
     Current^.DirE:=D;
     INC(FoundFiles);
     INC(TotalBytes,D.size);
   END;
   FindNext(D);
  UNTIL DOSERROR<>0;
 {$IFDEF CloseFind}
  FindClose(D);
 {$ENDIF}
 INC(Level);
 IF DirPtr^.SubDirs <> NIL THEN
  SearchForFile(DirPtr^.SubDirs);
 DEC(Level);
 AddDirExit(DirPtr^.Name);
 IF DirPtr^.NextDir <> NIL THEN     {Do next dir}
  SearchForFile(DirPtr^.NextDir);
END;

BEGIN
 IF ROOT=NIL THEN
  BEGIN
   Writeln('!! ODirTree.TTreeAddFiles.AddFiles called without initializing the tree with BuildTree');
   HALT;
  END;
 FoundFiles:=0;
 TotalBytes:=0;
 Level     :=0;
 SearchForFile(Root);
END;


END.

{

17-02-99  + Start in file logging for EDirTree/ODirTree
	  * Fixed major bug. When target2='.' then the root directory (which
	     doesn't have a . directory) wasn't scanned.
19-02-99  + Directories where scanned with *.* as filespec. Is now * (Linux and
	     LFN dirs with points in it)

22-02-99  - EDirTree copied to ODirTree, starting experimenting with OOP EDirtree
	  - Initial ODirTree working. haven't figured out details. Main problem
	      was OVERRIDE directive not being mentioned in manual.
	  + Levels (GetLevel) implementation. Handy for if you have a file per
	      directory level. (e.g. 00_index.txt), though a system based on
	      the tree building would be handier.
	    Odirtree is a bit bigger than EDirTree (in lines), but has some more
	    functionality (
	       - more searches at once possible because each instance it
			totally isolated,
	       - DirEnter, DirExit methods (4 lines only)
	       - Levels
	       - Everything is dynamical
	       - Also it's possible to put more data and methods in the
		  descendant object (like MainF in the indexer example).
	  * Fixed some problems with FindClose. After that, I included syseq.inc
	      to define FindClose.

27-02-99  - Initial untested Tree Classes. TTreeAddFiles N/I, several small
	     features (level etc) in other Tree classes N/I or incomplete.
	  * Several small fixes to ODirTree. Main fix was USEing ObjPas.
	  - Initial system seems to be working (some small cosmetic things,
	      but looks nice)
	      Also HeaptRC (and own memavail based test) doesn't seem to have
	      any unallocated mem. Cool! (but not really shocking. EDirtree
	      was already quite memory-safe)
	  * Fixed Levels for BuildTree method.
	  # Started with some simple docs. (Copying EDirTree.htm to ODirTree.htm)
	  - Implemented TTreeAddFiles class. Not tested yet.
	  * Some small fixes in levels and recursion behaviour of scantree.
	  # I've two different scenarios for DirsToo support in ScanTree.
		- One, reporting the currentdir to FileFound.
		- Two, reporting all subdirs of current dir to FileFound.
		Both have some advantages
	  + Fixed the above problem by allowing them both, depending on a boolean.
		}

