UNIT WinProc;
{ Window dependant, non PullDown routines, simple selector window procedures
   and array editting primitives. (C) Marco van de Voort and
                                           Andreas Reinicke 1989...1997
   Part of the XTDLIB package. Leftovers from tlib, revived and some procedures
   in the same style added

  Quite some overhead, for small dialogue screens 2x as much work as direct
   programming, however more quite flexible and easier extendable. Very fast
   development for bigger dialogue screens (specially if you have a template
   application to copy the routines from, like QuoSetup).

  A friend of mine who had a look at it didn't like it. He preferred a custom
  made dialogue routine (
   - more flexible queries,
   - hiding/unhiding questions (variable no of questions depending on input)
   - more intuitive cursor movement)

  He did however like the EditArray feature a lot.

  I agree partially with him, but I know how much work it is (from programming
  the helpfile routines). These routines are easy to use, quickly updated etc
  etc. You can always switch to a custom routine lateron, when the concept has
  moved past the initial stage.

This module is quite easy to use, but difficult to explain. See the
examples. E.g. QuoSetup }

INTERFACE

{$I Syseq.inc}

Uses EWindow,ELib,

 {$IFDEF Go32V2}
    Go32,
 {$ENDIF}
  Crt,
  {$IFDEF Turbo}
   BPGo32,
  {$ENDIF}
  EPasStr;
{$I keys.inc}

TYPE    ProcDisplayField  = PROCEDURE( Rec : LONGINT; Line : LONGINT; Color : BYTE );
        ProcManipRec      = PROCEDURE( VAR CurRec:LONGINT;VAR TotalRec:LONGINT);
        ProcLeftOverKeys  = PROCEDURE( Key:WORD;VAR CRec,NRec: LONGINT);
        ProcBackGr        = PROCEDURE;
        ProcEditField     = PROCEDURE( Field:LONGINT;Rec:LONGINT );


{Almost all what is needed have a record edited}
        EditRec= RECORD
               EditField   : ProcEditField;    { Invoked when editing a field of
                                                   the record}

               InitEdit    : ProcBackGr;       { Display background before record
                                                 is displayed}
               DisplEdit   : ProcDisplayField; { Display the contents of the
                                                  fields}
               X1,Y1,X2,Y2 : LONGINT;         { dimensions of the selector window}
               Frame       : BOOLEAN;          { if true then Frame= On? and the
                                                 following fields get meaning}
               FrameType   : String[10];         { Frame : type}
               ForeFrame,                      { colors of Frame}
               BackFrame,
               BackCol,                        { Color of edit background}
               ForeCol  : BYTE ;               { Color of edit (Default)}
               END;


  SelectEditRec= RECORD
                  X1,Y1,X2,Y2 : LONGINT; { Dimensions of selector window}
                  Frame       : BOOLEAN;  { Frame On? If yes then }
                  FrameType   : String[10]; { Type of Frame }
                  ForeFrame,              { Colors of Frame }
                  BackFrame,
                  BackCol,                { Default colors of window}
                  ForeCol     : BYTE;
                  BackGrProc  : ProcBackGr;    { Display background text in window}
                  DisplLine   : ProcDisplayField;    { Display one record }
                  DelRecProc,             { Delete one record }
                  InsRecProc  : ProcManipRec;   { Insert one record }
                  ElseProc    : ProcLeftOverKeys;   { All other keys are handed to
                                              this proc }
                  Fields      : LONGINT; { # Fields in record}
                  VisLines    : LONGINT; { # Lines visible in window. }
                  MaxRec      : LONGINT; { # Maximum records in array }
                  END;

CONST YesNo=#2+'YES@NO';        {Simple togglefielddata for toggler}

{ ---------------- Some general purpose routines ---------- }

{ Input data in a field on X,Y,XLength. If TextIn is true then the
(old) value to edit is in "S". CharAllow contains the characters allowed in the
input}
FUNCTION InputStr(VAR S:String;X,Y,Len:LONGINT;TextIn:BOOLEAN;CharAllow:CHARSET):BOOLEAN;

{ Set Attributes from one byte, TextColor and TextBackground in one,
faster under Linux. }
PROCEDURE SetAttrs(Col : BYTE);

{ ------------- The edit procedures -------- }


VAR   NormalColor,                { Some colors required by the Edit }
      HilightColor              { procedures }
      {EditColor  }   : BYTE;
      FieldSpace    : BYTE;       { Character used as background in InputStr.
                                      Default=177. }

{ Manages a selector window. Display displays one record, see
demo for other information. Kind of primitive PullDown.SubMenu system.
Requires a lot more code for a lot less features compared to Submenu.
However if SubMenu is limiting in features, it can be easy to develop your
own system starting with Selector. Is also used by the other procedures
}
FUNCTION Selector( Display : ProcDisplayField; MaxLines, Lines : LONGINT; VAR First,
                                               Curr : LONGINT ) : LONGINT;


{ Procedure to easy make "toggle" fields of type byte. For use with
    DisplayEdit and EditField procedures}
PROCEDURE Toggler(VAR Value : BYTE;InString : String;Increment:BOOLEAN);
{$IFDEF Turbo}
PROCEDURE TogglerB(VAR Value : BOOLEAN;InString : String;Increment:BOOLEAN);
{$ELSE}
PROCEDURE Toggler(VAR Value : BOOLEAN;InString : String;Increment:BOOLEAN);
{$ENDIF}


{ Edit a record, with several fields}
PROCEDURE EditRecord( VAR Edit:EditRec; MaxField : LONGINT;Rec:LONGINT );

{ Edit an array of records. }
PROCEDURE EditArray(VAR R : SelectEditRec;VAR Edit :EditRec;VAR NrRec:LONGINT);

{ -------- Screen setup system -------- }


CONST NrItemMax=100; {Increase if necessary}

TYPE

      BackGrRec = RECORD
                   Data  : PChar;
                   Items : LONGINT;    { Strings in Data }
                   Indices  : ARRAY[0..NrItemMax] OF LONGINT;
                                        { Indices to the strings }
                   Bytes : LONGINT;
                   END;
    PBackGrType = ^BackGrRec;


{ ----------------------------------------------------------------
I use these procedures, to set up the procedures required as parameters for
the "EditRecord" procedure both quick, and beautiful. }

{ Initialize a screen constant }
FUNCTION  InitBigStr(TS : PChar;Bytes:LONGINT):PBackGrType;

{ Remove BigStr system from memory, you loose the pointer in "TS", if you
loaded the fielddata dynamically make sure you dispose it before this
procedure or keep another pointerreference }
PROCEDURE KillBigStr(VAR TS : PBackGrType);

{ Return pointer to NrFoc'ed string }
FUNCTION  GetStrPtr(TS: PBackGrType;NrFoc : LONGINT):PChar;

{ Build the entire TS screen }
PROCEDURE BuildScreenFromPtr(TS:PBackGrType);

IMPLEMENTATION

{ Generic routines }

VAR ins : BOOLEAN;

PROCEDURE OutputStr(S:String;X,Y,XLength:LONGINT;AColor:BYTE);
{XLength is maximal length. not implemented}

BEGIN
 TextColor(AColor);GotoXY(X,Y);Write(S);
END ;

PROCEDURE SetAttrs(Col : BYTE);
BEGIN
 {$IFDEF Linux}
  TextAttr:=(Col AND 15);
  TextBackground(Col SHR 4);            {force Ansi-out}
 {$ELSE}
  TextAttr:=Col;
 {$ENDIF}
END;

FUNCTION  InputStr(VAR S:String;X,Y,Len:LONGINT;TextIn:BOOLEAN;CharAllow:CHARSET):BOOLEAN;

{
  Input a string from keyboard, in a nice way,
   allowed characters are in CHARSET CharAllow, but several editting
   keys are always allowed, see CASE loop.

Theoretical limit of 80 characters, but this isn't tested, so you could reach
132 characters if you wish, but you'll have to make "UitG" bigger.

Parameters:

   X,Y       Coordinates field
   Len       Lengte field
   TextIn    S already filled?}

VAR
    InGev                     : LONGINT; { No. of chars inputted }
    Posi                      : LONGINT; { Cursorposition}
{    Ins                       : BOOLEAN;}  { Insert yes/no }
    Key                       : LONGINT; { Last key as ELib.GetKey
                                            code <255 if normal key,
                                            >256 if special/function
                                            key. See keys.inc}
    Uitg                      : String;    {The inputted string}
    Full                      : BOOLEAN;   { Is the string full? }
    EndVal                    : WORD;

PROCEDURE ReWr; { Rewrite the field, using Uitg}

VAR    I                         : LONGINT;  { Temporary variabele }

BEGIN
 IF Length(Uitg)>Len THEN
  Uitg[0]:=CHR(Len);
 IF Length(Uitg)>0 THEN
  FOR I:= 1 TO Length(Uitg) DO
   BEGIN
    GotoXY(X+I-1,Y);
    
    IF Uitg[I]=CHR(32) THEN
     Write(CHR(FieldSpace))
    ELSE
     Write(Uitg[I]);
   END;
 IF Len<>Length(Uitg) THEN
  BEGIN
   GotoXY(X+Length(Uitg),Y);
   FOR I:= Length(Uitg) TO Len-1 DO
    Write(CHR(FieldSpace));
  END;
END;

PROCEDURE DoCursor; { Put Cursor in/out insert-mode }

BEGIN
 {$IFNDEF Linux}
 IF Ins THEN
  SetCursorSize($11E)
 ELSE
  SetCursorSize($71E);
 {$ENDIF}
END;

BEGIN
    { Init }

  InGev :=0;              { 0 chars untill now }
  Posi  :=1;               { Cursorposition 0 }
  Ins   :=TRUE;            { Insert according to parameters }
  DoCursor;        { Set cursor accordingly }
  Key   :=0;

       { put  padded field on screen }

  StrStr(Uitg,CHR(FieldSpace),Len);
  ReWr;
  GotoXY(X,Y);

  StrStr(Uitg,CHR(0),80);                  { Empty Uitg entirely }

  IF TextIn THEN
   BEGIN
    Uitg:=S;
    Posi:=Length(Uitg)+1;                        { Put a predefined }
    ReWr;                                   {  String on screen if specified }
   END;

  EndVal:=0;
  WHILE EndVal=0 DO
   BEGIN
    Full:=FALSE;
    IF ((Posi)>=Len) THEN
     BEGIN
      Full:=TRUE;
      Posi:=Len;
     END;
    GotoXY(X+Posi-1,Y);
    {$IFNDEF Linux}
    CursorOn;
    DoCursor;
    {$ENDIF}
    Key:=GetKey;
   {$IFNDEF Linux}
    CursorOff;
   {$ENDIF}
    CASE Key OF
          CR              : BEGIN
                             EndVal:=1;
                             S:=UitG;
                            END;
          ESC             : EndVal:=2;
          BS              : IF Posi>1 THEN       { BackSpace }
                              BEGIN
                               DEC(Posi);
                               Delete(Uitg,Posi,1);
                               DEC(InGev);
                               ReWr;
                              END;
          KDelete          : BEGIN
                              Delete(Uitg,Posi,1);
                              DEC(InGev);
                              ReWr;
                             END;
          ArrR            : IF (NOT Full) AND ((Posi-1)<InGev) THEN
                              BEGIN
                               INC (Posi);
                               GotoXY(X+Posi-1,Y);
                               END;
          KInsert          : BEGIN
                               Ins:= NOT Ins;
                               DoCursor;
                              END;
          ArrL            : IF (NOT (Posi=1)) THEN
                              BEGIN
                               DEC (Posi);
                               GotoXY(X+Posi-1,Y);
                              END;
          Home            : Posi:=1;
          KEnd            : Posi:=InGev-1;
          CtrlY           : BEGIN
                             Delete(Uitg,Posi,Length(Uitg)-Posi);
                             ReWr;
                            END;
          CtrlT           : BEGIN
                             Uitg[0]:=#0; Posi:=1; ReWr;
                            END;
    END; {Case}
   IF EndVal=0 THEN
    BEGIN
     IF (CHR(Key) IN CharAllow) THEN
      BEGIN
       IF Posi>Len THEN
        Posi:=Len;
       IF (Ins=FALSE) OR Full THEN
        BEGIN
         IF (ORD(Uitg[0])<Posi) THEN
           Uitg[0]:=CHR(Posi);
         Uitg[Posi]:=CHR(Key);
        END
       ELSE
        BEGIN
          InsertC(Uitg,CHR(Key),Posi);
        END;
       ReWr;
       INC(Posi);
      END;
     END;
    InGev:=Length(Uitg);
   END;
  InputStr:=Endval=1;
END;

FUNCTION  Selector( Display : ProcDisplayField; MaxLines, Lines : LONGINT; VAR First, Curr : LONGINT ) : LONGINT;
VAR
   Key, Entry : LONGINT;
   Tmp        : BOOLEAN;

   PROCEDURE DisplayLines( First, Lines : LONGINT );
   VAR
      Line : LONGINT;
   BEGIN

      FOR Line := First TO First+Lines-1 DO
       BEGIN
        IF (Line AND 7)=0 THEN
         BEGIN
         {$IFDEF Go32V2}
          while ((InPortb($3DA) AND 8) > 0) do;
          while ((InPortb($3DA) AND 8) = 0) do;
         {$ENDIF}
         {$IFDEF Turbo}
          while ((Port[$3DA] AND 8) > 0) do;
          while ((Port[$3DA] AND 8) = 0) do;
         {$ENDIF}
        END;
       IF Line<=MaxLines THEN
         Display( Line, Line-First, NormalColor );
        END;
   END;

BEGIN
   {$IFNDEF Linux}
    CursorOff;
   {$ENDIF}
   IF Curr > (First+Lines-1) THEN
    First := Curr-Lines+1;
   DisplayLines( First, Lines );
   Entry := Curr;
   Tmp:=FALSE;
   REPEAT
      Display( Entry, Entry-First, HilightColor );
      Key := GetKey;
      Display( Entry, Entry-First, NormalColor );
      CASE Key OF
       ArrD : IF Entry < MaxLines THEN
                  INC( Entry )
               ELSE
                  Entry := 1;
       ArrU : IF Entry > 1 THEN
                  DEC( Entry )
               ELSE
                  Entry := MaxLines;
       PgDn : IF Entry < First+Lines-1 THEN
               BEGIN
                  Entry := First+Lines-1;
                  IF Entry > MaxLines THEN
                     Entry := MaxLines;
                END
               ELSE
                IF Entry+Lines > MaxLines THEN
                  Entry := MaxLines
               ELSE
                  INC( Entry, Lines );
       PgUp : IF Entry > First THEN
                  Entry := First
                 ELSE
                  IF INTEGER(Entry)-INTEGER(Lines) < 1 THEN
                   Entry := 1
                  ELSE
                   DEC( Entry, Lines );
       Home : Entry := 1;
       KEnd  : Entry := MaxLines;
       ESC  :  Curr := 0;
       ELSE
         BEGIN
          Curr := Entry;
          Tmp:=TRUE;
         END;
        END;
     IF NOT (Key=ESC) OR NOT Tmp THEN
      BEGIN

      IF Entry < First THEN
        BEGIN
         First := Entry;
         DisplayLines( First, Lines );
        END
       ELSE
        IF Entry > First+Lines-1 THEN
         BEGIN
          First := Entry-Lines+1;
          DisplayLines( First, Lines );
       END;
      END;
     UNTIL (Key=ESC) OR Tmp;
  {$IFNDEF Linux}
   SetCursorSize($607);
  {$ENDIF}
   Selector:=Key;
END;

PROCEDURE EditRecord( VAR Edit:EditRec; MaxField : LONGINT;Rec:LONGINT );
VAR
   Key, Field : LONGINT;
   WD         : WinDef;
   Win        : WinType;

BEGIN
   WD.X1:=Edit.X1;               WD.Y1:=Edit.Y1;
   WD.X2:=Edit.X2;               WD.Y2:=Edit.Y2;
   WD.Background:=Edit.BackCol;  WD.Foreground:=Edit.ForeCol;
{   WD.CursorOn:=FALSE;           WD.WrapOn:=FALSE;
   WD.Hidden:=FALSE;    }         WD.FrameOn:=Edit.Frame;
   IF Edit.Frame THEN
    BEGIN
     WD.FrameDef :=Edit.FrameType; WD.FrameFore:=Edit.ForeFrame;
     WD.FrameBack:=Edit.BackFrame;
    END;
   Win:=WinOpen(WD);
   {$IFNDEF Linux}
   CursorOff;
   {$ENDIF}
   Edit.InitEdit;
   FOR Field := 1 TO MaxField DO
    Edit.DisplEdit( Field, Rec, NormalColor );
   Field := 1;
   REPEAT
      Edit.DisplEdit( Field, Rec, HilightColor );
      Key := GetKey;
      Edit.DisplEdit( Field, Rec, NormalColor );
      CASE Key OF
       ArrD : IF Field < MaxField THEN
                  INC( Field )
               ELSE
                 Field := 1;
       ArrU : IF Field > 1 THEN
                  DEC( Field )
               ELSE
                  Field := MaxField;
       PgDn : Field := MaxField;
       PgUp : Field := 1;
       CR   : Edit.EditField( Field,Rec );
      END;
   UNTIL Key=ESC;
   {$IFNDEF Linux}
   SetCursorSize($607);
   {$ENDIF}
   WinClose(Win);
END;

PROCEDURE Toggler(VAR Value : BYTE;InString : String;Increment:BOOLEAN);

VAR SubStr : String;

BEGIN
 IF Increment THEN
  Value:=(Value+1) MOD ORD(InString[1]);
 ItemS(SubStr,InString,'@',Value);
 IF Value=0 THEN
  Delete(SubStr,1,1);
 Write(SubStr);
END;

{$IFDEF Turbo}
PROCEDURE TogglerB(VAR Value : BOOLEAN;InString : String;Increment:BOOLEAN);
{$ELSE}
PROCEDURE Toggler(VAR Value : BOOLEAN;InString : String;Increment:BOOLEAN);
{$ENDIF}

VAR SubStr : String;
    OrdVal : WORD;

BEGIN
 OrdVal:=0;
 IF Increment THEN
  Value:=NOT Value;
 IF Value THEN
  OrdVal:=1;
 ItemS(SubStr,InString,'@',OrdVal);
 IF NOT Value THEN
  Delete(SubStr,1,1);           {Remove the leading number of items}
 Write(SubStr);
END;



PROCEDURE EditArray(VAR R : SelectEditRec;VAR Edit :EditRec;VAR NrRec:LONGINT);

VAR First,CRec,
    Key       : LONGINT;
    Win       : WinType;
    WD        : WinDef;

BEGIN
 WD.X1:=R.X1;               WD.Y1:=R.Y1;
 WD.X2:=R.X2;               WD.Y2:=R.Y2;
 WD.Background:=R.BackCol;  WD.Foreground:=R.ForeCol;
 WD.FrameOn:=R.Frame;
 IF R.Frame THEN
  BEGIN
  WD.FrameDef:=R.FrameType; WD.FrameFore:=R.ForeFrame;
  WD.FrameBack:=R.BackFrame;
  END;
 {$IFNDEF Linux}
 CursorOff;
 {$ENDIF}
 Win:=WinOpen(WD);
 Use(Win);
 R.BackGrProc;
 First:=1;
 CRec:=First;
 WITH R DO
  BEGIN
   REPEAT
      Key := Selector(DisplLine, NrRec, R.VisLines, First, CRec);
      CASE Key OF
        CR     : BEGIN
                  EditRecord( Edit, R.Fields ,CRec);
                  Use(Win);
                 END;
        KDelete : IF Assigned(DelRecProc) THEN
                   DelRecProc( CRec, NrRec );
        KInsert : IF NrRec<R.MaxRec THEN
                   IF Assigned(InsRecProc) THEN
                    BEGIN
                    InsRecProc( CRec, NrRec );
                    EditRecord( Edit, R.Fields,NrRec );
                    Use(Win);
                    CRec:=NrRec;
                    END;
         ESC    : Key:=ESC;                     {To avoid processing by ELSE}
      ELSE
       IF Assigned(ElseProc) THEN
        ElseProc(Key,CRec,NrRec);
      END;
   UNTIL Key=ESC;
 END;
 WinClose(Win);
 {$IFNDEF Linux}
 CursorOn;
 {$ENDIF}
END;


{ --------     Screen setup system -------- }
FUNCTION InitBigStr(TS : PChar;Bytes:LONGINT):PBackGrType;

VAR SysDat : PBackGrType;
    Nr     : LONGINT;
    Ind    : LONGINT;
    Fl     : BOOLEAN;

BEGIN
 NEW(SysDat);
 SysDat^.Data:=TS;
 Nr:=0; Ind:=0;
 Sysdat^.bytes:=Bytes;
 Fl:=FALSE;

 WHILE (TS[Ind]<>CHR(0)) AND NOT Fl DO
  BEGIN
   SysDat^.Indices[Nr]:=Ind;
   INC(Nr);
   INC(Ind,4);
   IF Ind>Bytes THEN  { safety }
    Fl:=TRUE;
   IF NOT Fl THEN
     Ind:=Ind+ScanR(TS[Ind],0,Bytes-Ind)+1;
  END;
 SysDat^.Items:=Nr;
 InitBigStr:=SysDat;
END;

PROCEDURE KillBigStr(VAR TS : PBackGrType);
BEGIN
 DISPOSE(TS);
END;

FUNCTION GetStrPtr(TS: PBackGrType;NrFoc : LONGINT):PChar;

BEGIN
 GetStrPtr:= @TS^.Data[TS^.Indices[NrFoc]];
END;

PROCEDURE BuildScreenFromPtr(TS:PBackGrType);

VAR A : LONGINT;
    P : PChar;

BEGIN
 FOR A:=0 TO TS^.Items-1 DO
  BEGIN
   P:=GetStrPtr(TS,A);
   GotoXY(ORD(P[0]),ORD(P[1]));
   SetAttrs(ORD(P[3]));
   P:=@P[4];
   Write(P);
  END;
END ;

BEGIN
   NormalColor  := LightGray+Black*16;
   HilightColor := Black+White*16;
   {EditColor    := Yellow+Green*16;}
   FieldSpace:=177;
END.

