{ example for : DOS memory access
		Translation services
		registers record
		rm_addr record
		tb_size()
		tb_address()
		seg_memcpy()
		dpmi_call_rm_procedure_with_retf_frame()				

This demo (C) 1998 by Thomas Schatzl

Author:               Alexander Stohr (AlexS@Freepage.de), contributed to
                      the Tom32 package.
License               LGPL, see the file docs/COPYING.TXT, included in
                      this distribution for details about the copyright.
Description           Calls a 16 bit real mode procedure in protected mode
                      via the translation services provided by DPMI. Uses
                      the internal transfer buffer to store the procedure
                      temporarily in DOS memory (a unique DOS memory block
                      could be used too. See the call16.pas example how to
                      do this).


  To compile the provided 16 bit asm code used by this example, you need to
  issue the following steps :

  tasm dos16.asm
  tlink /t dos16
  ren dos16.com dos16.bin

  A precompiled DOS16.BIN should be included within this package.
}

program DosCall16;

uses dpmi, memory;

const proc_org = $100; { we set the code origin of the real mode proc to
                         $100 (equals the 'org 100h' directive in dos16.asm) }

var Reg16 : Registers; { the registers record used for calling 
			 the real mode procedure  }
    rm_Mem : rm_Addr;  { storage for the calculated real mode memory address 
			 where the real mode procedure is copied to }	

    l : DWord;         { temporary variable }
    pM : Pointer;
    Fin : File Of Byte;

begin
     { check if DOS16.BIN exists anywhere in the path and error out if not }
     Assign(Fin,'DOS16.BIN');
     Reset(Fin);
     if (ioresult<>0) then begin
        Writeln('Error accessing DOS16.BIN containing the real mode procedure to call. Halting.');
        halt;
     end;
     l := filesize(fin);

     { check if the internal transfer buffer is big enough for our proc. If it
       isn't you need to allocate a DOS memory via DPMI as shown in the call16.pas
       example
     }
     if ((l + proc_org) > tb_size) then begin
        Writeln('The transfer buffer is too small for your procedure. Use a unique DOS memory');
        Writeln('block for your program. Halting.');
        halt;
     end;

     { we need to use a temporary buffer on heap to store the real mode code
       since it isn't possible to read it directly to DOS memory. Then we move
       it to our desired address and de-allocate the temporary buffer on heap
       again
     }
     getmem(pM, l);
     blockread(Fin, pM^, l);
     seg_memcpy(dseg, DWord(pM), fseg, tb_address + proc_org, l);
     freemem(pM, l);

     { close the file afterwards }
     close(Fin);

    { set up the real mode address where we want to copy the called real mode
      procedure too. Since the address returned by tb_address() is a linear
      address, we have to convert it to a valid real mode segment:offset
      address.
     }
     rm_Mem.segment := tb_address shr 4;
     rm_Mem.offset  := tb_address and $f;
     { set up the parameters for the procedure. Since the data used by the
       real mode procedure lies in its code segment, we can directly set
       DS to the code segment of the real mode code.
     }
     Reg16.DS := rm_Mem.Segment;
     Reg16.CS := rm_Mem.Segment;
     Reg16.IP := rm_Mem.Offset + proc_org;
     { Execute the real mode code }
     dpmi_call_rm_procedure_with_retf_frame(Reg16);
end.

