{ Example for : seg_memcpy()
                seg_memset()
                DOS memory access

This demo (C) 1998 by Thomas Schatzl

Author:               Thomas Schatzl
License               LGPL, see the file docs/COPYING.TXT, included in
                      this distribution for details about the copyright.
Description           This example copies around some blocks of memory in
                      DOS memory space.

  In more detail, the program copies a string randomly to the text mode
  screen. Aditionally it messes around a bit with the color attributes of
  the string.
  Before doing this it saves the entire screen contents to the heap and
  restores it afterwards.

  Some additional background:

  The text screen of a VGA card has it's address space at $B800:0; screen
  memory is organized in a linear fashion, e.g. the second line comes
  directly after the first, where each cell occupies 2 bytes of memory
  (1 byte character data, 1 byte attributes). It is 32 kb in size.

  Hence the offset of a single memory cell from its origin is:

     Y*columns*2 + X*2

  where X and Y mark the point and columns is the number of character cells
  per line
}

uses
	crt, { keypressed(), gotoxy(), randomize(), random(), wherex(),
            wherey() }
    memory, dpmi;

const
	text = '! Hello world !'; { sample text string }

var
	columns, rows : Longint;
    screensize : Longint;
	textofs : Longint;
    save_screen : ^byte; { this variable holds the entire screen contents }
    curx, cury : Integer; { These two hold the previous cursor coordinates }

begin
	 { initialize some variables }
	 columns := lo(windmax) + 1;
     rows := hi(windmax) + 1;
     screensize := columns * rows * 2;

     getmem(save_screen, screensize);

     randomize;
     { save screen contents to save_screen variable }
     seg_memcpy(fseg, $B8000, dseg, DWord(save_screen), screensize);
     { save current cursor coordinates }
     curx := wherex; cury := wherey;
     { This is our demo text }
     gotoxy(1, 1); Write(text);
     { calculate the address in offscreen memory (to be sure it won't be over
       written by the copy process later, we don't put it exactly at the end
       of the visible screen area) }
     textofs := screensize + length(text)*2;
     { copy it to offscreen memory }
     seg_memcpy (fseg, $B8000, fseg, $B8000 + textofs, length(text)*2);
     { clear the screen by writing zeros on the whole visible screen }
     seg_memsetb(fseg, $B8000, screensize, 0);
     while (not keypressed) do begin
           { set the attribute field (byte 2 of every cell) of the text in
             offscreen memory to random values }
           seg_memsetb(fseg, $B8000 + textofs + random(length(text))*2 + 1,
                       1, random(255));
           { copy the string from offscreen to visibly screen by calculating
             it's destination address randomly }
           seg_memcpy(fseg, $B8000 + textofs,
                      fseg, $B8000 + random(columns)*2+random(rows)*columns*2,
                      length(text)*2);
     end;
     { clear the keyboard buffer }
     readkey;
     { wait for a keypress }
     readkey;
     { restore old screen contents afterwards }
     seg_memcpy(dseg, dword(save_screen), fseg, $B8000, screensize);
     gotoxy(curx, cury);

     freemem(save_screen, screensize);
end.
