{ example for :
     Executing software interrupts
     tb_size()
     tb_address()
     trealregs / registers record
     realintr()
     DOS memory access
     copyfromdos()
     copytodos()

This demo (C) 1998 by Thomas Schatzl

Author:               Thomas Schatzl
License               LGPL, see the file docs/COPYING.TXT, included in
                      this distribution for details about the copyright.
Description           This program demonstrates the usage of DOS real mode
                      memory by executing a software interrupt which needs
                      a buffer to store data into. Because these interrupts
                      are real mode funcs, the buffer must be located in
                      real mode memory space (first MB of memory). Contrary
                      to BUFFER.PAS it doesn't allocate a new DOS memory
                      buffer, but uses the built-in transfer buffer.

  In more detail this program tries to detect a VESA 2.0 BIOS extension of
  your graphics card and outputs its version.

  Here's the necessary interrupt call description:

  Int 10h 4f00h : VESA BIOS extension installation check
  Input : AX = 4F00h
          ES:DI = pointer to 512 byte information buffer
  Output : AX = 004Fh if successful
           ES:DI = pointer to filled buffer

  Buffer structure : (relevant to this example)

           must be 'VESA' in the first 4 chars of the buffer to be valid
           VBE version in the next word

  Note : to request VBE 2.0 information, the first 4 bytes of the buffer must
         contain 'VBE2' prior to the interrupt call.

         (this makes the problem a bit tougher; we first have to copy the
         buffer with the 'VBE2' id to dos memory...)
}

uses memory, dpmi;

type VBEInfoBuf = record
                Signature : array[0..3] of char; { contains 'VESA' if successful }
                Version : Word;
                reserved : array[0..505] of byte; { pad to 512 bytes length }
     end;

var
    r : registers;  { register structure to issue a software interrupt }
    infobuf : VBEInfoBuf;

begin
     if (tb_size < sizeof(VBEInfoBuf)) then begin
        Writeln('Transfer buffer is too small to use');
        { In this case we would have to allocate our own DOS memory  }
        halt;
     end;
     { first we reset the registers and infobuf variable }
     memsetb(r, sizeof(r), 0);
     memsetb(infobuf, sizeof(VBEInfoBuf), 0);
     { request VBE 2.0 information, fill out information buffer }
     infobuf.Signature := 'VBE2';
     { copy buffer to the allocated real mode memory }
     seg_memcpy(dseg, dword(@infobuf), fseg, tb_address, sizeof(infobuf));
     { issue the interrupt }
     r.ax := $4f00;
     { convert the linear address returned by transfer_buffer() to a real
      mode segment:offset address }
     r.es := tb_address div 16;
     r.di := tb_address and $F;
     realintr($10, r);
     { check if interrupt call was successful }
     if (r.ax <> $4f) then begin
        { write message and exit, because the infobuf doesn't contain any
          useful data we could tell the user }
        Writeln('VBE BIOS extension not available, function call failed');
        halt;
     end;
     { copy buffer to our infobuf variable again }
     seg_memcpy(fseg, tb_address, dseg, dword(@infobuf), sizeof(infobuf));
     { check if buffer is valid }
     if (infobuf.signature[0] = 'V') and (infobuf.signature[1] = 'E') and
        (infobuf.signature[2] = 'S') and (infobuf.signature[3] = 'A') then begin
        Writeln('VBE version ', hi(infobuf.version), '.', lo(infobuf.version), ' detected');
     end;
end.